/*
 * $Id: bullet.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import javax.media.opengl.GL;

import abagames.util.ActorPool;
import abagames.util.Vector2;
import abagames.util.Vector3;

//private import std.math;
//private import opengl;
//private import abagames.util.vector;
//private import abagames.util.actor;
//private import abagames.util.math;
//private import abagames.ttn.token;
//private import abagames.ttn.field;
//private import abagames.ttn.shape;
//private import abagames.ttn.player;
//private import abagames.ttn.particle;
//private import abagames.ttn.enemy;
//private import abagames.ttn.screen;

/**
 * Enemies' bullets.
 */
class BulletPool extends ActorPool<Bullet> {
	public BulletPool() {
		super(new ActorCreator<Bullet>() {
			public Bullet create() {
				return new Bullet();
			}
			public Bullet[] createArray(int n) {
				return new Bullet[n];
			}
		});
	}

	public void move() {
		super.move();
		BulletState.move();
	}
}

public class Bullet extends Token<BulletState, BulletSpec> {
	public void init(Object[] args) {
		super.init(args);
		state = new BulletState();
	}
	
	public void setWaitCnt(int c) {
		state.waitCnt = c;
	}
}

class BulletState extends TokenState {
	static int colorCnt = 0;
	static float colorAlpha = 0;
	Vector2 ppos;
	Vector2 tailPos;
	int cnt;
	int waitCnt;
	float speedRatio;

	public BulletState() {
		super();
		ppos = new Vector2();
		tailPos = new Vector2();
	}

	public static void move() {
		colorCnt++;
		int c = colorCnt % 30;
		if (c < 15)
			colorAlpha = (float) c / 15;
		else
			colorAlpha = 1 - (float) (c - 15) / 15;
	}

	public void clear() {
		ppos.x = ppos.y = 0;
		tailPos.x = tailPos.y = 0;
		cnt = 0;
		waitCnt = 0;
		speedRatio = 0;
		super.clear();
	}
}

class BulletSpec extends TokenSpec<BulletState> {
	static final float DISAPPEAR_CNT = 300;
	Player player;
	EnemyPool enemies;
	ParticlePool particles;
	Shape lineShape;

	public BulletSpec(Field field, Player player, EnemyPool enemies, ParticlePool particles,
			Shape shape, Shape lineShape) {
		this.field = field;
		this.player = player;
		this.enemies = enemies;
		this.particles = particles;
		this.shape = shape;
		this.lineShape = lineShape;
	}

	public void set(BulletState bs) {
//		with (bs) {
	{
		bs.ppos.x = bs.pos.x;
		bs.ppos.y = bs.pos.y;
		bs.tailPos.x = bs.pos.x;
		bs.tailPos.y = bs.pos.y;
//		assert(deg <>= 0);
	}
	}

	public boolean move(BulletState bs) {
//		with (bs) {
		{
			if (bs.waitCnt > 0) {
				bs.waitCnt--;
				return true;
			}
			bs.ppos.x = bs.pos.x;
			bs.ppos.y = bs.pos.y;
			float sp = bs.speed;
			if (bs.cnt < 30)
				sp *= ((float) (bs.cnt + 10) / 40);
			bs.tailPos.x -= Math.sin(bs.deg) * sp * 0.7f;
			bs.tailPos.y += Math.cos(bs.deg) * sp * 0.7f;
			bs.pos.x -= Math.sin(bs.deg) * sp;
			bs.pos.y += Math.cos(bs.deg) * sp;
			field.addSlowdownRatio(bs.speed * 0.04f);
			bs.pos.x = field.normalizeX(bs.pos.x);
			if (!field.containsOuter(bs.pos))
				return false;
			if (!field.contains(bs.pos) || bs.cnt >= DISAPPEAR_CNT * 0.9f) {
				bs.tailPos.x += (bs.pos.x - bs.tailPos.x) * 0.1f;
				bs.tailPos.y += (bs.pos.y - bs.tailPos.y) * 0.1f;
			}
			bs.tailPos.x = field.normalizeX(bs.tailPos.x);
			if (enemies.checkBulletHit(bs.pos, bs.ppos))
				return false;
			if (player.checkBulletHit(bs.pos, bs.ppos))
				return false;
			bs.cnt++;
			if (bs.cnt >= DISAPPEAR_CNT)
				return false;
			return true;
		}
	}

	public void draw(GL gl, BulletState bs) {
//		with (bs) {
		{
			if (bs.waitCnt > 0)
				return;
			Vector3 p;
			gl.glBegin(GL.GL_LINES);
			Screen.setColor(gl, 0.1f, 0.4f, 0.4f, 0.5f);
			p = field.calcCircularPos(bs.tailPos);
			Screen.glVertex(gl, p);
			Screen.setColor(gl, 0.2f * BulletState.colorAlpha, 0.8f * BulletState.colorAlpha, 0.8f * BulletState.colorAlpha);
			p = field.calcCircularPos(bs.pos);
			Screen.glVertex(gl, p);
			gl.glEnd();
			p = field.calcCircularPos(bs.pos);
			float cd = field.calcCircularDeg(bs.pos.x);
			((BulletShapeBase)shape).draw(gl, p, cd, bs.deg, bs.cnt * 3.0f);
			Screen.setColor(gl, 0.6f * BulletState.colorAlpha, 0.9f * BulletState.colorAlpha, 0.9f * BulletState.colorAlpha);
			((BulletShapeBase)lineShape).draw(gl, p, cd, bs.deg, bs.cnt * 3.0f);
		}
	}
}
