/*
 * $Id: enemy.d,v 1.2 2006/11/23 02:29:43 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import javax.media.opengl.GL;

import abagames.util.ActorPool;
import abagames.util.MathUtil;
import abagames.util.Rand;
import abagames.util.Vector2;
import abagames.util.Vector3;

/**
 * Enemies and turrets.
 */
class EnemyPool extends ActorPool<Enemy> {
	public static boolean trailEffect = false;
	Field _field;

	public EnemyPool() {
		super(new ActorPool.ActorCreator<Enemy>() {
			public Enemy create() {
				return new Enemy();
			}
			public Enemy[] createArray(int n) {
				return new Enemy[n];
			}
		});
	}
	public Enemy getNearestEnemy(Vector2 p) {
		float dst = 99999;
		Enemy ne = null;
		for (Enemy e: actors) {
			if (e.exists() && !e.isBeingCaptured())
				if (_field.calcCircularDist(e.pos(), p) < dst) {
					dst = _field.calcCircularDist(e.pos(), p);
					ne = e;
				}
		}
		return ne;
	}

	public boolean checkShotHit(Vector2 p, float deg) {
		return checkShotHit(p, deg, 1.0f);
	}

	public boolean checkShotHit(Vector2 p, float deg, float widthRatio) {
		Enemy e = getNearestEnemy(p);
		if (e != null) {
			float ox = _field.normalizeX(e.pos().x - p.x);
			float oy = e.pos().y - p.y;
			if (Math.abs(ox) < 1.0f * e.state.size.x && Math.abs(oy) < 1.0f * e.state.size.y * widthRatio) {
				e.hitShot(deg);
				return true;
			}
		}
		return false;
	}

	public boolean checkBulletHit(Vector2 p, Vector2 pp) {
		boolean hitf = false;
		for (Enemy e: actors) {
			if (e.exists() && e.isCaptured())
				if (_field.checkHitDist(e.pos(), p, pp, EnemySpec.BULLET_HIT_WIDTH)) {
					e.hitCaptured();
					hitf = true;
				}
		}
		return hitf;
	}

	public boolean checkEnemyHit(Vector2 p, Vector2 size) {
		boolean hitf = false;
		for (Enemy e: actors) {
			if (e.exists() && e.isCaptured()) {
				float ox = _field.normalizeX(e.pos().x - p.x);
				float oy = e.pos().y - p.y;
				if (Math.abs(ox) < 0.5f * (e.state.size.x + size.x) &&
						Math.abs(oy) < 0.5f * (e.state.size.y + size.y)) {
					e.hitCaptured();
					hitf = true;
				}
			}
		}
		return hitf;
	}

	public int num() {
		int n = 0;
		for (Enemy e: actors)
			if (e.exists() && !e.isCaptured())
				n++;
		return n;
	}

	public int numInAttack() {
		int n = 0;
		for (Enemy e: actors)
			if (e.exists() && e.isInAttack())
				n++;
		return n;
	}

	public int numInScreen() {
		int n = 0;
		for (Enemy e: actors)
			if (e.exists() && e.isInScreen())
				n++;
		return n;
	}

	public void drawFront(GL gl) {
		if (trailEffect)
			for (Enemy a: actors)
				if (a.exists() && a.state.pos.y <= _field.size().y * 1.5f)
					a.drawTrails(gl);
		for (Enemy a: actors)
			if (a.exists() && a.state.pos.y <= _field.size().y * 1.5f)
				a.draw(gl);
	}

	public void drawBack(GL gl) {
		if (trailEffect)
			for (Enemy a: actors)
				if (a.exists() &&
						a.state.pos.y > _field.size().y * 1.5f &&
						(a.state.pos.x <= _field.circularDistance() / 4 &&
								a.state.pos.x >= -_field.circularDistance() / 4))
					a.drawTrails(gl);
		for (Enemy a: actors)
			if (a.exists() &&
					a.state.pos.y > _field.size().y * 1.5f &&
					(a.state.pos.x <= _field.circularDistance() / 4 &&
							a.state.pos.x >= -_field.circularDistance() / 4))
				a.draw(gl);
	}

	public void drawPillarBack(GL gl) {
		if (trailEffect)
			for (Enemy a: actors)
				if (a.exists() &&
						a.state.pos.y > _field.size().y * 1.5f &&
						(a.state.pos.x > _field.circularDistance() / 4 ||
								a.state.pos.x < -_field.circularDistance() / 4))
					a.drawTrails(gl);
		for (Enemy a: actors)
			if (a.exists() &&
					a.state.pos.y > _field.size().y * 1.5f &&
					(a.state.pos.x > _field.circularDistance() / 4 ||
							a.state.pos.x < -_field.circularDistance() / 4))
				a.draw(gl);
	}

	public Field field(Field v) {
		return _field = v;
	}
}

public class Enemy extends Token<EnemyState, EnemySpec> {
	public void init(Object[] args) {
		state = new EnemyState();
//		spec = new EnemySpec();
		super.init(args);
		state.enemy = this;
	}

	public void setSmallEnemyState(float baseSpeed, float angVel, int waitCnt, int appPattern) {
		setSmallEnemyState(baseSpeed, angVel, waitCnt, appPattern, 0, 0, false, null);
	}
	public void setSmallEnemyState(float baseSpeed, float angVel, int waitCnt, int appPattern,
			float er, float ed, boolean gd) {
		setSmallEnemyState(baseSpeed, angVel, waitCnt, appPattern, 0, 0, gd, null);

	}
	public void setSmallEnemyState(float baseSpeed, float angVel, int waitCnt, int appPattern,
			float er, float ed, boolean gd,
			Enemy firstEnemy) {
		state.baseBaseSpeed = state.baseSpeed = baseSpeed;
		state.baseAngVel = state.angVel = angVel;
		state.waitCnt = waitCnt;
		state.ellipseRatio = er;
		state.ellipseDeg = ed;
		state.isGoingDownBeforeStandBy = gd;
		switch (appPattern) {
		case 0:
			state.phase = -200;
			break;
		case 1:
			state.phase = -100;
			break;
		}
		if (firstEnemy != null) {
			((SmallEnemySpec) spec).init(state, firstEnemy.state);
			state.isFirstEnemy = false;
		} else {
			spec.init(state);
			state.isFirstEnemy = true;
		}
	}

	public void setMiddleEnemyState(float baseSpeed, float angVel) {

		setMiddleEnemyState(baseSpeed, angVel, 0, 0);
	}
	public void setMiddleEnemyState(float baseSpeed, float angVel,
			float er, float ed) {
		state.baseBaseSpeed = state.baseSpeed = baseSpeed;
		state.baseAngVel = state.angVel = angVel;
		state.ellipseRatio = er;
		state.ellipseDeg = ed;
		spec.init(state);
	}

	public void hitShot() {
		hitShot(0);
	}
	public void hitShot(float deg) {
		if (spec.hitShot(state, deg))
			remove();
	}

	public void hitCaptured() {
		SmallEnemySpec ses = (SmallEnemySpec) spec;
		if (ses != null)
			ses.hitCaptured(state);
	}

	public void destroyed() {
		spec.destroyed(state);
		_exists = false;
	}

	public boolean isInAttack() {
		if (spec.isBeingCaptured(state))
			return false;
		return spec.isInAttack(state);
	}

	public boolean isInScreen() {
		if (spec.isBeingCaptured(state))
			return false;
		return spec.isInScreen(state);
	}

	public boolean isBeingCaptured() {
		return spec.isBeingCaptured(state);
	}

	public boolean isCaptured() {
		if (spec instanceof SmallEnemySpec) {
			SmallEnemySpec ses = (SmallEnemySpec) spec;
			return ses.isCaptured(state);
		} else {
			return false;
		}
	}

	public void drawTrails(GL gl) {
		spec.drawTrails(gl, state);
	}

	public Vector2 pos() {
		return state.pos;
	}
}

class EnemyState extends TokenState {
	static final int TRAIL_NUM = 64;
	static final int TRAIL_INTERVAL = 8;
	TurretState turretState;
	Enemy enemy;
	Vector2 vel;
	Vector2 centerPos, centerVel, standByPos;
	float baseBaseSpeed, baseSpeed;
	float baseAngVel, angVel;
	int waitCnt;
	int cnt;
	float turretDeg;
	float ellipseRatio, ellipseDeg;
	float shield;
	int phase;
	int phaseCnt, nextPhaseCnt;
	int captureState, captureIdx;
	boolean isGoingDownBeforeStandBy;
	Vector2 size, targetSize, sizeVel;
	Trail[] trails;
	int trailIdx;
	boolean trailLooped;
	boolean isFirstEnemy;

//	invariant {
//	if (isInitialized) {
//	assert(baseBaseSpeed >= 0);
//	assert(baseSpeed >= 0);
//	assert(baseAngVel >= 0);
//	assert(angVel >= 0);
//	assert(turretDeg <>= 0);
//	assert(centerPos.x <>= 0);
//	assert(centerPos.y <>= 0);
//	assert(centerVel.x <>= 0);
//	assert(centerVel.y <>= 0);
//	assert(shield <>= 0);
//	assert(captureState >= 0);
//	assert(size.x <>= 0);
//	assert(size.y <>= 0);
//	assert(targetSize.x <>= 0);
//	assert(targetSize.y <>= 0);
//	assert(sizeVel.x <>= 0);
//	assert(sizeVel.y <>= 0);
//	}
//	}

	public EnemyState() {
		super();
		turretState = new TurretState();
		vel = new Vector2();
		centerPos = new Vector2();
		centerVel = new Vector2();
		standByPos = new Vector2();
		size = new Vector2();
		targetSize = new Vector2();
		sizeVel = new Vector2();
		trails = new Trail[TRAIL_NUM];
		for (int i = 0; i < TRAIL_NUM; i++)
			trails[i] = new Trail();
	}

	public void clear() {
		turretState.clear();
		vel.x = vel.y = 0;
		centerPos.x = centerPos.y = 0;
		centerVel.x = centerVel.y = 0;
		standByPos.x = standByPos.y = 0;
		baseBaseSpeed = baseSpeed = 0;
		baseAngVel = angVel = 0;
		waitCnt = 0;
		cnt = 0;
		turretDeg = 0;
		ellipseRatio = 0;
		ellipseDeg = 0;
		shield = 0;
		phase = 0;
		phaseCnt = nextPhaseCnt = 0;
		captureState = 0;
		captureIdx = 0;
		isGoingDownBeforeStandBy = false;
		size.x = size.y = 1;
		targetSize.x = targetSize.y = 1;
		sizeVel.x = sizeVel.y = 0;
		trailIdx = 0;
		trailLooped = false;
		isFirstEnemy = false;
		super.clear();
	}

	public void move() {
		cnt++;
	}

	public void recordTrail() {
		trails[trailIdx].set(pos.x, pos.y, deg, cnt);
		trailIdx++;
		if (trailIdx >= TRAIL_NUM) {
			trailIdx = 0;
			trailLooped = true;
		}
	}

	public void drawTrails(GL gl, EnemyShape s, float r, float g, float b, Vector2 size, Field field) {
		int ti = trailIdx;
		float a = 1.0f;
		for (int i = 0; i < TRAIL_NUM / TRAIL_INTERVAL; i++) {
			ti -= TRAIL_INTERVAL;
			if (ti < 0) {
				if (trailLooped)
					ti += TRAIL_NUM;
				else
					break;
			}
			Trail t = trails[ti];
			Screen.setColor(gl, r * a, g * a, b * a, a * 0.66f);
			Vector3 p = field.calcCircularPos(t.pos);
			float cd = field.calcCircularDeg(t.pos.x);
			s.draw(gl, p, cd, t.deg, t.cnt, size);
			a *= 0.7f;
		}
	}
}

//class EnemySpec<T extends EnemyState> extends TokenSpec<T> {
abstract class EnemySpec extends TokenSpec<EnemyState> {
//	mixin StaticRandImpl;
	public static Rand rand = new Rand();
	static final float BULLET_HIT_WIDTH = 0.8f;
	static final float NEXT_PHASE_DIST = 5;
	BulletPool bullets;
	Player player;
	ParticlePool particles;
	EnemyPool enemies;
	Stage stage;
	EnemyShape trailShape;
	BulletSpec bulletSpec;
	TurretSpec turretSpec;
	float shield = 1;
	float rank = 0;
	boolean capturable;
	int score;
	boolean aimAllDirection = false;
	float aimDirectionRange = (float)Math.PI / 4;
	String explosionSeName;

//	invariant {
//	assert(shield > 0);
//	assert(rank >= 0);
//	assert(aimDirectionRange >= 0);
//	}

	public EnemySpec(Field field, BulletPool bullets, Player player,
			ParticlePool particles, EnemyPool enemies, Stage stage,
			Shape shape, EnemyShape trailShape, BulletSpec bulletSpec) {
		this.field = field;
		this.bullets = bullets;
		this.player = player;
		this.particles = particles;
		this.enemies = enemies;
		this.stage = stage;
		this.shape = shape;
		this.trailShape = trailShape;
		this.bulletSpec = bulletSpec;
	}

	public void set(EnemyState es) {
		es.shield = shield;
	}

	public boolean move(EnemyState es) {
//		with (es) 
		{
			es.move();
			if (isInScreen(es) && es.isFirstEnemy) {
				Sound.playSe("flying_down.wav");
				es.isFirstEnemy = false;
			}
			if (es.captureState > 0) {
				moveCaptured(es);
				return true;
			}
			if (enemies.checkEnemyHit(es.pos, es.size)) {
				destroyed(es);
				return false;
			}
			if (player.checkEnemyHit(es.pos, es.vel, es.size)) {
				destroyed(es);
				return false;
			}
			if (capturable)
				checkCaptured(es);
			float er = (1 - es.ellipseRatio) + (float)Math.abs(Math.sin(es.deg + es.ellipseDeg)) * es.ellipseRatio * 2;
			float rk = rank;
			es.vel.x -= Math.sin(es.deg) * es.speed * er * 0.1f * rk;
			es.vel.y += Math.cos(es.deg) * es.speed * er * 0.1f * rk;
			es.vel.opMulAssign(0.9f);
			es.pos.opAddAssign(es.vel);
			if (isInScreen(es))
				field.addSlowdownRatio(es.speed * 0.04f * rk);
			es.pos.x = field.normalizeX(es.pos.x);
			es.recordTrail();
			if (es.phase >= -50 && es.phase < 0 && !field.containsIncludingPit(es.pos))
				return false;
			if (es.waitCnt > 0) {
				es.waitCnt--;
			} else {
				Vector2 cp = es.centerPos;
				es.centerPos.x = field.normalizeX(es.centerPos.x);
				es.phaseCnt++;
				if (field.calcCircularDist(es.centerPos, es.pos) < NEXT_PHASE_DIST) {
					es.nextPhaseCnt--;
					if (es.nextPhaseCnt <= 0) {
						es.phase++;
						if (!gotoNextPhase(es))
							return false;
					}
				}
				cp.x = field.normalizeX(cp.x);
				float dst = field.calcCircularDist(cp, es.pos);
				es.speed += ((es.baseSpeed * (1 + dst * 0.1f)) - es.speed) * 0.05f;
				float av = es.angVel * rk;
				float td = (float)Math.atan2(field.normalizeX(-(cp.x - es.pos.x)), cp.y - es.pos.y);
				float ad = (float)MathUtil.normalizeDeg(td - es.deg);
				av *= (2.5f - er);
				if (ad > av || ad < -Math.PI * 0.8f)
					es.deg += av;
				else if (ad < -av)
					es.deg -= av;
				else
					es.deg = td;
//				assert(deg <>= 0);
				es.turretDeg = (float)Math.atan2(field.normalizeX(-(player.pos().x - es.pos.x)), player.pos().y - es.pos.y);
				if (!aimAllDirection) {
					if (es.turretDeg >= 0 && es.turretDeg < Math.PI - aimDirectionRange)
						es.turretDeg = (float)Math.PI - aimDirectionRange;
					else if (es.turretDeg < 0 && es.turretDeg > -Math.PI + aimDirectionRange)
						es.turretDeg = -(float)Math.PI + aimDirectionRange;
				}
				es.turretState.update(es.pos.x, es.pos.y, es.turretDeg);
				movePhase(es);
				es.sizeVel.x += (es.targetSize.x - es.size.x) * 0.2f;
				es.sizeVel.y += (es.targetSize.y - es.size.y) * 0.2f;
				es.size.opAddAssign(es.sizeVel);
				es.sizeVel.opMulAssign(0.95f);
			}
			return true;
		}
	}

	private void moveCaptured(EnemyState es) {
//		with (es) {
		{
			float cx;
			switch (es.captureState) {
			case 1:
				es.vel.x += (player.pos().x - es.pos.x) * 0.03f;
				es.vel.y += (player.pos().y - es.pos.y) * 0.03f;
				es.pos.x += (player.pos().x - es.pos.x) * 0.03f;
				es.pos.y += (player.pos().y - es.pos.y) * 0.03f;
				es.deg *= 0.95f;
				if (player.pos().dist(es.pos) < 1)
					es.captureState = 2;
				break;
			case 2:
				cx = calcCapturePosX(es.captureIdx);
				es.vel.x += (player.pos().x + cx - es.pos.x) * 0.03f;
				es.pos.x += (player.pos().x + cx - es.pos.x) * 0.1f;
				es.pos.y += (player.pos().y - es.pos.y) * 0.33f;
				es.vel.y *= 0.6f;
				es.deg *= 0.95f;
				if (Math.abs(player.pos().x + cx - es.pos.x) < 0.2f)
					es.captureState = 3;
				break;
			case 3:
				cx = calcCapturePosX(es.captureIdx);
				es.pos.x = player.pos().x + cx;
				es.pos.y = player.pos().y;
				es.deg = player.deg();
				break;
			}
			es.vel.opMulAssign(0.9f);
			es.pos.opAddAssign(es.vel);
		}
	}

	private float calcCapturePosX(int idx) {
		if (idx % 2 == 0)
			return ((idx / 2) + 0.5f) * PlayerSpec.CAPTURED_ENEMIES_INTERVAL_LENGTH * player.capturedEnemyWidth();
		else
			return -((idx / 2) + 0.5f) * PlayerSpec.CAPTURED_ENEMIES_INTERVAL_LENGTH * player.capturedEnemyWidth();
	}

	private void checkCaptured(EnemyState es) {
//		with (es) {
		{
			if (player.isInTractorBeam(es.pos)) {
				int idx = player.addCapturedEnemy(es.enemy);
				if (idx >= 0) {
					es.captureState = 1;
					es.captureIdx = idx;
				}
			}
		}
	}

	public void hitCaptured(EnemyState es) {
		player.destroyCapturedEnemies(es.captureIdx);
	}

	public boolean isBeingCaptured(EnemyState es) {
		return (es.captureState > 0);
	}

	public boolean isCaptured(EnemyState es) {
		return (es.captureState == 3);
	}

	public boolean hitShot(EnemyState es) {
		return hitShot(es, 0);
	}

	public boolean hitShot(EnemyState es, float dd) {
//		with (es) {
		{
			shield--;
			float r = 0.5f + rand.nextFloat(0.5f);
			float g = 0.1f + rand.nextFloat(0.5f);
			float b = 0.5f + rand.nextFloat(0.5f);
			for (int i = 0; i < 10; i++) {
				Particle p;
				float d;
				p = particles.getInstanceForced();
				d = dd + rand.nextSignedFloat((float)Math.PI / 4);
				p.set(Particle.LINE, es.pos.x, es.pos.y, d, 0.1f + rand.nextFloat(0.5f), 1,
						r, g, b, 30 + rand.nextInt(30));
				p = particles.getInstanceForced();
				d = dd + (float)Math.PI + rand.nextSignedFloat((float)Math.PI / 4);
				p.set(Particle.LINE, es.pos.x, es.pos.y, d, 0.1f + rand.nextFloat(0.5f), 1,
						r, g, b, 30 + rand.nextInt(30));
			}
			if (shield <= 0) {
				destroyed(es, dd);
				return true;
			}
			es.targetSize.x *= 1.15f;
			es.targetSize.y *= 1.15f;
			es.sizeVel.x = 0.3f;
			es.sizeVel.y = 0.3f;
			return false;
		}
	}

	public void destroyed(EnemyState es) {
		destroyed(es, 0);
	}

	public void destroyed(EnemyState es, float dd) {
//		with (es) {
		{
			float r = 0.5f + rand.nextFloat(0.5f);
			float g = 0.1f + rand.nextFloat(0.5f);
			float b = 0.5f + rand.nextFloat(0.5f);
			float sz = (es.size.x + es.size.y) / 2;
			sz = (sz - 1) * 2 + 1;
			int n = 3 + rand.nextInt(2);
			n *= sz;
			for (int i = 0; i < n; i++) {
				Particle p = particles.getInstanceForced();
				float d = dd + rand.nextSignedFloat((float)Math.PI / 5);
				p.set(Particle.TRIANGLE, es.pos.x, es.pos.y, d, 0.5f,
						(2 + rand.nextFloat(0.5f)) * sz, r, g, b, 50 + rand.nextInt(100));
			}
			for (int i = 0; i < n; i++) {
				Particle p = particles.getInstanceForced();
				float d = rand.nextFloat((float)Math.PI * 2);
				p.set(Particle.QUAD, es.pos.x, es.pos.y, d, 0.1f + rand.nextFloat(0.1f),
						(1 + rand.nextFloat(0.5f)) * sz, r, g, b, 50 + rand.nextInt(100));
			}
			if (!isBeingCaptured(es)) {
				player.addScore(score);
				/*Particle p = particles.getInstanceForced();
        p.set(Particle.Shape.BONUS, pos.x, pos.y, dd, 0.2f,
              1, 1, 1, 1, 60, false, player.multiplier);*/
				player.addMultiplier(0.1f);
			}
			Sound.playSe(explosionSeName);
		}
	}

	protected boolean gotoNextPhaseInAppearing(EnemyState es) {
//		with (es) {
		{
			switch (es.phase) {
			case -300:
				es.centerPos.x = rand.nextSignedFloat(field.size().x * 0.2f);
				es.centerPos.y = field.size().y * 2.0f;
				es.standByPos.x = rand.nextSignedFloat(field.size().x * 0.2f);
				es.standByPos.y = field.size().y * (0.7f + rand.nextFloat(0.1f));
				es.nextPhaseCnt = 15;
				es.baseSpeed = es.baseBaseSpeed * 1.5f;
				es.angVel = es.baseAngVel * 1.5f;
				es.phase = -50;
				break;

			case -200:
				es.centerPos.x = rand.nextSignedFloat(field.size().x * 0.1f);
				es.centerPos.y = field.size().y * 1.6f;
				if (es.centerPos.x < 0)
					es.standByPos.x = field.size().x * (rand.nextSignedFloat(0.4f) + 0.4f);
				else
					es.standByPos.x = field.size().x * (rand.nextSignedFloat(0.4f) - 0.4f);
				es.standByPos.y = field.size().y * (0.5f + rand.nextFloat(0.3f));
				es.nextPhaseCnt = 60;
				es.baseSpeed = es.baseBaseSpeed * 1.0f;
				es.angVel = es.baseAngVel * 1.0f;
				break;
			case -199:
				if (es.standByPos.x < 0)
					es.centerPos.x = field.size().x * 0.75f;
				else
					es.centerPos.x = -field.size().x * 0.75f;
				es.centerPos.y = 0;
				if (es.isGoingDownBeforeStandBy)
					es.nextPhaseCnt = 20;
				else
					es.nextPhaseCnt = 60;
				es.baseSpeed = es.baseBaseSpeed * 2;
				es.angVel = es.baseAngVel * 2;
				es.phase = -50;
				break;

			case -100:
				es.centerPos.x = field.size().x * 4.0f;
				if (rand.nextInt(2) == 0)
					es.centerPos.x *= -1;
				es.centerPos.y = field.size().y * 1.6f;
				if (es.centerPos.x < 0)
					es.standByPos.x = field.size().x * (rand.nextSignedFloat(0.4f) + 0.4f);
				else
					es.standByPos.x = field.size().x * (rand.nextSignedFloat(0.4f) - 0.4f);
				es.standByPos.y = field.size().y * (0.5f + rand.nextFloat(0.3f));
				es.nextPhaseCnt = 20;
				es.baseSpeed = es.baseBaseSpeed * 2.0f;
				es.angVel = es.baseAngVel * 2.0f;
				break;
			case -99:
				if (es.centerPos.x > 0)
					es.centerPos.x = field.size().x * 2.0f;
				else
					es.centerPos.x = -field.size().x * 2.0f;
				es.centerPos.y = -field.size().y * 1.2f;
				es.nextPhaseCnt = 20;
				es.baseSpeed = es.baseBaseSpeed * 2;
				es.angVel = es.baseAngVel * 2;
				break;
			case -98:
				if (es.centerPos.x > 0)
					es.centerPos.x = field.size().x * 0.5f;
				else
					es.centerPos.x = -field.size().x * 0.5f;
				es.centerPos.y = 0;
				es.nextPhaseCnt = 30;
				es.phase = -50;
				break;

			case -49:
				if (es.isGoingDownBeforeStandBy) {
					es.centerPos.x = es.centerPos.x / 2;
					es.centerPos.y = 0;
					es.phase = -30;
					es.nextPhaseCnt = 10;
					break;
				}
				es.centerPos.x = es.standByPos.x;
				es.centerPos.y = es.standByPos.y;
				es.nextPhaseCnt = calcStandByTime(es);
				es.baseSpeed = es.baseBaseSpeed;
				es.angVel = es.baseAngVel;
				es.phase = -10;
				break;

			case -29:
				es.centerPos.x = (es.centerPos.x + player.pos().x * 2) / 3;
				es.centerPos.y = -field.size().y * 1.1f;
				es.baseSpeed = es.baseBaseSpeed * 1.2f;
				es.angVel = es.baseAngVel * 1.2f;
				es.nextPhaseCnt = 10;
				break;
			case -28:
				es.centerPos.y = -field.size().y * 1.5f;
				es.nextPhaseCnt = 10;
				break;
			case -9:
				es.phase = 0;
				break;
			default:
				return false;
			}
			es.nextPhaseCnt /= rank;
			es.phaseCnt = 0;
		}
		return true;
	}

	public void movePhase(EnemyState es) {
//		with (es) {
		{
			switch (es.phase) {
			case -200:
			case -100:
				if (es.pos.y < field.size().y * 1.5f)
					es.pos.y = field.size().y * 1.5f;
				break;
			case -99:
				if (es.centerPos.x < 0 && es.pos.x > -field.size().x)
					es.pos.x += (-field.size().x - es.pos.x) * 0.2f;
				else if (es.centerPos.x > 0 && es.pos.x < field.size().x)
					es.pos.x += (field.size().x - es.pos.x) * 0.2f;
				break;
			case -50:
			case -49:
			case -10:
				if (es.pos.y < -field.size().y * 0.5f)
					es.pos.y += (-field.size().y * 0.5f - es.pos.y) * 0.2f;
				break;
			default:
				break;
			}
			if (isInAttack(es))
				if (es.phase >= 0 || rand.nextInt(5) == 0)
					turretSpec.move(es.turretState, rank);
		}
	}

	public boolean isInScreen(EnemyState es) {
		return (field.size().contains(es.pos));
	}

	public abstract void setRank(float rank);
	public abstract void init(EnemyState es);
	public abstract boolean gotoNextPhase(EnemyState es);
	public abstract boolean isInAttack(EnemyState es);
	protected abstract int calcStandByTime(EnemyState es);

	public void draw(GL gl, EnemyState es) {
		Vector3 p = field.calcCircularPos(es.pos);
		float cd = field.calcCircularDeg(es.pos.x);
		((EnemyShape) shape).draw(gl, p, cd, es.deg, es.cnt, es.size);
	}

	public void drawTrails(GL gl, EnemyState es) {
		if (es.captureState > 0)
			return;
		es.drawTrails(gl, trailShape, 0.2f, 0.2f, 0.8f, es.size, field);
	}
}

class Trail {
	//private:
	Vector2 pos;
	float deg;
	int cnt;

//	invariant {
//	assert(pos.x <>= 0);
//	assert(pos.y <>= 0);
//	assert(deg <>= 0);
//	}

	public Trail() {
		pos = new Vector2();
		deg = 0;
	}

	public void set(float x, float y, float d, int c) {
		pos.x = x;
		pos.y = y;
		deg = d;
		cnt = c;
	}
}

class MiddleEnemySpec extends EnemySpec {
	//private:

	public MiddleEnemySpec(Field field, BulletPool bullets, Player player,
			ParticlePool particles, EnemyPool enemies, Stage stage,
			Shape shape, EnemyShape trailShape, BulletSpec bulletSpec) {
		super(field, bullets, player, particles, enemies, stage,
				shape, trailShape, bulletSpec);
		turretSpec = new TurretSpec(field, bullets, player, enemies, particles,
				stage, bulletSpec);
		shield = 3;
		score = 100;
		capturable = false;
		aimAllDirection = false;
		explosionSeName = "explosion3.wav";
	}

	public void init(EnemyState es) {
//		with (es) {
		{
			es.size.x = es.size.y = 1.33f;
			es.phase = -300;
			gotoNextPhaseInAppearing(es);
		}
	}

	public void setRank(float r) {
		rank = (float)Math.sqrt(r);
		if (rank < 1.5f)
			rank = 1.5f;
		turretSpec.setRankMiddle(r * 3);
	}

	public boolean gotoNextPhase(EnemyState es) {
//		with (es) {
		{
			if (es.phase < 0)
				return gotoNextPhaseInAppearing(es);
			switch (es.phase) {
			case 1:
				if (!player.hasCollision()) {
					es.phase = 0;
					es.nextPhaseCnt = calcStandByTime(es);
					break;
				}
				Sound.playSe("flying_down.wav");
				es.centerPos.x = field.size().x * (0.6f + rand.nextSignedFloat(0.1f));
				if (rand.nextInt(2) == 0)
					es.centerPos.x *= -1;
				es.centerPos.y = field.size().y * (0.2f + rand.nextFloat(0.2f));
				es.nextPhaseCnt = 60;
				break;
			case 2:
				es.centerPos.x *= -0.9f;
				es.centerPos.y = field.size().y * (0.2f + rand.nextFloat(0.2f));
				es.nextPhaseCnt = 60;
				break;
			case 3:
				es.centerPos.x = es.standByPos.x;
				es.centerPos.y = es.standByPos.y;
				es.phase = 0;
				es.nextPhaseCnt = calcStandByTime(es);
				break;
			}
			es.nextPhaseCnt /= rank;
			es.phaseCnt = 0;
		}
		return true;
	}

	public boolean isInAttack(EnemyState es) {
		return (es.phase == 1 || es.phase == 2);
	}

	protected int calcStandByTime(EnemyState es) {
		if (es.phase < 0)
			return 30 + rand.nextInt(30);
		else
			return 200 + rand.nextInt(150);
	}
}

abstract class SmallEnemySpec extends EnemySpec {
	//private:

	public SmallEnemySpec(Field field, BulletPool bullets, Player player,
			ParticlePool particles, EnemyPool enemies, Stage stage,
			Shape shape, EnemyShape trailShape, BulletSpec bulletSpec) {
		super(field, bullets, player, particles, enemies, stage,
				shape, trailShape, bulletSpec);
		turretSpec = new TurretSpec(field, bullets, player, enemies, particles,
				stage, bulletSpec);
		shield = 1;
		score = 10;
		capturable = true;
		aimAllDirection = false;
	}

	public void init(EnemyState es) {
		gotoNextPhaseInAppearing(es);
	}

	public void init(EnemyState es, EnemyState fes) {
//		with (es) {
		{
			es.centerPos.x = fes.centerPos.x;
			es.centerPos.y = fes.centerPos.y;
			es.standByPos.x = fes.standByPos.x;
			es.standByPos.y = fes.standByPos.y;
			es.nextPhaseCnt = fes.nextPhaseCnt;
			es.baseSpeed = fes.baseSpeed;
			es.angVel = fes.angVel;
			es.phase = fes.phase;
			es.size.x = es.size.y = 1.25f;
		}
	}

	protected int calcStandByTime(EnemyState es) {
		return 60 + rand.nextInt(120);
	}
}

class SE1Spec extends SmallEnemySpec {
//	private:

	public SE1Spec(Field field, BulletPool bullets, Player player,
			ParticlePool particles, EnemyPool enemies, Stage stage,
			Shape shape, EnemyShape trailShape, BulletSpec bulletSpec) {
		super(field, bullets, player, particles, enemies, stage,
				shape, trailShape, bulletSpec);
		explosionSeName = "explosion1.wav";
	}

	public void setRank(float r) {
		rank = (float)Math.sqrt(r * 0.5f);
		if (rank < 1)
			rank = 1;
		turretSpec.setRankNormal(r * 2);
	}

	public boolean gotoNextPhase(EnemyState es) {
//		with (es) {
		{
			if (es.phase < 0)
				return gotoNextPhaseInAppearing(es);
			switch (es.phase) {
			case 1:
				if (!player.hasCollision() || enemies.numInAttack() > stage.attackSmallEnemyNum()) {
					es.phase = 0;
					es.nextPhaseCnt = calcStandByTime(es);
					break;
				}
				Sound.playSe("flying_down.wav");
				es.centerPos.y = 0;
				es.centerPos.x = (es.standByPos.x + player.pos().x) / 2;
				es.nextPhaseCnt = 60;
				break;
			case 2:
				es.centerPos.y = -field.size().y * 0.7f;
				es.centerPos.x = player.pos().x;
				es.nextPhaseCnt = 30;
				break;
			case 3:
				es.centerPos.x = es.standByPos.x;
				es.centerPos.y = es.standByPos.y;
				es.phase = 0;
				es.nextPhaseCnt = calcStandByTime(es);
				break;
			}
			es.nextPhaseCnt /= rank;
			es.phaseCnt = 0;
		}
		return true;
	}

	public boolean isInAttack(EnemyState es) {
		return (es.phase < -10 || es.phase == 1 || es.phase == 2);
	}
}

class SE2Spec extends SmallEnemySpec {
//	private:

	public SE2Spec(Field field, BulletPool bullets, Player player,
			ParticlePool particles, EnemyPool enemies, Stage stage,
			Shape shape, EnemyShape trailShape, BulletSpec bulletSpec) {
		super(field, bullets, player, particles, enemies, stage,
				shape, trailShape, bulletSpec);
		explosionSeName = "explosion2.wav";
	}

	public void setRank(float r) {
		rank = (float)Math.sqrt(r * 0.5f);
		if (rank < 1)
			rank = 1;
		turretSpec.setRankNormal(r * 2, true);
	}

	public boolean gotoNextPhase(EnemyState es) {
//		with (es) {
		{
			if (es.phase < 0)
				return gotoNextPhaseInAppearing(es);
			switch (es.phase) {
			case 1:
				if (!player.hasCollision() || enemies.numInAttack() > stage.attackSmallEnemyNum()) {
					es.phase = 0;
					es.nextPhaseCnt = calcStandByTime(es);
					break;
				}
				Sound.playSe("flying_down.wav");
				es.centerPos.y = -field.size().y * 0.3f;
				es.centerPos.x = (es.standByPos.x + player.pos().x) / 2;
				es.baseSpeed = es.baseBaseSpeed;
				es.angVel = es.baseAngVel;
				es.nextPhaseCnt = 30 + rand.nextInt(60);
				break;
			case 2:
				es.centerPos.y = -field.size().y * 1.3f;
				es.centerPos.x *= -1;
				es.nextPhaseCnt = 30;
				break;
			case 3:
				es.centerPos.y = -field.size().y * 1.0f;
				if (es.centerPos.x < 0)
					es.centerPos.x = -field.size().x * 1.5f;
				else
					es.centerPos.x = field.size().x * 1.5f;
				es.baseSpeed = es.baseBaseSpeed * 1.5f;
				es.angVel = es.baseAngVel * 1.5f;
				es.nextPhaseCnt = 30;
				break;
			case 4:
				es.centerPos.x = es.standByPos.x;
				es.centerPos.y = es.standByPos.y;
				es.phase = 0;
				es.nextPhaseCnt = calcStandByTime(es);
				break;
			}
			es.nextPhaseCnt /= rank;
			es.phaseCnt = 0;
		}
		return true;
	}

	public void movePhase(EnemyState es) {
		super.movePhase(es);
//		with (es) {
		{
			if (es.phase == 3) {
				if (es.centerPos.x < 0) {
					if (es.pos.x > -field.size().x * 1.2f)
						es.pos.x += (es.centerPos.x - es.pos.x) * 0.2f;
				} else {
					if (es.pos.x < field.size().x * 1.2f)
						es.pos.x += (es.centerPos.x - es.pos.x) * 0.2f;
				}
			}
		}
	}

	public boolean isInAttack(EnemyState es) {
		return (es.phase < -10 || es.phase == 1 || es.phase == 2 || es.phase == 3);
	}
}

class TurretState extends TokenState {
	//private:
	float fireCnt, burstCnt;
	int burstNum;
	int nwaySpeedAccelDir;

//	invariant {
//	if (isInitialized) {
//	assert(fireCnt <>= 0);
//	assert(burstCnt <>= 0);
//	}
//	}

	public void clear() {
		fireCnt = burstCnt = 0;
		burstNum = 0;
		nwaySpeedAccelDir = 1;
		super.clear();
	}

	public void update(float x, float y, float d) {
		pos.x = x;
		pos.y = y;
		deg = d;
	}
}

class TurretSpec<T extends TurretState> extends TokenSpec<T> {
//	mixin StaticRandImpl;
	public static Rand rand = new Rand();
	//private:
	static final float INTERVAL_MAX = 90.0f;
	static final float SPEED_RATIO = 5.0f;
	BulletSpec bulletSpec;
	BulletPool bullets;
	Player player;
	Stage stage;
	int interval;
	float speed;
	float speedAccel;
	int burstNum, burstInterval;
	int nway;
	float nwayAngle;
	float nwayBaseDeg;
	float nwaySpeedAccel;
	boolean fireingAtATime;

//	invariant {
//	assert(interval > 0);
//	assert(speed > 0);
//	assert(speedAccel < 1 && speedAccel > -1);
//	assert(burstNum >= 1);
//	assert(burstInterval >= 0);
//	assert(nway >= 1);
//	assert(nwayAngle >= 0);
//	assert(nwayBaseDeg <>= 0);
//	assert(nwaySpeedAccel <>= 0);
//	}

	public TurretSpec(Field field, BulletPool bullets, Player player,
			EnemyPool enemies, ParticlePool particles,
			Stage stage, BulletSpec bulletSpec) {
		this.bulletSpec = bulletSpec;
		this.field = field;
		this.bullets = bullets;
		this.player = player;
		this.stage = stage;
		initParam();
	}

	private void initParam() {
		interval = 99999;
		speed = 1;
		speedAccel = 0;
		burstNum = 1;
		burstInterval = 99999;
		nway = 1;
		nwayAngle = 0;
		nwayBaseDeg = 0;
		nwaySpeedAccel = 0;
		fireingAtATime = false;
	}

	public void setRankNormal(float rank) {
		setRankNormal(rank, false);
	}

	public void setRankNormal(float rank, boolean isWide) {
		initParam();
		float rr = rand.nextFloat(0.5f);
		float nsr = 0.5f + rand.nextSignedFloat(0.3f);
		float nr, br, ir;
		if (isWide) {
			nr = (rank * nsr * rr);
			br = 0;
			ir = (rank * nsr * (1 - rr));
		} else {
			nr = 0;
			br = (rank * nsr * rr);
			ir = (rank * nsr * (1 - rr));
		}
		burstNum = (int) Math.sqrt(br) + 1;
		burstInterval = 3 + rand.nextInt(2);
		nway = (int) Math.sqrt(nr) + 1;
		interval = (int) (INTERVAL_MAX / (ir * 2 + 1)) + 1;
		float sr = rank - nway + 1 - burstNum + 1 - ir;
		if (sr < 0.01f)
			sr = 0.01f;
		speed = (float)Math.sqrt(sr * 0.66f);
		assert(speed > 0);
		speed *= 0.2f;
		if (speed < 0.1f)
			speed = 0.1f;
		else
			speed = (float)Math.sqrt(speed * 10) / 10;
		assert(speed > 0);
		speed *= 0.33f;
		nwayAngle = (1.66f + rand.nextFloat(0.33f)) / (1 + nway * 0.7f) * 0.06f;
		fireingAtATime = true;
	}

	public void setRankMiddle(float rank) {
		initParam();
		float nr, br, ir;
		if (rand.nextInt(3) == 0) {
			nr = 0;
			br = (rank * 0.4f) * (1.0f + rand.nextSignedFloat(0.2f));
			ir = (rank * 0.5f) * (1.0f + rand.nextSignedFloat(0.2f));
		} else {
			nr = (rank * 0.3f) * (1.0f + rand.nextSignedFloat(0.2f));
			br = (rank * 0.3f) * (1.0f + rand.nextSignedFloat(0.2f));
			ir = (rank * 0.3f) * (1.0f + rand.nextSignedFloat(0.2f));
		}
		boolean acf = false;
		burstNum = (int) Math.sqrt(br) + 1;
		if (burstNum > 1 && rand.nextInt(3) > 0) {
			acf = true;
			nr *= 0.7f;
			ir *= 0.7f;
			rank *= 0.8f;
		}
		burstInterval = 3 + rand.nextInt(2);
		nway = (int) Math.sqrt(nr) + 1;
		interval = (int) (INTERVAL_MAX / (Math.sqrt(ir * 2 + 1))) + 1;
		float sr = rank - burstNum + 1 - nway + 1 - ir;
		if (sr < 0.01f)
			sr = 0.01f;
		speed = (float)Math.sqrt(sr * 0.66f);
		assert(speed > 0);
		speed *= 0.2f;
		if (speed < 0.1f)
			speed = 0.1f;
		else
			speed = (float)Math.sqrt(speed * 10) / 10;
		assert(speed > 0);
		speed *= 0.33f;
		if (acf) {
			speedAccel = (speed * (0.2f + burstNum * 0.05f + rand.nextFloat(0.1f))) / (burstNum - 1);
			if (rand.nextInt(2) == 0)
				speedAccel *= -1;
		}
		if (nway > 1 && rand.nextInt(3) == 0) {
			nwaySpeedAccel = (speed * (0.05f + nway * 0.02f + rand.nextFloat(0.02f))) / (nway - 1);
			if (rand.nextInt(2) == 0)
				nwaySpeedAccel *= -1;
		}
		nwayAngle = (1.66f + rand.nextFloat(0.33f)) / (1 + nway * 0.7f) * 0.06f;
		if (rand.nextInt(3) == 0)
			fireingAtATime = true;
	}

	public boolean move(TurretState ts) {
		return move(ts, 1);
	}

	public boolean move(TurretState ts, float time) {
		if (fireingAtATime) {
			ts.fireCnt -= time;
			if (ts.fireCnt <= 0) {
				ts.fireCnt = interval;
				if (ts.fireCnt < 3)
					ts.fireCnt = 3;
				if (ts.pos.y > 0) {
					float sp = speed - speedAccel * (burstNum - 1) / 2;
					for (int i = 0; i < burstNum; i++) {
						float d = ts.deg - nwayAngle * (nway - 1) / 2 + nwayBaseDeg;
						float nsp = sp - nwaySpeedAccel * ts.nwaySpeedAccelDir * (nway - 1) / 2;
						for (int j = 0; j < nway; j++) {
							Bullet b = bullets.getInstance();
							if (b == null)
								break;
							b.set(bulletSpec, ts.pos, d, nsp * SPEED_RATIO);
							b.setWaitCnt(i * burstInterval);
							d += nwayAngle;
							nsp += nwaySpeedAccel * ts.nwaySpeedAccelDir;
						}
						sp += speedAccel;
					}
					ts.nwaySpeedAccelDir *= -1;
				}
			}
		} else {
			if (ts.burstNum <= 0) {
				ts.fireCnt -= time;
				if (ts.fireCnt <= 0) {
					ts.fireCnt = interval;
					if (ts.fireCnt < 3)
						ts.fireCnt = 3;
					ts.burstNum = burstNum;
					ts.burstCnt = 0;
					ts.speed = speed - speedAccel * (ts.burstNum - 1) / 2;
					ts.nwaySpeedAccelDir *= -1;
				}
			}
			if (ts.burstNum > 0) {
				ts.burstCnt -= time;
				if (ts.burstCnt <= 0) {
					ts.burstCnt = burstInterval;
					ts.burstNum--;
					if (ts.pos.y > 0) {
						float d = ts.deg - nwayAngle * (nway - 1) / 2 + nwayBaseDeg;
						float nsp = ts.speed - nwaySpeedAccel * ts.nwaySpeedAccelDir * (nway - 1) / 2;
						for (int i = 0; i < nway; i++) {
							Bullet b = bullets.getInstance();
							if (b == null)
								break;
							b.set(bulletSpec, ts.pos, d, nsp * SPEED_RATIO);
							d += nwayAngle;
							nsp += nwaySpeedAccel * ts.nwaySpeedAccelDir;
						}
					}
					ts.speed += speedAccel;
				}
			}
		}
		return true;
	}
}
