/*
 * $Id: field.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import java.nio.FloatBuffer;

import javax.media.opengl.GL;
import javax.media.opengl.glu.GLU;

import com.sun.opengl.util.BufferUtil;

import abagames.util.MathUtil;
import abagames.util.Vector2;
import abagames.util.Vector3;

//private import std.math;
//private import opengl;
//private import openglu;
//private import abagames.util.vector;
//private import abagames.util.math;
//private import abagames.ttn.screen;
//private import abagames.ttn.frame;

/**
 * Game field.
 */
public class Field {
	public static final float PIT_SIZE_Y_RATIO = 12.0f;
	static final float CIRCLE_RADIUS = 64.0f;
	static final float EYE_POS_DIST_RATIO = 1.25f;
	static final float X_EXPANSION_RATIO = 1.0f;
	static final float SIDEWALL_WIDTH = 145;
	static final float TORUS_Y = -24.0f;
	Frame frame;
	Screen screen;
	Vector2 _size, _outerSize;
	Vector3 _eyePos;
	float eyeDeg;
	Vector3 circlePos;
	int cnt;
	GLU glu = new GLU();

	public Field(Frame frame, Screen screen) {
		this.frame = frame;
		this.screen = screen;
		_size = new Vector2(12, 12);
		_outerSize = new Vector2(13, 13);
		_eyePos = new Vector3();
		circlePos = new Vector3();
		set();
		initTorus();
	}

	public void set() {
		_eyePos.x = _eyePos.y = _eyePos.z = 0;
		eyeDeg = 0;
		cnt = 0;
	}

	public boolean contains(Vector2 p) {
		return contains(p.x, p.y);
	}

	public boolean contains(float x, float y) {
		return _size.contains(x, y);
	}

	public boolean containsOuter(Vector2 p) {
		return containsOuter(p.x, p.y);
	}

	public boolean containsOuter(float x, float y) {
		return _outerSize.contains(x, y);
	}

	public boolean containsOuterY(float y) {
		return (y >= -_outerSize.y && y <= _outerSize.y);
	}

	public boolean containsIncludingPit(Vector2 p) {
		return (p.y >= -_outerSize.y && p.y <= _size.y * PIT_SIZE_Y_RATIO * 1.1f);
	}

	public float normalizeX(float x) {
		float rx = x;
		float hd = CIRCLE_RADIUS * (float)Math.PI / X_EXPANSION_RATIO;
		if (rx < -hd)
			rx = hd * 2 - (-rx % (hd * 2));
		return (rx + hd) % (hd * 2) - hd;
	}

	public float calcCircularDist(Vector2 p1, Vector2 p2) {
		float ax = Math.abs(normalizeX(p1.x - p2.x));
		float ay = Math.abs(p1.y - p2.y);
		if (ax > ay)
			return ax + ay / 2;
		else
			return ay + ax / 2;
	}

	public float circularDistance() {
		return CIRCLE_RADIUS * (float)Math.PI * 2 / X_EXPANSION_RATIO;
	}

	public Vector3 calcCircularPos(Vector2 p) {
		return calcCircularPos(p.x, p.y);
	}

	public Vector3 calcCircularPos(float x, float y) {
		float d = calcCircularDeg(x);
		if (y < _size.y) {
			circlePos.x = (float)Math.sin(d) * CIRCLE_RADIUS;
			circlePos.z = (float)Math.cos(d) * CIRCLE_RADIUS;
			circlePos.y = y;
		} else if (y < _size.y * 3) {
			float cd = (y - _size.y) * (float)Math.PI / 2 / (_size.y * 2);
			float cr = CIRCLE_RADIUS * (0.8f + 0.2f * (float)Math.cos(cd));
			circlePos.x = (float)Math.sin(d) * cr;
			circlePos.z = (float)Math.cos(d) * cr;
			circlePos.y = _size.y + (float)Math.sin(cd) * CIRCLE_RADIUS * 0.2f;
		} else if (y < _size.y * 7) {
			float cd = (y - _size.y * 3) * (float)Math.PI / 2 / (_size.y * 4);
			float cr = CIRCLE_RADIUS * (0.8f - 0.4f * (float)Math.sin(cd));
			circlePos.x = (float)Math.sin(d) * cr;
			circlePos.z = (float)Math.cos(d) * cr;
			circlePos.y = _size.y - CIRCLE_RADIUS * 0.2f + (float)Math.cos(cd) * CIRCLE_RADIUS * 0.4f;
		} else {
			float cr = CIRCLE_RADIUS * 0.4f;
			circlePos.x = (float)Math.sin(d) * cr;
			circlePos.z = (float)Math.cos(d) * cr;
			circlePos.y = _size.y - CIRCLE_RADIUS * 0.2f - (y - _size.y * 7);
		}
		return circlePos;
	}

	public float calcCircularDeg(float x) {
		return x * X_EXPANSION_RATIO / CIRCLE_RADIUS;
	}

	public float calcCircularDist(float d) {
		return d * CIRCLE_RADIUS / X_EXPANSION_RATIO ;
	}

	public boolean checkHitDist(Vector2 pos, Vector2 p, Vector2 pp, float dist) {
		float bmvx, bmvy, inaa;
		bmvx = pp.x;
		bmvy = pp.y;
		bmvx -= p.x;
		bmvy -= p.y;
		bmvx = normalizeX(bmvx);
		inaa = bmvx * bmvx + bmvy * bmvy;
		if (inaa > 0.00001) {
			float sofsx, sofsy, inab, hd;
			sofsx = pos.x;
			sofsy = pos.y;
			sofsx -= p.x;
			sofsy -= p.y;
			sofsx = normalizeX(sofsx);
			inab = bmvx * sofsx + bmvy * sofsy;
			if (inab >= 0 && inab <= inaa) {
				hd = sofsx * sofsx + sofsy * sofsy - inab * inab / inaa;
				if (hd >= 0 && hd <= dist)
					return true;
			}
		}
		return false;
	}

	public void addSlowdownRatio(float sr) {
		frame.addSlowdownRatio(sr);
	}

	public void setEyePos(Vector2 p) {
		eyeDeg = calcCircularDeg(p.x) * 0.25f;
		_eyePos.x = (float)Math.sin(eyeDeg) * CIRCLE_RADIUS * EYE_POS_DIST_RATIO;
		_eyePos.z = (float)Math.cos(eyeDeg) * CIRCLE_RADIUS * EYE_POS_DIST_RATIO;
	}

	public void setLookAt(GL gl) {
		gl.glMatrixMode(GL.GL_PROJECTION);
		screen.setPerspective();
		glu.gluLookAt(_eyePos.x, _eyePos.y, _eyePos.z, 0, _eyePos.y, 0, 0, 1, 0);
		gl.glMatrixMode(GL.GL_MODELVIEW);
	}

	public void resetLookAt(GL gl) {
		gl.glMatrixMode(GL.GL_PROJECTION);
		screen.setPerspective();
		glu.gluLookAt(0, 0, 1, 0, 0, 0, 0, 1, 0);
		gl.glMatrixMode(GL.GL_MODELVIEW);
	}

	public void beginDrawingFront(GL gl) {
		gl.glMatrixMode(GL.GL_PROJECTION);
		gl.glLoadIdentity();
		gl.glOrtho(0, 640, 480, 0, -1, 1);
		gl.glMatrixMode(GL.GL_MODELVIEW);
		drawSidewall(gl);
	}

	private void drawSidewall(GL gl) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		Screen.setColor(gl, 0.25f, 0.25f, 0.25f, 0.5f);
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3f(0, 0, 0);
		gl.glVertex3f(SIDEWALL_WIDTH, 0, 0);
		gl.glVertex3f(SIDEWALL_WIDTH, 480, 0);
		gl.glVertex3f(0, 480, 0);
		gl.glEnd();
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3f(640, 0, 0);
		gl.glVertex3f(640 - SIDEWALL_WIDTH, 0, 0);
		gl.glVertex3f(640 - SIDEWALL_WIDTH, 480, 0);
		gl.glVertex3f(640, 480, 0);
		gl.glEnd();
		Screen.setColor(gl, 1.0f, 1.0f, 1.0f, 0.8f);
		gl.glBegin(GL.GL_LINES);
		gl.glVertex3f(SIDEWALL_WIDTH, 0, 0);
		gl.glVertex3f(SIDEWALL_WIDTH, 480, 0);
		gl.glVertex3f(640 - SIDEWALL_WIDTH, 0, 0);
		gl.glVertex3f(640 - SIDEWALL_WIDTH, 480, 0);
		gl.glEnd();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
	}

	public void move() {
		cnt++;
	}

	public void drawBack(GL gl) {
		gl.glPushMatrix();
		gl.glTranslatef(0, TORUS_Y, 0);
		drawTorusShape(gl, (float)Math.PI / 2);
		gl.glPopMatrix();
	}

	public void drawFront(GL gl) {
		gl.glPushMatrix();
		gl.glTranslatef(0, TORUS_Y, 0);
		drawTorusShape(gl, -(float)Math.PI / 2);
		gl.glPopMatrix();
	}

	Vector3 _cp = new Vector3();
	Vector3 _ringOfs = new Vector3();
	
	float[] torusVertexArray;
	float[] torusColorArray;
	FloatBuffer torusVertex;
	FloatBuffer torusColor;
	float[] torusLineVertexArray;
	FloatBuffer torusLineVertex;
	
	private void initTorus() {
		torusVertexArray = new float[16 * 16 * 4 * 3];
		torusColorArray = new float[16 * 16 * 4 * 4];
		torusVertex = BufferUtil.newFloatBuffer(torusVertexArray.length);
		torusColor = BufferUtil.newFloatBuffer(torusColorArray.length);
		torusLineVertexArray = new float[16 * 16 * 4 * 3];
		torusLineVertex = BufferUtil.newFloatBuffer(torusLineVertexArray.length);
	}
	
	private void setTorusShape(float d1s) {
//		long t0 = System.nanoTime();
		Vector3 cp = _cp; //new Vector3();
		cp.clear();
		cp.y = 0;
		Vector3 ringOfs = _ringOfs;//new Vector3();
		ringOfs.clear();
		float torusRad = CIRCLE_RADIUS * 0.9f;
		float ringRad;
		float d1;
		ringRad = CIRCLE_RADIUS * 0.3f;
		d1 = d1s;
		int ptrc = 0;
		int ptrv = 0;
		final float dd1 = (float)Math.PI * 2 / 32;
		final float dd2 = (float)Math.PI * 2 / 16;
		for (int i = 0; i < 16; i++, d1 += dd1) {
			float d2 = cnt * 0.003f;
			for (int j = 0; j < 16; j++, d2 += dd2) {
				cp.x = (float)sin(d1) * torusRad;
				cp.z = (float)cos(d1) * torusRad;
				createRingOffset(ringOfs, cp, ringRad, d1, d2);
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.8f;
				torusVertexArray[ptrv++] = ringOfs.x;
				torusVertexArray[ptrv++] = ringOfs.y;
				torusVertexArray[ptrv++] = ringOfs.z;

				createRingOffset(ringOfs, cp, ringRad, d1, d2 + dd2);
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.8f;
				torusVertexArray[ptrv++] = ringOfs.x;
				torusVertexArray[ptrv++] = ringOfs.y;
				torusVertexArray[ptrv++] = ringOfs.z;
				
				cp.x = (float)sin(d1 + dd1) * torusRad;
				cp.z = (float)cos(d1 + dd1) * torusRad;
				createRingOffset(ringOfs, cp, ringRad, d1 + dd1, d2 + dd2);
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.8f;
				torusVertexArray[ptrv++] = ringOfs.x;
				torusVertexArray[ptrv++] = ringOfs.y;
				torusVertexArray[ptrv++] = ringOfs.z;
				
				createRingOffset(ringOfs, cp, ringRad, d1 + dd1, d2);
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.3f;
				torusColorArray[ptrc++] = 0.2f;
				torusVertexArray[ptrv++] = ringOfs.x;
				torusVertexArray[ptrv++] = ringOfs.y;
				torusVertexArray[ptrv++] = ringOfs.z;
			}
		}
//		long t1 = System.nanoTime();
		ringRad = CIRCLE_RADIUS * 0.3f;
		d1 = d1s;
		ptrv = 0;
		for (int i = 0; i < 16; i++, d1 += dd1) {
			float d2 = cnt * 0.003f;
			for (int j = 0; j < 16; j++, d2 += dd2) {
				cp.x = (float)sin(d1 + dd1 * 0.1f) * torusRad;
				cp.z = (float)cos(d1 + dd1 * 0.1f) * torusRad;
				createRingOffset(ringOfs, cp, ringRad, d1 + dd1 * 0.1f, d2 + dd2 * 0.1f);
				torusLineVertexArray[ptrv++] = ringOfs.x;
				torusLineVertexArray[ptrv++] = ringOfs.y;
				torusLineVertexArray[ptrv++] = ringOfs.z;
				
				createRingOffset(ringOfs, cp, ringRad, d1 + dd1 * 0.1f, d2 + dd2 * 0.9f);
				torusLineVertexArray[ptrv++] = ringOfs.x;
				torusLineVertexArray[ptrv++] = ringOfs.y;
				torusLineVertexArray[ptrv++] = ringOfs.z;
				
				cp.x = (float)sin(d1 + dd1 * 0.9f) * torusRad;
				cp.z = (float)cos(d1 + dd1 * 0.9f) * torusRad;
				createRingOffset(ringOfs, cp, ringRad, d1 + dd1 * 0.9f, d2 + dd1 * 0.1f);
				torusLineVertexArray[ptrv++] = ringOfs.x;
				torusLineVertexArray[ptrv++] = ringOfs.y;
				torusLineVertexArray[ptrv++] = ringOfs.z;
				
				createRingOffset(ringOfs, cp, ringRad, d1 + dd1 * 0.9f, d2 + dd2 * 0.9f);
				torusLineVertexArray[ptrv++] = ringOfs.x;
				torusLineVertexArray[ptrv++] = ringOfs.y;
				torusLineVertexArray[ptrv++] = ringOfs.z;
			}
		}
//		long t2 = System.nanoTime();
		
		torusVertex.put(torusVertexArray);
		torusVertex.rewind();
		torusColor.put(torusColorArray);
		torusColor.rewind();
		torusLineVertex.put(torusLineVertexArray);
		torusLineVertex.rewind();
//		long t3 = System.nanoTime();
//		System.out.println("setup " + (t1 - t0) / 1000000.0 + " " + (t2 - t1) / 1000000.0 + " " + (t3 - t2) / 1000000.0);
	}
	
	private void drawTorusShape(GL gl, float d1s) {
//		long t0 = System.nanoTime();
		setTorusShape(d1s);
//		long t1 = System.nanoTime();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
		gl.glEnableClientState(GL.GL_COLOR_ARRAY);
		gl.glVertexPointer(3, GL.GL_FLOAT, 0, torusVertex);
		gl.glColorPointer(4, GL.GL_FLOAT, 0, torusColor);
		gl.glDrawArrays(GL.GL_QUADS, 0, 16 *16 * 4);

//		long t2 = System.nanoTime();
		gl.glDisableClientState(GL.GL_COLOR_ARRAY);
		
		gl.glVertexPointer(3, GL.GL_FLOAT, 0, torusLineVertex);
		Screen.setColor(gl, 0.1f, 0.1f, 0.1f);
		gl.glDrawArrays(GL.GL_LINE_STRIP, 0, 16 *16 * 4);
		
		gl.glDisableClientState(GL.GL_VERTEX_ARRAY);
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);

//		long t3 = System.nanoTime();
//		System.out.println("drawTorusShape " + (t1 - t0) / 1000000.0 + " " + (t2 - t1) / 1000000.0 + " " + (t3 - t2) / 1000000.0);
//		Vector3 cp = _cp; //new Vector3();
//		cp.clear();
//		cp.y = 0;
//		Vector3 ringOfs = _ringOfs;//new Vector3();
//		ringOfs.clear();
//		float torusRad = CIRCLE_RADIUS * 0.9f;
//		float ringRad;
//		float d1;
//		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
//		gl.glBegin(GL.GL_QUADS);
//		ringRad = CIRCLE_RADIUS * 0.3f;
//		d1 = d1s;
//		for (int i = 0; i < 16; i++, d1 += (float)Math.PI * 2 / 32) {
//			float d2 = cnt * 0.003f;
//			for (int j = 0; j < 16; j++, d2 += (float)Math.PI * 2 / 16) {
//				cp.x = (float)Math.sin(d1) * torusRad;
//				cp.z = (float)Math.cos(d1) * torusRad;
//				createRingOffset(ringOfs, cp, ringRad, d1, d2);
//				Screen.setColor(gl, 0.3f, 0.3f, 0.3f, 0.8f);
//				Screen.glVertex(gl, ringOfs);
//				createRingOffset(ringOfs, cp, ringRad, d1, d2 + (float)Math.PI * 2 / 16);
//				Screen.glVertex(gl, ringOfs);
//				cp.x = (float)Math.sin(d1 + (float)Math.PI * 2 / 32) * torusRad;
//				cp.z = (float)Math.cos(d1 + (float)Math.PI * 2 / 32) * torusRad;
//				createRingOffset(ringOfs, cp, ringRad, d1 + (float)Math.PI * 2 / 32, d2 + (float)Math.PI * 2 / 16);
//				Screen.glVertex(gl, ringOfs);
//				createRingOffset(ringOfs, cp, ringRad, d1 + (float)Math.PI * 2 / 32, d2);
//				Screen.setColor(gl, 0.3f, 0.3f, 0.3f, 0.2f);
//				Screen.glVertex(gl, ringOfs);
//			}
//		}
//		gl.glEnd();
//		gl.glBegin(GL.GL_LINE_STRIP);
//		ringRad = CIRCLE_RADIUS * 0.3f;
//		Screen.setColor(gl, 0.1f, 0.1f, 0.1f);
//		d1 = d1s;
//		for (int i = 0; i < 16; i++, d1 += (float)Math.PI * 2 / 32) {
//			float d2 = cnt * 0.003f;
//			for (int j = 0; j < 16; j++, d2 += (float)Math.PI * 2 / 16) {
//				cp.x = (float)Math.sin(d1 + (float)Math.PI * 2 / 32 * 0.1f) * torusRad;
//				cp.z = (float)Math.cos(d1 + (float)Math.PI * 2 / 32 * 0.1f) * torusRad;
//				createRingOffset(ringOfs, cp, ringRad, d1 + (float)Math.PI * 2 / 32 * 0.1f, d2 + (float)Math.PI * 2 / 16 * 0.1f);
//				Screen.glVertex(gl, ringOfs);
//				createRingOffset(ringOfs, cp, ringRad, d1 + (float)Math.PI * 2 / 32 * 0.1f, d2 + (float)Math.PI * 2 / 16 * 0.9f);
//				Screen.glVertex(gl, ringOfs);
//				cp.x = (float)Math.sin(d1 + (float)Math.PI * 2 / 32 * 0.9f) * torusRad;
//				cp.z = (float)Math.cos(d1 + (float)Math.PI * 2 / 32 * 0.9f) * torusRad;
//				createRingOffset(ringOfs, cp, ringRad, d1 + (float)Math.PI * 2 / 32 * 0.9f, d2 + (float)Math.PI * 2 / 32 * 0.1f);
//				Screen.glVertex(gl, ringOfs);
//				createRingOffset(ringOfs, cp, ringRad, d1 + (float)Math.PI * 2 / 32 * 0.9f, d2 + (float)Math.PI * 2 / 16 * 0.9f);
//				Screen.glVertex(gl, ringOfs);
//			}
//		}
//		gl.glEnd();
//		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
	}

	public final void createRingOffset(Vector3 ringOfs, Vector3 centerPos,
			float rad, float d1, float d2) {
		ringOfs.x = 0;
		ringOfs.y = 0;
		ringOfs.z = rad;
		ringOfs.rollX2(d2);
		ringOfs.rollY2(-d1);
		ringOfs.opAddAssign(centerPos);
	}

	public Vector3 eyePos() {
		return _eyePos;
	}

	public Vector2 size() {
		return _size;
	}

	public static float sin(double t) {
//		if (true)
//			return (float)Math.sin(t);
		t = MathUtil.normalizeDeg(t);
		if (t >= 0) {
			return sinTable[(int)(t * scale)];
		} else {
			return -sinTable[(int)(- t * scale)];
		}
	}

	public static float cos(double t) {
		return sin(t + Math.PI / 2);
	}
	
	private static int TABLE_SIZE = 1024;
	private static float[] sinTable;
	private static final float scale;
	
	static {
		scale = (float)(1 / Math.PI * TABLE_SIZE);
		sinTable = new float[TABLE_SIZE + 1];
		for (int i = 0; i < TABLE_SIZE + 1; i++) {
			sinTable[i] = (float)Math.sin(Math.PI * i / TABLE_SIZE);
		}
	}
}
