/*
 * $Id: letter.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import javax.media.opengl.GL;

import abagames.util.sdl.DisplayList;

/**
 * Letters.
 */
public class Letter {
 
	public static DisplayList displayList;
	public static final float LETTER_WIDTH = 2.1f;
	public static final float LETTER_HEIGHT = 3.0f;
	public static final int TO_RIGHT = 0, TO_DOWN = 1, TO_LEFT = 2, TO_UP = 3;
 
	static final int LETTER_NUM = 44;
	static final int DISPLAY_LIST_NUM = LETTER_NUM * 3;
	static final int NORMAL = 0, POLYGON = 1, LINE = 2;

	public static void init(GL gl) {
		displayList = new DisplayList(gl, DISPLAY_LIST_NUM);
		displayList.resetList();
		for (int j = 0; j < 3; j++) {
			for (int i = 0; i < LETTER_NUM; i++) {
				displayList.newList(gl);
				setLetter(gl, i, j);
				displayList.endList(gl);
			}
		}
	}

	public static void close(GL gl) {
		displayList.close(gl);
	}

	public static float getWidth(int n, float s) {
		return n * s * LETTER_WIDTH;
	}

	public static float getWidthNum(int num, float s) {
		//int dg = 1;
		int n = num;
		int c = 1;
		for (;;) {
			if (n < 10)
				break;
			n /= 10;
			c++;
		}
		return c * s * LETTER_WIDTH;
	}

	public static float getHeight(float s) {
		return s * LETTER_HEIGHT;
	}

	public static void drawLetter(GL gl, int n) {
		displayList.call(gl, n);
	}

	private static void drawLetter(GL gl, int n, float x, float y, float s, float d) {
		gl.glPushMatrix();
		gl.glTranslatef(x, y, 0);
		gl.glScalef(s, s, s);
		gl.glRotatef(d, 0, 0, 1);
		displayList.call(gl, n);
		gl.glPopMatrix();
	}

	private static void drawLetterRev(GL gl, int n, float x, float y, float s, float d) {
		gl.glPushMatrix();
		gl.glTranslatef(x, y, 0);
		gl.glScalef(s, -s, s);
		gl.glRotatef(d, 0, 0, 1);
		displayList.call(gl, n);
		gl.glPopMatrix();
	}

	public static int convertCharToInt(char c) {
		int idx = 0;
		if (c >= '0' && c <='9') {
			idx = c - '0';
		} else if (c >= 'A' && c <= 'Z') {
			idx = c - 'A' + 10;
		} else if (c >= 'a' && c <= 'z') {
			idx = c - 'a' + 10;
		} else if (c == '.') {
			idx = 36;
		} else if (c == '-') {
			idx = 38;
		} else if (c == '+') {
			idx = 39;
		} else if (c == '_') {
			idx = 37;
		} else if (c == '!') {
			idx = 42;
		} else if (c == '/') {
			idx = 43;
		}
		return idx;
	}

	public static void drawString(GL gl, String str, float lx, float y, float s) {
		drawString(gl, str, lx, y, s,
				TO_RIGHT,
				false, 0,
				1, 1, 1);
	}
	public static void drawString(GL gl, String str, float lx, float y, float s,
			int d,
			boolean rev, float od,
			float r, float g, float b) {
		lx += LETTER_WIDTH * s / 2;
		y += LETTER_HEIGHT * s / 2;
		float x = lx;
		int idx;
		float ld = 0;
		switch (d) {
		case TO_RIGHT:
			ld = 0;
			break;
		case TO_DOWN:
			ld = 90;
			break;
		case TO_LEFT:
			ld = 180;
			break;
		case TO_UP:
			ld = 270;
			break;
		}
		ld += od;
//		for (char c: str.toCharArray()) {
		for (int i = 0; i < str.length(); i++) {
			char c = str.charAt(i);
			if (c != ' ') {
				idx = convertCharToInt(c);
				if (r == 1 && g == 1 && b == 1) {
					if (rev)
						drawLetterRev(gl, idx, x, y, s, ld);
					else
						drawLetter(gl, idx, x, y, s, ld);
				} else {
					Screen.setColor(gl, r, g, b, 0.5f);
					if (rev)
						drawLetterRev(gl, idx + LETTER_NUM, x, y, s, ld);
					else
						drawLetter(gl, idx + LETTER_NUM, x, y, s, ld);
					Screen.setColor(gl, r, g, b);
					if (rev)
						drawLetterRev(gl, idx + LETTER_NUM * 2, x, y, s, ld);
					else
						drawLetter(gl, idx + LETTER_NUM * 2, x, y, s, ld);
				}
			}
			if (od == 0) {
				switch(d) {
				case TO_RIGHT:
					x += s * LETTER_WIDTH;
					break;
				case TO_DOWN:
					y += s * LETTER_WIDTH;
					break;
				case TO_LEFT:
					x -= s * LETTER_WIDTH;
					break;
				case TO_UP:
					y -= s * LETTER_WIDTH;
					break;
				}
			} else {
				x += Math.cos(ld * Math.PI / 180) * s * LETTER_WIDTH;
				y += Math.sin(ld * Math.PI / 180) * s * LETTER_WIDTH;
			}
		}
	}

	public static void drawNum(GL gl, int num, float lx, float y, float s) {
		drawNum(gl, num, lx, y, s,
				0, -1, -1);
	}
	public static void drawNum(GL gl, int num, float lx, float y, float s,
			int dg,
			int headChar, int floatDigit) {
		lx += LETTER_WIDTH * s / 2;
		y += LETTER_HEIGHT * s / 2;
		int n = num;
		float x = lx;
		float ld = 0;
		int digit = dg;
		int fd = floatDigit;
		for (;;) {
			if (fd <= 0) {
				drawLetter(gl, n % 10, x, y, s, ld);
				x -= s * LETTER_WIDTH;
			} else {
				drawLetter(gl, n % 10, x, y + s * LETTER_WIDTH * 0.25f, s * 0.5f, ld);
				x -= s * LETTER_WIDTH * 0.5f;
			}
			n /= 10;
			digit--;
			fd--;
			if (n <= 0 && digit <= 0 && fd < 0)
				break;
			if (fd == 0) {
				drawLetter(gl, 36, x, y + s * LETTER_WIDTH * 0.25f, s * 0.5f, ld);
				x -= s * LETTER_WIDTH * 0.5f;
			}
		}
		if (headChar >= 0)
			drawLetter(gl, headChar, x + s * LETTER_WIDTH * 0.2f, y + s * LETTER_WIDTH * 0.2f,
					s * 0.6f, ld);
	}

	public static void drawNumSign(GL gl, int num, float lx, float ly, float s) {
		drawNumSign(gl, num, lx, ly, s,
				-1, -1, 0);
	}

	public static void drawNumSign(GL gl, int num, float lx, float ly, float s,
			int headChar, int floatDigit, int type) {
		float x = lx;
		float y = ly;
		int n = num;
		int fd = floatDigit;
		for (;;) {
			if (fd <= 0) {
				drawLetterRev(gl, n % 10 + type * LETTER_NUM, x, y, s, 0);
				x -= s * LETTER_WIDTH;
			} else {
				drawLetterRev(gl, n % 10 + type * LETTER_NUM, x, y - s * LETTER_WIDTH * 0.25f, s * 0.5f, 0);
				x -= s * LETTER_WIDTH * 0.5f;
			}
			n /= 10;
			if (n <= 0)
				break;
			fd--;
			if (fd == 0) {
				drawLetterRev(gl, 36 + type * LETTER_NUM, x, y - s * LETTER_WIDTH * 0.25f, s * 0.5f, 0);
				x -= s * LETTER_WIDTH * 0.5f;
			}
		}
		if (headChar >= 0)
			drawLetterRev(gl, headChar + type * LETTER_NUM, x + s * LETTER_WIDTH * 0.2f, y - s * LETTER_WIDTH * 0.2f,
					s * 0.6f, 0);
	}

	public static void drawTime(GL gl, int time, float lx, float y, float s) {
		int n = time;
		if (n < 0)
			n = 0;
		float x = lx;
		for (int i = 0; i < 7; i++) {
			if (i != 4) {
				drawLetter(gl, n % 10, x, y, s, TO_RIGHT);
				n /= 10;
			} else {
				drawLetter(gl, n % 6, x, y, s, TO_RIGHT);
				n /= 6;
			}
			if ((i & 1) == 1 || i == 0) {
				switch (i) {
				case 3:
					drawLetter(gl, 41, x + s * 1.16f, y, s, TO_RIGHT);
					break;
				case 5:
					drawLetter(gl, 40, x + s * 1.16f, y, s, TO_RIGHT);
					break;
				default:
					break;
				}
				x -= s * LETTER_WIDTH;
			} else {
				x -= s * LETTER_WIDTH * 1.3f;
			}
			if (n <= 0)
				break;
		}
	}

	private static void setLetter(GL gl, int idx) {
		setLetter(gl, idx, NORMAL);
	}

	private static void setLetter(GL gl, int idx, int type) {
		float x, y, length, size, t;
		float deg;
		for (int i = 0;; i++) {
			deg = (int) spData[idx][i][4];
			if (deg > 99990)
				break;
			x = -spData[idx][i][0];
			y = -spData[idx][i][1];
			size = spData[idx][i][2];
			length = spData[idx][i][3];
			y *= 0.9;
			size *= 1.4;
			length *= 1.05;
			x = -x;
			y = y;
			deg %= 180;
			switch (type) {
			case NORMAL:
				drawSegment(gl, x, y, size, length, deg);
				break;
			case POLYGON:
				drawSegmentPolygon(gl, x, y, size, length, deg);
				break;
			case LINE:
				drawSegmentLine(gl, x, y, size, length, deg);
				break;
			}
		}
	}

	private static void drawSegment(GL gl, float x, float y, float width, float height, float deg) {
		gl.glPushMatrix();
		gl.glTranslatef(x - width / 2, y, 0);
		gl.glRotatef(deg, 0, 0, 1);
		Screen.setColor(gl, 1, 1, 1, 0.5f);
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		drawSegmentPart(gl, width, height);
		gl.glEnd();
		Screen.setColor(gl, 1, 1, 1);
		gl.glBegin(GL.GL_LINE_LOOP);
		drawSegmentPart(gl, width, height);
		gl.glEnd();
		gl.glPopMatrix();
	}

	private static void drawSegmentPolygon(GL gl, float x, float y, float width, float height, float deg) {
		gl.glPushMatrix();
		gl.glTranslatef(x - width / 2, y, 0);
		gl.glRotatef(deg, 0, 0, 1);
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		drawSegmentPart(gl, width, height);
		gl.glEnd();
		gl.glPopMatrix();
	}

	private static void drawSegmentLine(GL gl, float x, float y, float width, float height, float deg) {
		gl.glPushMatrix();
		gl.glTranslatef(x - width / 2, y, 0);
		gl.glRotatef(deg, 0, 0, 1);
		gl.glBegin(GL.GL_LINE_LOOP);
		drawSegmentPart(gl, width, height);
		gl.glEnd();
		gl.glPopMatrix();
	}

	private static void drawSegmentPart(GL gl, float width, float height) {
		gl.glVertex3f(-width / 2, 0, 0);
		gl.glVertex3f(-width / 3 * 1, -height / 2, 0);
		gl.glVertex3f( width / 3 * 1, -height / 2, 0);
		gl.glVertex3f( width / 2, 0, 0);
		gl.glVertex3f( width / 3 * 1,  height / 2, 0);
		gl.glVertex3f(-width / 3 * 1,  height / 2, 0);
	}

	private static float[][][] spData = 
	{{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.6f, 0.55f, 0.65f, 0.3f, 90}, {0.6f, 0.55f, 0.65f, 0.3f, 90},
		{-0.6f, -0.55f, 0.65f, 0.3f, 90}, {0.6f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{0.5f, 0.55f, 0.65f, 0.3f, 90},
		{0.5f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{//A
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{-0.18f, 1.15f, 0.45f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.45f, 0.55f, 0.65f, 0.3f, 90},
		{-0.18f, 0, 0.45f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{-0.15f, 1.15f, 0.45f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.45f, 0.45f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{//F
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0.05f, 0, 0.3f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 0.55f, 0.65f, 0.3f, 90},
		{0, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0.65f, -0.55f, 0.65f, 0.3f, 90}, {-0.7f, -0.7f, 0.3f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{//K
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.4f, 0.55f, 0.65f, 0.3f, 100},
		{-0.25f, 0, 0.45f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.6f, -0.55f, 0.65f, 0.3f, 80},
		{0, 0, 0, 0, 99999},
	},{
		{-0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{-0.5f, 1.15f, 0.3f, 0.3f, 0}, {0.1f, 1.15f, 0.3f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0.55f, 0.65f, 0.3f, 90},
		{0, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{//P
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0.05f, -0.55f, 0.45f, 0.3f, 60},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.2f, 0, 0.45f, 0.3f, 0},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.45f, -0.55f, 0.65f, 0.3f, 80},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{-0.65f, 0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0.65f, 0.3f, 0},
		{0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{-0.5f, 1.15f, 0.55f, 0.3f, 0}, {0.5f, 1.15f, 0.55f, 0.3f, 0},
		{0.1f, 0.55f, 0.65f, 0.3f, 90},
		{0.1f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{//U
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.5f, -0.55f, 0.65f, 0.3f, 90}, {0.5f, -0.55f, 0.65f, 0.3f, 90},
		{-0.1f, -1.15f, 0.45f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{
		{-0.65f, 0.55f, 0.65f, 0.3f, 90}, {0.65f, 0.55f, 0.65f, 0.3f, 90},
		{-0.65f, -0.55f, 0.65f, 0.3f, 90}, {0.65f, -0.55f, 0.65f, 0.3f, 90},
		{-0.5f, -1.15f, 0.3f, 0.3f, 0}, {0.1f, -1.15f, 0.3f, 0.3f, 0},
		{0, 0.55f, 0.65f, 0.3f, 90},
		{0, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{-0.4f, 0.6f, 0.85f, 0.3f, 360-120},
		{0.4f, 0.6f, 0.85f, 0.3f, 360-60},
		{-0.4f, -0.6f, 0.85f, 0.3f, 360-240},
		{0.4f, -0.6f, 0.85f, 0.3f, 360-300},
		{0, 0, 0, 0, 99999},
	},{
		{-0.4f, 0.6f, 0.85f, 0.3f, 360-120},
		{0.4f, 0.6f, 0.85f, 0.3f, 360-60},
		{-0.1f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{
		{0, 1.15f, 0.65f, 0.3f, 0},
		{0.3f, 0.4f, 0.65f, 0.3f, 120},
		{-0.3f, -0.4f, 0.65f, 0.3f, 120},
		{0, -1.15f, 0.65f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{//.
		{0, -1.15f, 0.3f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{//_
		{0, -1.15f, 0.8f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{//-
		{0, 0, 0.9f, 0.3f, 0},
		{0, 0, 0, 0, 99999},
	},{//+
		{-0.5f, 0, 0.45f, 0.3f, 0}, {0.45f, 0, 0.45f, 0.3f, 0},
		{0.1f, 0.55f, 0.65f, 0.3f, 90},
		{0.1f, -0.55f, 0.65f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{//'
		{0, 1.0f, 0.4f, 0.2f, 90},
		{0, 0, 0, 0, 99999},
	},{//''
		{-0.19f, 1.0f, 0.4f, 0.2f, 90},
		{0.2f, 1.0f, 0.4f, 0.2f, 90},
		{0, 0, 0, 0, 99999},
	},{//!
		{0.56f, 0.25f, 1.1f, 0.3f, 90},
		{0, -1.0f, 0.3f, 0.3f, 90},
		{0, 0, 0, 0, 99999},
	},{// /
		{0.8f, 0, 1.75f, 0.3f, 120},
		{0, 0, 0, 0, 99999},
	}};
}
