/*
 * $Id: particle.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import javax.media.opengl.GL;

import abagames.util.ActorPool;
import abagames.util.Rand;
import abagames.util.Vector2;
import abagames.util.Vector3;

/**
 * Particles (Triangle / Line / Quad / Bonus).
 */
class ParticlePool extends ActorPool<Particle> {
	public ParticlePool() {
		super(new ActorCreator<Particle>() {
			public Particle create() {
				return new Particle();
			}
			public Particle[] createArray(int n) {
				return new Particle[n];
			}
			
		});
	}
}

public class Particle extends Token<ParticleState, ParticleSpec> {
	public static final int TRIANGLE = 0, LINE = 1, QUAD = 2, BONUS = 3;
	private TriangleParticleSpec triangleParticleSpec;
	private LineParticleSpec lineParticleSpec;
	private QuadParticleSpec quadParticleSpec;
	private BonusParticleSpec bonusParticleSpec;

	public void init(Object[] args) {
		super.init(args);
		state = new ParticleState();
		triangleParticleSpec = (TriangleParticleSpec) args[0];
		lineParticleSpec = (LineParticleSpec) args[1];
		quadParticleSpec = (QuadParticleSpec) args[2];
		bonusParticleSpec = (BonusParticleSpec) args[3];
	}

	  public void set(int type,
	                  float x, float y, float deg, float speed,
	                  float sz, float r, float g, float b) {
		  set(type, x, y, deg, speed, sz, r, g, b, 60, true, 0);
	  }

	  public void set(int type,
	                  float x, float y, float deg, float speed,
	                  float sz, float r, float g, float b,
	                  int c) {
		  set(type, x, y, deg, speed, sz, r, g, b, c, true, 0);
	  }
	  public void set(int type,
                  float x, float y, float deg, float speed,
                  float sz, float r, float g, float b,
                  int c, boolean ebg, float num) {
		  switch (type) {
		  case TRIANGLE:
			  spec = triangleParticleSpec;
			  break;
		  case LINE:
			  spec = lineParticleSpec;
			  break;
		  case QUAD:
			  spec = quadParticleSpec;
			  break;
		  case BONUS:
			  spec = bonusParticleSpec;
			  break;
		  }
//		  this.spec = spec;
		  super.set(x, y, deg, speed);
		  state.size = sz;
		  state.vel.x = -(float)Math.sin(deg) * speed;
		  state.vel.y = (float)Math.cos(deg) * speed;
		  state.r = r;
		  state.g = g;
		  state.b = b;
		  state.cnt = state.startCnt = c;
		  state.effectedByGravity = ebg;
		  state.trgNum = num;
		  if (type == BONUS)
			  ((BonusParticleSpec) spec).setSize(state);
	  }

	  public void setByVelocity(float x, float y, float vx, float vy,
			  float sz, float r, float g, float b, float a) {
		  setByVelocity(x, y, vx, vy,
				  sz, r, g, b, a,
				  60, true);
	  }
	  
	  public void setByVelocity(float x, float y, float vx, float vy,
			  float sz, float r, float g, float b, float a,
			  int c, boolean ebg) {
		  spec = triangleParticleSpec;
		  super.set(x, y, 0, 0);
		  state.vel.x = vx;
		  state.vel.y = vy;
		  state.size = sz;
		  state.r = r;
		  state.g = g;
		  state.b = b;
		  state.a = a;
		  state.cnt = state.startCnt = c;
		  state.effectedByGravity = ebg;
	  }
}

class ParticleState extends TokenState {
  Vector2 vel;
  Vector2 tailPos;
  float size;
  int cnt, startCnt;
  float r, g, b, a;
  float d1, d2;
  float vd1, vd2;
  boolean effectedByGravity;
  float num, trgNum;
  float trgSize;

//  invariant {
//    if (isInitialized) {
//      assert(pos.x <>= 0);
//      assert(pos.y <>= 0);
//      assert(vel.x <>= 0);
//      assert(vel.y <>= 0);
//      assert(tailPos.x <>= 0);
//      assert(tailPos.y <>= 0);
//      assert(size > 0 && size < 10);
//      assert(r >= 0 && r <= 1);
//      assert(g >= 0 && g <= 1);
//      assert(b >= 0 && b <= 1);
//      assert(a >= 0 && a <= 1);
//      assert(d1 <>= 0);
//      assert(d2 <>= 0);
//      assert(vd1 <>= 0);
//      assert(vd2 <>= 0);
//      assert(num <>= 0);
//      assert(trgNum <>= 0);
//      assert(trgSize > 0);
//    }
//  }

  public ParticleState() {
    super();
    vel = new Vector2();
    tailPos = new Vector2();
  }

  public void clear() {
    vel.x = vel.y = 0;
    size = 1;
    cnt = 0;
    r = g = b = 0;
    a = 1;
    d1 = d2 = 0;
    vd1 = vd2 = 0;
    effectedByGravity = false;
    num = trgNum = 1;
    trgSize = 1;
    super.clear();
  }
}

class ParticleSpec extends TokenSpec<ParticleState> {
	public static Rand rand = new Rand();
//  mixin StaticRandImpl;
}

class TriangleParticleSpec extends ParticleSpec {
	static final float SLOW_DOWN_RATIO = 0.05f;
  static final float GRAVITY = 0.003f;
  Shape particleShape;
  ParticlePool particles;

  public TriangleParticleSpec(GL gl, Field field) {
    this.field = field;
    particleShape = new TriangleParticleShape(gl);
  }

  public void setParticles(ParticlePool particles) {
    this.particles = particles;
  }

  public void set(ParticleState ps) {
//    with (ps) {
	  {
      ps.d1 = rand.nextFloat((float)Math.PI * 2);
      ps.d2 = rand.nextFloat((float)Math.PI * 2);
      ps.vd1 = rand.nextSignedFloat(0.1f);
      ps.vd2 = rand.nextSignedFloat(0.1f);
    }
  }

  public boolean move(ParticleState ps) {
//    with (ps) {
	  {
		  ps.pos.opAddAssign(ps.vel);
		  ps.pos.x = field.normalizeX(ps.pos.x);
      if (ps.effectedByGravity)
    	  ps.vel.y -= GRAVITY;
      ps.vel.opMulAssign(1 - SLOW_DOWN_RATIO);
      ps.d1 += ps.vd1;
      ps.d2 += ps.vd2;
      ps.vd1 *= (1 - SLOW_DOWN_RATIO * 0.2f);
      ps.vd2 *= (1 - SLOW_DOWN_RATIO * 0.2f);
      float cfr = 1.0f - (1.0f / (float) ps.startCnt);
      if (cfr < 0)
        cfr = 0;
      ps.r *= cfr;
      ps.g *= cfr;
      ps.b *= cfr;
      ps.a *= cfr;
      float fs = 0;
      if (ps.size > 2.0f && rand.nextInt(45) == 0)
        fs = 0.5f - rand.nextFloat(0.2f);
      else if (ps.size > 0.5f && rand.nextInt(10) == 0)
        fs = 0.1f + rand.nextSignedFloat(0.05f);
      if (fs > 0) {
        float vx = ps.vel.x * rand.nextSignedFloat(0.8f);
        float vy = ps.vel.y * rand.nextSignedFloat(0.8f);
        ps.vel.x -= vx * fs;
        ps.vel.y -= vy * fs;
        float cr = 1 - fs * 0.2f;
        ps.vel.opDivAssign(cr);
        Particle p = particles.getInstanceForced();
        int nc = (int) (ps.cnt * (0.8f + fs * 0.2f));
        if (nc > 0)
          p.setByVelocity(ps.pos.x, ps.pos.y, vx, vy, ps.size * fs, ps.r, ps.g, ps.b, ps.a,
                          nc, ps.effectedByGravity);
        ps.size *= (1 - fs);
        ps.cnt *= cr;
      }
      ps.cnt--;
      if (ps.cnt <= 0)
        return false;
      return true;
    }
  }

  public void draw(GL gl, ParticleState ps) {
//    with (ps) {
	  {
      Vector3 p = field.calcCircularPos(ps.pos);
      Screen.setColor(gl, ps.r, ps.g, ps.b, ps.a);
      particleShape.draw(gl, p, ps.d1, ps.d2);
    }
  }
}

class LineParticleSpec extends ParticleSpec {
  static final float SLOW_DOWN_RATIO = 0.03f;

  public LineParticleSpec(Field field) {
    this.field = field;
  }

  public void set(ParticleState ps) {
//    with (ps) {
	  {
      ps.tailPos.x = ps.pos.x;
      ps.tailPos.y = ps.pos.y;
    }
  }


  public boolean move(ParticleState ps) {
//    with (ps) {
	  {
      ps.stepForward();
      ps.tailPos.x += (ps.pos.x - ps.tailPos.x) * 0.05f;
      ps.tailPos.y += (ps.pos.y - ps.tailPos.y) * 0.05f;
      ps.speed *= (1 - SLOW_DOWN_RATIO);
      ps.pos.x = field.normalizeX(ps.pos.x);
      float cfr = 1.0f - (1.0f / (float) ps.startCnt);
      if (cfr < 0)
        cfr = 0;
      ps.r *= cfr;
      ps.g *= cfr;
      ps.b *= cfr;
      ps.a *= cfr;
      ps.cnt--;
      if (ps.cnt <= 0)
        return false;
      return true;
    }
  }

  public void draw(GL gl, ParticleState ps) {
//    with (ps) {
	  {
      Vector3 p;
      gl.glBegin(GL.GL_LINES);
      Screen.setColor(gl, ps.r, ps.g, ps.b, ps.a);
      p = field.calcCircularPos(ps.pos);
      Screen.glVertex(gl, p);
      p = field.calcCircularPos(ps.tailPos);
      Screen.glVertex(gl, p);
      gl.glEnd();
    }
  }
}

class QuadParticleSpec extends ParticleSpec {
// private:
  static final float SLOW_DOWN_RATIO = 0.07f;
  static final float GRAVITY = 0.002f;

  public QuadParticleSpec(Field field) {
    this.field = field;
  }

  public boolean move(ParticleState ps) {
//    with (ps) {
	  {
      ps.pos.opAddAssign(ps.vel);
      ps.pos.x = field.normalizeX(ps.pos.x);
      if (ps.effectedByGravity)
    	  ps.vel.y -= GRAVITY;
      ps.vel.opMulAssign((1 - SLOW_DOWN_RATIO));
      float cfr = 1.0f - (1.0f / (float) ps.startCnt);
      if (cfr < 0)
        cfr = 0;
      ps.r *= cfr;
      ps.g *= cfr;
      ps.b *= cfr;
      ps.a *= cfr;
      ps.size *= (1 - (1 - cfr) * 0.5f);
      ps.cnt--;
      if (ps.cnt <= 0)
        return false;
      return true;
    }
  }

  public void draw(GL gl, ParticleState ps) {
//    with (ps) {
	  {
      Vector3 p;
      float sz = ps.size * 0.5f;
      Screen.setColor(gl, ps.r, ps.g, ps.b, ps.a);
      gl.glBegin(GL.GL_QUADS);
      p = field.calcCircularPos(ps.pos.x - sz, ps.pos.y - sz);
      Screen.glVertex(gl, p);
      p = field.calcCircularPos(ps.pos.x + sz, ps.pos.y - sz);
      Screen.glVertex(gl, p);
      p = field.calcCircularPos(ps.pos.x + sz, ps.pos.y + sz);
      Screen.glVertex(gl, p);
      p = field.calcCircularPos(ps.pos.x - sz, ps.pos.y + sz);
      Screen.glVertex(gl, p);
      gl.glEnd();
      gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
      Screen.setColor(gl, 0, 0, 0, ps.a * 0.66f);
      gl.glBegin(GL.GL_LINE_LOOP);
      p = field.calcCircularPos(ps.pos.x - sz, ps.pos.y - sz);
      Screen.glVertex(gl, p);
      p = field.calcCircularPos(ps.pos.x + sz, ps.pos.y - sz);
      Screen.glVertex(gl, p);
      p = field.calcCircularPos(ps.pos.x + sz, ps.pos.y + sz);
      Screen.glVertex(gl, p);
      p = field.calcCircularPos(ps.pos.x - sz, ps.pos.y + sz);
      Screen.glVertex(gl, p);
      gl.glEnd();
      gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
    }
  }
}

class BonusParticleSpec extends ParticleSpec {
  static final float SLOW_DOWN_RATIO = 0.04f;

  public BonusParticleSpec(Field field) {
    this.field = field;
  }

  public void setSize(ParticleState ps) {
//    with (ps) {
	  {
		  ps.trgSize = 0.5f;
		  ps.size = 0.1f;
    }
  }

  public boolean move(ParticleState ps) {
//    with (ps) {
	  {
      ps.stepForward();
      ps.speed *= (1 - SLOW_DOWN_RATIO);
      ps.pos.x = field.normalizeX(ps.pos.x);
      float cfr = 1.0f - (1.0f / (float) ps.startCnt);
      if (cfr < 0)
        cfr = 0;
      ps.a *= cfr;
      ps.num += (ps.trgNum - ps.num) * 0.2f;
      ps.size += (ps.trgSize - ps.size) * 0.1f;
      ps.cnt--;
      if (ps.cnt <= 0)
        return false;
      return true;
    }
  }

  public void draw(GL gl, ParticleState ps) {
//    with (ps) {
	  {
      gl.glPushMatrix();
      Vector3 p = field.calcCircularPos(ps.pos);
      Screen.setColor(gl, 1, 1, 1, ps.a * 0.5f);
      Screen.glTranslate(gl, p);
      Letter.drawNumSign(gl, (int) (ps.num * 100), 0, 0, ps.size, 33, 2, 1);
      Screen.setColor(gl, 1, 1, 1, ps.a);
      Letter.drawNumSign(gl, (int) (ps.num * 100), 0, 0, ps.size, 33, 2, 2);
      gl.glPopMatrix();
    }
  }
}
