/*
 * $Id: pillar.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import java.util.Arrays;

import javax.media.opengl.GL;

import abagames.util.ActorPool;

/**
 * Pillars at the center and on the background.
 */
class PillarPool extends ActorPool<Pillar> {
	public PillarPool() {
		super(new ActorCreator<Pillar>() {
			public Pillar create() {
				return new Pillar();
			}

			public Pillar[] createArray(int n) {
				return new Pillar[n];
			}
		});
	}

	public void setEnd() {
		for (Pillar a: actors)
			if (a.exists())
				a.setEnd();
	}

  public void drawCenter(GL gl) {
	  Pillar[] sas = (Pillar[])actors.clone();
	  Arrays.sort(sas);
//	  Pillar[] sas = actors.sort();
    for (Pillar a: sas)
      if (a.exists() && !a.state.isOutside)
        a.draw(gl);
  }

  public void drawOutside(GL gl) {
    for (Pillar a: actors)
      if (a.exists() && a.state.isOutside)
        a.draw(gl);
  }
}

public class Pillar extends Token<PillarState, PillarSpec> implements Comparable<Pillar>{
	public Pillar() {
		state = new PillarState();
	}
	
	public void set(PillarSpec ps, float y, float maxY, Pillar pp, PillarShape s, float vdeg) {
		set(ps, y, maxY, pp, s, vdeg, false);
	}
	public void set(PillarSpec ps, float y, float maxY, Pillar pp, PillarShape s, float vdeg, boolean outside) {
		super.set(ps, 0, y, 0, 0);
		state.maxY = maxY;
		state.previousPillar = pp;
		state.pshape = s;
		state.vdeg = vdeg;
		state.isOutside = outside;
	}

	public void setEnd() {
		state.isEnded = true;
	}

//	public int opCmp(Object o) {
//	Pillar p = (Pillar) o;
//	if (p == null)
//	return 0;
//	return (int) (Math.abs(p.pos().y) - Math.abs(pos().y));
//	}
	public int compareTo(Pillar p) {
		if (p == null)
			return 0;
		return (int) (Math.abs(p.pos().y) - Math.abs(pos().y));
	}
}

class PillarState extends TokenState {
	Pillar previousPillar;
	float vy, vdeg;
	float maxY;
	PillarShape pshape;
	boolean isEnded;
	boolean isOutside;

	public void clear() {
		previousPillar = null;
		vy = 0;
		vdeg = 0;
		maxY = 0;
		isEnded = false;
		isOutside = false;
		super.clear();
	}
}

class PillarSpec extends TokenSpec<PillarState> {
	static final float VELOCITY_Y = 0.025f;

	public PillarSpec(Field field) {
		this.field = field;
	}

	public boolean move(PillarState ps) {
//		with (ps) {
		{
			if (!ps.isOutside) {
				ps.vy += VELOCITY_Y;
				ps.vy *= 0.98f;
				ps.pos.y += ps.vy;
				if (ps.vy > 0) {
					float ty;
					if (ps.previousPillar != null && ps.previousPillar.exists())
						ty = ps.previousPillar.pos().y - PillarShape.TICKNESS;
					else
						ty = ps.maxY;
					ty -= PillarShape.TICKNESS;
					if (!ps.isEnded && ps.pos.y > ty) {
						ps.vy *= -0.5f;
						ps.pos.y += (ty - ps.pos.y) * 0.5f;
						if (ps.previousPillar != null)
							ps.previousPillar.state.vy -= ps.vy * 0.5f;
					}
					if (ps.pos.y > 100)
						return false;
				}
			} else {
				ps.pos.y -= 0.2f;
				if (ps.pos.y < -50)
					return false;
			}
			ps.deg += ps.vdeg;
			return true;
		}
	}

	public void draw(GL gl, PillarState ps) {
		ps.pshape.draw(gl, ps.pos.y, ps.deg);
	}
}
