/*
 * $Id: player.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import java.util.ArrayList;

import javax.media.opengl.GL;

import abagames.util.ActorPool;
import abagames.util.Rand;
import abagames.util.Vector2;
import abagames.util.Vector3;
import abagames.util.sdl.NoRecordDataException;
import abagames.util.sdl.Pad;
import abagames.util.sdl.Pad.PadState;
import abagames.util.sdl.Pad.RecordablePad;

/**
 * Player and shots.
 */
public class Player extends Token<PlayerState, PlayerSpec> {
	Vector2 hitOffset;

	public Player(GL gl, PlayerSpec spec) {
		state = new PlayerState();
		this.spec = spec;
		spec.setState(gl, state);
		hitOffset = new Vector2();
	}

	public boolean replayMode(boolean v) {
		return state.replayMode = v;
	}

	public void set() {
		state.set();
		spec.clear();
		hitOffset.x = hitOffset.y = 0;
		spec.field.setEyePos(pos());
	}

	public boolean checkBulletHit(Vector2 p, Vector2 pp) {
//		with (state) {
		{
			if (!hasCollision())
				return false;
			if (spec.field.checkHitDist(pos(), p, pp, PlayerSpec.BULLET_HIT_WIDTH)) {
				destroy();
				return true;
			}
			return false;
		}
	}

	public boolean checkEnemyHit(Vector2 p, Vector2 v, Vector2 size) {
//		with (state) {
		{
			if (!hasCollision())
				return false;
			if (Math.abs(pos().x - p.x) < size.x && Math.abs(pos().y - p.y) < size.y) {
				//destroy();
				hitOffset.x = pos().x - p.x;
				hitOffset.y = pos().y - p.y;
				spec.addVelocity(state, v, hitOffset);
				return true;
			}
			return false;
		}
	}

	public void destroy() {
		remove();
		spec.destroyed(state);
	}

	public void destroyCapturedEnemies(int idx) {
		state.destroyCapturedEnemies(idx);
	}

	public boolean isInTractorBeam(Vector2 p) {
		return spec.tractorBeam.contains(p);
	}

	public int addCapturedEnemy(Enemy e) {
		return state.addCapturedEnemy(e);
	}

	public float capturedEnemyWidth() {
		return state.capturedEnemyWidth;
	}

	public void addScore(int sc) {
		spec.addScore(sc);
	}

	public void addMultiplier(float mp) {
		spec.addMultiplier(mp);
	}

	public float multiplier() {
		return spec.multiplier();
	}

	public float deg() {
		return state.deg;
	}

	public boolean isActive() {
		return state.isActive();
	}

	public boolean hasCollision() {
		return state.hasCollision();
	}
}

class PlayerState extends TokenState {
	public boolean replayMode;
	static final int RESPAWN_INTERVAL = 72;
	static final int INVINCIBLE_INTERVAL_RESPAWN = 240;
	static final int MAX_CAPTURED_ENEMIES_NUM = 10;
	Enemy[] capturedEnemies;
	int capturedEnemyNum;
	int respawnCnt;
	boolean isInRespawn;
	int invincibleCnt;
	boolean isInvincible;
	int shotCnt;
	int capturedEnemyShotCnt;
	boolean aPressed, bPressed;
	Vector2 vel;
	float capturedEnemyWidth;
	int colorCnt;
	boolean isFirstShot;
	
//	invariant {
//		if (isInitialized) {
//			assert(vel.x <>= 0);
//			assert(vel.y <>= 0);
//			assert(capturedEnemyWidth >= 0);
//		}
//	}

	public PlayerState() {
		capturedEnemies = new Enemy[MAX_CAPTURED_ENEMIES_NUM];
		vel = new Vector2();
	}

	public void set() {
		reset();
		pos.x = 0;
		respawnCnt = 0;
		isInRespawn = false;
		aPressed = bPressed = true;
		shotCnt = 60;
	}

	public void clear() {
		capturedEnemyNum = 0;
		respawnCnt = invincibleCnt = 0;
		isInRespawn = isInvincible = false;
		shotCnt = 0;
		capturedEnemyShotCnt = 0;
		vel.x = vel.y = 0;
		capturedEnemyWidth = 1.0f;
		colorCnt = 0;
		isFirstShot = false;
		super.clear();
	}

	private void reset() {
		float x = pos.x;
		clear();
		pos.x = x;
		pos.y = -10.0f;
		speed = PlayerSpec.BASE_SPEED;
		invincibleCnt = INVINCIBLE_INTERVAL_RESPAWN;
		isInvincible = true;
		isFirstShot = true;
	}

	public void move() {
		colorCnt++;
		if (isInRespawn) {
			respawnCnt--;
			if (respawnCnt <= 0) {
				reset();
				isInRespawn = false;
			}
		} else if (isInvincible) {
			invincibleCnt--;
			if (invincibleCnt <= 0)
				isInvincible = false;
		}
	}

	public boolean isActive() {
		return !isInRespawn;
	}

	public boolean hasCollision() {
		return (!isInRespawn && !isInvincible);
	}

	public boolean hasShape() {
		if (isInRespawn)
			return false;
		if (!isInvincible)
			return true;
		if (invincibleCnt % 60 < 30)
			return false;
		else
			return true;
	}

	public void destroyed() {
		respawnCnt = RESPAWN_INTERVAL;
		destroyCapturedEnemies(0);
		isInRespawn = true;
	}

	public int addCapturedEnemy(Enemy e) {
		if (isInRespawn || capturedEnemyNum >= MAX_CAPTURED_ENEMIES_NUM)
			return -1;
		capturedEnemies[capturedEnemyNum] = e;
		capturedEnemyNum++;
		return capturedEnemyNum - 1;
	}

	public void destroyCapturedEnemies(int idx) {
		for (int i = idx; i < capturedEnemyNum; i++)
			if (capturedEnemies[i].exists())
				capturedEnemies[i].destroyed();
		capturedEnemyNum = idx;
	}
}

class PlayerSpec extends TokenSpec<PlayerState> {
//	mixin StaticRandImpl;
	public static Rand rand = new Rand();
	public static final float BASE_SPEED = 0.15f;
	public static final float BASE_VELOCITY = 0.03f;
	public static final float BULLET_HIT_WIDTH = 0.2f;
	public static final float CAPTURED_ENEMIES_INTERVAL_LENGTH = 1.2f;
	public static final int SHOT_MAX_NUM = 3;
	private static final float TILT_DEG = 1.0f;
	private static final float SHOT_INTERVAL = 3;
	private static final float FIRST_SHOT_INTERVAL = 6;
	private static final int TWIN_SHOT_MAX_NUM = 2;
	ShotPool shots, capturedEnemiesShots;
	ShotSpec shotSpec;
	EnemyPool enemies;
	BulletPool bullets;
	ParticlePool particles;
	RecordablePad pad;
	GameState gameState;
	PlayerState playerState;
	TractorBeam tractorBeam;
	Shape lineShape;

	public PlayerSpec(GL gl, Pad pad, GameState gameState, Field field,
			EnemyPool enemies, BulletPool bullets, ParticlePool particles) {
		this.pad = (RecordablePad) pad;
		this.gameState = gameState;
		this.field = field;
		this.enemies = enemies;
		this.bullets = bullets;
		this.particles = particles;
		shots = new ShotPool(new ActorPool.ActorCreator<Shot>() {
			public Shot create() {
				return new Shot();
			}

			public Shot[] createArray(int n) {
				return new Shot[n];
			}
		});
		shots.init(SHOT_MAX_NUM, null);
		capturedEnemiesShots = new ShotPool(new ActorPool.ActorCreator<Shot>() {
			public Shot create() {
				return new Shot();
			}

			public Shot[] createArray(int n) {
				return new Shot[n];
			}
		});
		capturedEnemiesShots.init(SHOT_MAX_NUM * PlayerState.MAX_CAPTURED_ENEMIES_NUM, null);
		shotSpec = new ShotSpec(gl, field, enemies, bullets, gameState);
		shape = new PlayerShape(gl);
		lineShape = new PlayerLineShape(gl);
	}

	public void setState(GL gl, PlayerState ps) {
		playerState = ps;
		shotSpec.setPlayerState(ps);
		tractorBeam = new TractorBeam(gl, field, ps);
	}

	public void close(GL gl) {
		((PlayerShape) shape).close(gl);
		shotSpec.close(gl);
	}

	public void clear() {
		tractorBeam.clear();
		shots.clear();
		capturedEnemiesShots.clear();
	}

	public boolean move(PlayerState ps) {
//		with (ps) {
		{
			PadState input;
			if (!ps.replayMode) {
				input = pad.getState();
			} else {
				try {
					input = pad.input.replay();
				} catch (NoRecordDataException e) {
//					e.printStackTrace();
					gameState.startGameOverWithoutRecording();
					input = pad.getNullState();
				}
			}
			shots.move();
			capturedEnemiesShots.move();
			capturedEnemiesShots.checkParent();
			if (gameState.isGameOver()) {
				if ((input.button & PadState.A) != 0) {
					if (!ps.aPressed) {
						ps.aPressed = true;
						if (!ps.replayMode)
							gameState.backToTitle();
					}
				} else {
					ps.aPressed = false;
				}
				return true;
			}
			ps.move();
			if (!ps.isActive())
				return true;
			float vx = 0, vy = 0;
			if ((input.dir & PadState.RIGHT) != 0)
				vx = 1;
			else if ((input.dir & PadState.LEFT) != 0)
				vx = -1;
			if ((input.dir & PadState.UP) != 0)
				vy = 1;
			else if ((input.dir & PadState.DOWN) != 0)
				vy = -1;
			if (vx != 0 && vy != 0) {
				vx *= 0.7f;
				vy *= 0.7f;
			}
			ps.pos.x += (vx * ps.speed);
			ps.pos.y += (vy * ps.speed);
			if ((input.button & PadState.B) == 0)
				ps.deg += (-TILT_DEG * (vx * ps.speed) - ps.deg) * 0.1f;
//			assert(deg <>= 0);
			ps.pos.opAddAssign(ps.vel);
			ps.vel.opMulAssign(0.9f);
			if ((input.button & PadState.A) != 0&& (input.button & Pad.PadState.B) == 0) {
				if (ps.shotCnt <= 0) {
					Shot s = shots.getInstance();
					if (s != null) {
						s.set(shotSpec, ps.pos, ps.deg, 0.66f);
						if (ps.isFirstShot) {
							ps.isFirstShot = false;
							ps.shotCnt += FIRST_SHOT_INTERVAL;
						} else {
							ps.shotCnt += SHOT_INTERVAL;
						}
						gameState.countShotFired();
						addShotParticle(ps.pos, ps.deg);
						Sound.playSe("shot.wav");
						for (int i = 0; i < ps.capturedEnemyNum; i++) {
							if (ps.capturedEnemies[i].isCaptured()) {
								Shot ces = capturedEnemiesShots.getInstance();
								if (ces == null)
									break;
								ces.set(shotSpec, ps.capturedEnemies[i].pos(), ps.deg, 0.66f);
								ces.setParent(s);
								addShotParticle(ps.capturedEnemies[i].pos(), ps.deg);
							}
						}
					}
				}
			} else {
				ps.isFirstShot = true;
			}
			if ((input.button & PadState.B) != 0) {
				ps.speed += (BASE_SPEED * 1.2f - ps.speed) * 0.33f;
				ps.deg *= 0.9f;
				/*capturedEnemyWidth -= 0.05f;
        if (capturedEnemyWidth < 0.2f)
          capturedEnemyWidth = 0.2f;*/
			} else {
				ps.speed += (BASE_SPEED * 2.0f - ps.speed) * 0.33f;
				/*capturedEnemyWidth += 0.05f;
        if (capturedEnemyWidth > 1)
          capturedEnemyWidth = 1;*/
			}
			if ((input.button & PadState.B) != 0&&
					ps.capturedEnemyNum < PlayerState.MAX_CAPTURED_ENEMIES_NUM) {
				tractorBeam.extendLength();
			} else {
				tractorBeam.reduceLength();
			}
			tractorBeam.move();
			if (ps.shotCnt > 0)
				ps.shotCnt--;
			if (ps.capturedEnemyShotCnt > 0)
				ps.capturedEnemyShotCnt--;
			//if (pos.y > field.size.y)
			//pos.y = field.size.y;
			if (ps.pos.y > 0)
				ps.pos.y = 0;
			else if (ps.pos.y < -field.size().y)
				ps.pos.y = -field.size().y;
			if (ps.pos.x > field.size().x)
				ps.pos.x = field.size().x;
			else if (ps.pos.x < -field.size().x)
				ps.pos.x = -field.size().x;
			ps.pos.x = field.normalizeX(ps.pos.x);
			field.setEyePos(ps.pos);
			return true;
		}
	}

	private void addShotParticle(Vector2 p, float d) {
		for (int i = 0; i < 5; i++) {
			Particle pt;
			pt = particles.getInstanceForced();
			pt.set(Particle.LINE, p.x - 0.5f, p.y,
					-d + rand.nextSignedFloat(0.5f), 0.25f + rand.nextFloat(0.75f),
					1, 1.0f, 0.25f, 0.5f, 10);
			pt = particles.getInstanceForced();
			pt.set(Particle.LINE, p.x + 0.5f, p.y,
					-d + rand.nextSignedFloat(0.5f), 0.25f + rand.nextFloat(0.75f),
					1, 1.0f, 0.25f, 0.5f, 10);
		}
	}

	public void addVelocity(PlayerState ps, Vector2 v, Vector2 o) {
		Vector2 rv = v.getElement(o, 0.05f, 0.25f);
		rv.opMulAssign(5);
		ps.vel.opAddAssign(rv);
		float d = (float)Math.atan2(rv.x, -rv.y);
		float sp = rv.vctSize();
		for (int i = 0; i < 36; i++) {
			Particle pt;
			pt = particles.getInstanceForced();
			float r, g, b;
			r = 0.5f + rand.nextFloat(0.5f);
			g = 0.3f + rand.nextFloat(0.3f);
			b = 0.8f + rand.nextFloat(0.2f);
			pt.set(Particle.LINE, ps.pos.x, ps.pos.y,
					d + rand.nextSignedFloat(0.3f), sp * (1 + rand.nextFloat(2)),
					1, r, g, b, 30 + rand.nextInt(30));
		}
		Sound.playSe("flick.wav");
	}

	public void destroyed(PlayerState ps) {
//		with (ps) {
		{
			if (!ps.isActive())
				return;
			ps.destroyed();
			tractorBeam.clear();
			gameState.destroyedPlayer();
			float r, g, b;
			r = 0.5f + rand.nextFloat(0.5f);
			g = 0.3f + rand.nextFloat(0.3f);
			b = 0.8f + rand.nextFloat(0.2f);
			for (int i = 0; i < 100; i++) {
				Particle p = particles.getInstanceForced();
				p.set(Particle.QUAD, ps.pos.x, ps.pos.y, rand.nextFloat((float)Math.PI * 2), 0.01f + rand.nextFloat(1.0f),
						1 + rand.nextFloat(4), r, g, b, 10 + rand.nextInt(200));
			}
			r = 0.5f + rand.nextFloat(0.5f);
			g = 0.3f + rand.nextFloat(0.3f);
			b = 0.8f + rand.nextFloat(0.2f);
			for (int i = 0; i < 30; i++) {
				Particle p = particles.getInstanceForced();
				p.set(Particle.TRIANGLE, ps.pos.x, ps.pos.y, rand.nextFloat((float)Math.PI * 2), 0.03f + rand.nextFloat(0.3f),
						3, r, g, b, 50 + rand.nextInt(150));
			}
			r = 0.5f + rand.nextFloat(0.5f);
			g = 0.3f + rand.nextFloat(0.3f);
			b = 0.8f + rand.nextFloat(0.2f);
			for (int i = 0; i < 300; i++) {
				Particle p = particles.getInstanceForced();
				p.set(Particle.LINE, ps.pos.x, ps.pos.y, rand.nextFloat((float)Math.PI * 2), 0.07f + rand.nextFloat(0.7f),
						1, r, g, b, 100 + rand.nextInt(100));
			}
			Sound.playSe("player_explosion.wav");
		}
	}

	public void addScore(int sc) {
		gameState.addScore(sc);
	}

	public void addMultiplier(float mp) {
		gameState.addMultiplier(mp);
	}

	public float multiplier() {
		return gameState.multiplier();
	}

	public void draw(GL gl, PlayerState ps) {
//		with (ps) {
		{
			shots.draw(gl);
			capturedEnemiesShots.draw(gl);
			tractorBeam.draw(gl);
			if (!ps.isActive())
				return;
			Vector3 p = field.calcCircularPos(ps.pos);
			float cd = field.calcCircularDeg(ps.pos.x);
			if (ps.hasShape())
				shape.draw(gl, p, cd, ps.deg);
			int c = ps.colorCnt % 60;
			float a;
			if (c < 30)
				a = (float) c / 30;
			else
				a = 1 - (float) (c - 30) / 30;
			Screen.setColor(gl, a, a, a);
			lineShape.draw(gl, p, cd, ps.deg);
		}
	}
}

class ShotPool extends ActorPool<Shot> {
	public ShotPool(ActorCreator<Shot> c) {
		super(c);
	}

	public void checkParent() {
		for (Shot a: actors)
		if (a.exists())
			if (!a.spec.checkParent(a.state))
				a.remove();
	}
}

class Shot extends Token<ShotState, ShotSpec> {
	public void init(Object[] args) {
		super.init(args);
		state = new ShotState();
	}
	public void setParent(Shot s) {
		spec.setParent(state, s);
	}
}

class ShotState extends TokenState {
	Shot parent;
	int cnt;

	public void clear() {
		parent = null;
		cnt = 0;
		super.clear();
	}
}

class ShotSpec extends TokenSpec<ShotState> {
	EnemyPool enemies;
	BulletPool bullets;
	PlayerState playerState;
	GameState gameState;

	public ShotSpec(GL gl, Field field, EnemyPool enemies, BulletPool bullets, GameState gameState) {
		this.field = field;
		this.enemies = enemies;
		this.bullets = bullets;
		this.gameState = gameState;
		shape = new ShotShape(gl);
	}

	public void setPlayerState(PlayerState ps) {
		playerState = ps;
	}

	public void close(GL gl) {
		((ShotShape) shape).close(gl);
	}

	public void set(ShotState ss) {
		ss.parent = null;
		ss.cnt = 0;
	}

	public void setParent(ShotState ss, Shot s) {
		ss.parent = s;
	}

	public boolean move(ShotState ss) {
//		with (ss) {
		{
			if (ss.parent != null)
				if (ss.parent.exists() == false)
					return false;
			ss.stepForward();
			ss.pos.x = field.normalizeX(ss.pos.x);
			if (!field.containsOuterY(ss.pos.y)) {
				return false;
			}
			if (enemies.checkShotHit(ss.pos, ss.deg, 2)) {
				if (ss.parent != null)
					ss.parent.remove();
				gameState.countShotHit();
				return false;
			}
			ss.cnt++;
			return true;
		}
	}

	public boolean checkParent(ShotState ss) {
		if (ss.parent != null)
			if (ss.parent.exists() == false)
				return false;
		return true;
	}
}

class TractorBeam {
	static final float MAX_LENGTH = 10;
	static final float WIDTH = 1.5f;
	static final float SHAPE_INTERVAL_TIME = 10;
	static final float SHAPE_INTERVAL_LENGTH = 0.5f;
	Field field;
	PlayerState playerState;
	ArrayList<TractorBeamShape> shapes;
	float length = 0;
	int cnt;
	boolean isExtending;

//	invariant {
//		assert(length <>= 0);
//	}

	public TractorBeam(GL gl, Field field, PlayerState playerState) {
		this.field = field;
		this.playerState = playerState;
		shapes = new ArrayList<TractorBeamShape>();
		shapes.add(new TractorBeamShapeRed(gl));
		shapes.add(new TractorBeamShapeBlue(gl));
		shapes.add(new TractorBeamShapePurple(gl));
		clear();
	}

	public void clear() {
		length = 0;
		cnt = 0;
		isExtending = false;
	}

	public void move() {
		if (length <= 0)
			return;
		cnt++;
		if (cnt % 12 == 0 && isExtending)
			Sound.playSe("tractor.wav");
	}

	public void extendLength() {
		length += (MAX_LENGTH - length) * 0.05f;
		isExtending = true;
	}

	public void reduceLength() {
		length += (0 - length) * 0.1f;
		if (length < 0.1f)
			length = 0;
		isExtending = false;
	}

	public boolean contains(Vector2 p) {
		if (length <= 0)
			return false;
		return (p.x > playerState.pos.x - WIDTH / 2 &&
				p.x < playerState.pos.x + WIDTH / 2 &&
				p.y > playerState.pos.y && p.y < playerState.pos.y + length + WIDTH);
	}

	public void draw(GL gl) {
		if (length <= 0)
			return;
		float y = SHAPE_INTERVAL_LENGTH -
		(cnt % SHAPE_INTERVAL_TIME) * SHAPE_INTERVAL_LENGTH / SHAPE_INTERVAL_TIME;
		int c = (int) (cnt / SHAPE_INTERVAL_TIME);
		for (;;) {
			if (y > length)
				break;
			gl.glPushMatrix();
			Vector3 p = field.calcCircularPos(playerState.pos.x, playerState.pos.y + y);
			Screen.glTranslate(gl, p);
			float s = y;
			if (s > 1)
				s = 1;
			gl.glScalef(s, s, s);
			shapes.get(c % 3).draw(gl);
			c++;
			gl.glPopMatrix();
			y += SHAPE_INTERVAL_LENGTH;
		}
	}
}
