/*
 * $Id: shape.d,v 1.2 2006/11/23 02:29:43 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import javax.media.opengl.GL;

import abagames.util.Vector2;
import abagames.util.Vector3;
import abagames.util.sdl.DisplayList;
import abagames.util.sdl.Screen3D;

//private import std.math;
//private import opengl;
//private import abagames.util.vector;
//private import abagames.util.sdl.displaylist;
//private import abagames.ttn.screen;
//private import abagames.ttn.field;

/**
 * 3D shapes of a player, enemies, particles, etc.
 */
public interface Shape {
	public void draw(GL gl, Vector3 pos, float cd, float deg);
}

abstract class DisplayListShape implements Shape {
	protected DisplayList displayList;

	public DisplayListShape(GL gl) {
		displayList = new DisplayList(gl, 1);
		displayList.beginNewList(gl);
		drawList(gl);
		displayList.endNewList(gl);
	}

	protected abstract void drawList(GL gl);

	public void draw(GL gl) {
		drawList(gl);
	}

	public void draw(GL gl, Vector3 pos, float cd, float deg) {
		gl.glPushMatrix();
		Screen3D.glTranslate(gl, pos);
		gl.glRotatef(cd * 180 / (float)Math.PI, 0, 1, 0);
		Screen3D.glRotate(gl, deg);
		displayList.call(gl);
		gl.glPopMatrix();
	}

	public void close(GL gl) {
		displayList.close(gl);
	}
}

class PyramidShape {
	public static void draw(GL gl) {
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3f(0, 0, 0);
		gl.glVertex3f(1, 1, 1);
		gl.glVertex3f(1, 1, -1);
		gl.glVertex3f(-1, 1, -1);
		gl.glVertex3f(-1, 1, 1);
		gl.glVertex3f(1, 1, 1);
		gl.glEnd();
		Screen3D.setColor(gl, 0.1f, 0.1f, 0.1f, 0.5f);
		gl.glBegin(GL.GL_LINE_STRIP);
		gl.glVertex3f(0, 0, 0);
		gl.glVertex3f(1, 1, 1);
		gl.glVertex3f(1, 1, -1);
		gl.glVertex3f(0, 0, 0);
		gl.glVertex3f(-1, 1, -1);
		gl.glVertex3f(-1, 1, 1);
		gl.glVertex3f(0, 0, 0);
		gl.glEnd();
		gl.glBegin(GL.GL_LINES);
		gl.glVertex3f(1, 1, 1);
		gl.glVertex3f(-1, 1, 1);
		gl.glVertex3f(1, 1, -1);
		gl.glVertex3f(-1, 1, -1);
		gl.glEnd();
	}

	public static void drawShadow(GL gl, float r, float g, float b) {
		drawShadow(gl, r, g, b, false);
	}
	
	public static void drawShadow(GL gl, float r, float g, float b, boolean noAlpha) {
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		Screen3D.setColor(gl, r, g, b);
		gl.glVertex3f(0, 0, 0);
		if (!noAlpha)
			Screen3D.setColor(gl, r * 0.75f, g * 0.75f, b * 0.75f, 0.33f);
		else
			Screen3D.setColor(gl, r * 0.75f, g * 0.75f, b * 0.75f, 0.75f);
		gl.glVertex3f(1, 1, 1);
		gl.glVertex3f(1, 1, -1);
		gl.glVertex3f(-1, 1, -1);
		gl.glVertex3f(-1, 1, 1);
		gl.glVertex3f(1, 1, 1);
		gl.glEnd();
	}

	public static void drawPolygonShape(GL gl) {
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3f(0, 0, 0);
		gl.glVertex3f(1, 1, 1);
		gl.glVertex3f(1, 1, -1);
		gl.glVertex3f(-1, 1, -1);
		gl.glVertex3f(-1, 1, 1);
		gl.glVertex3f(1, 1, 1);
		gl.glEnd();
	}

	public static void drawLineShape(GL gl) {
		gl.glBegin(GL.GL_LINE_STRIP);
		gl.glVertex3f(0, 0, 0);
		gl.glVertex3f(1, 1, 1);
		gl.glVertex3f(1, 1, -1);
		gl.glVertex3f(0, 0, 0);
		gl.glVertex3f(-1, 1, -1);
		gl.glVertex3f(-1, 1, 1);
		gl.glVertex3f(0, 0, 0);
		gl.glEnd();
		gl.glBegin(GL.GL_LINES);
		gl.glVertex3f(1, 1, 1);
		gl.glVertex3f(-1, 1, 1);
		gl.glVertex3f(1, 1, -1);
		gl.glVertex3f(-1, 1, -1);
		gl.glEnd();
	}
}

class PlayerShape extends DisplayListShape {
	public PlayerShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(0, -0.6f, 0);
		gl.glScalef(0.4f, 1.3f, 0.4f);
		PyramidShape.drawShadow(gl, 1, 0.5f, 0.5f, true);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(0.5f, -0.2f, 0);
		gl.glScalef(0.3f, 0.9f, 0.3f);
		PyramidShape.drawShadow(gl, 1, 1, 1, true);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(-0.5f, -0.2f, 0);
		gl.glScalef(0.3f, 0.9f, 0.3f);
		PyramidShape.drawShadow(gl, 1, 1, 1, true);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 1, 0.5f, 0.5f);
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(0, -0.6f, 0);
		gl.glScalef(0.3f, 1.2f, 0.3f);
		PyramidShape.drawPolygonShape(gl);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 1, 1, 1);
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(0.5f, -0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.drawPolygonShape(gl);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 1, 1, 1);
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(-0.5f, -0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.drawPolygonShape(gl);
		gl.glPopMatrix();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
	}
}

class PlayerLineShape extends DisplayListShape {
	public PlayerLineShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(0, -0.6f, 0);
		gl.glScalef(0.3f, 1.2f, 0.3f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(0.5f, -0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(-0.5f, -0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
	}
}

class ShotShape extends DisplayListShape {
	public ShotShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(0.5f, -0.5f, 0);
		gl.glScalef(0.1f, 1.0f, 0.1f);
		Screen3D.setColor(gl, 0.4f, 0.2f, 0.8f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(180, 0, 0, 1);
		gl.glTranslatef(-0.5f, -0.5f, 0);
		gl.glScalef(0.1f, 1.0f, 0.1f);
		Screen3D.setColor(gl, 0.4f, 0.2f, 0.8f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
	}
}

abstract class TractorBeamShape extends DisplayListShape {
	public TractorBeamShape(GL gl) {
		super(gl);
	}
	protected void drawTractorBeam(GL gl, float r, float g, float b) {
		Screen3D.setColor(gl, r, g, b, 0.5f);
		gl.glBegin(GL.GL_QUADS);
		gl.glVertex3f(-1, 0, -1);
		gl.glVertex3f(1, 0, -1);
		gl.glVertex3f(1, 0, 1);
		gl.glVertex3f(-1, 0, 1);
		gl.glEnd();
		Screen3D.setColor(gl, r, g, b);
		gl.glBegin(GL.GL_LINE_LOOP);
		gl.glVertex3f(-1, 0, -1);
		gl.glVertex3f(1, 0, -1);
		gl.glVertex3f(1, 0, 1);
		gl.glVertex3f(-1, 0, 1);
		gl.glEnd();
	}
}

class TractorBeamShapeRed extends TractorBeamShape {
	public TractorBeamShapeRed(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		drawTractorBeam(gl, 0.5f, 0.2f, 0.2f);
	}
}

class TractorBeamShapeBlue extends TractorBeamShape {
	public TractorBeamShapeBlue(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		drawTractorBeam(gl, 0.2f, 0.2f, 0.5f);
	}
}

class TractorBeamShapePurple extends TractorBeamShape {
	public TractorBeamShapePurple(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		drawTractorBeam(gl, 0.5f, 0.2f, 0.5f);
	}
}

abstract class BulletShapeBase extends DisplayListShape {
	public BulletShapeBase(GL gl) {
		super(gl);
	}	
	public void draw(GL gl, Vector3 pos, float cd, float deg, float rd) {
		gl.glPushMatrix();
		Screen3D.glTranslate(gl, pos);
		gl.glRotatef(cd * 180 / (float)Math.PI, 0, 1, 0);
		Screen3D.glRotate(gl, deg);
		gl.glRotatef(rd, 0, 1, 0);
		displayList.call(gl);
		gl.glPopMatrix();
	}
}

class BulletShape extends BulletShapeBase {
	public BulletShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		Screen3D.setColor(gl, 0, 0, 0);
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3f(0, 0.5f, 0);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glEnd();
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glEnd();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
		gl.glScalef(1.2f, 1.2f, 1.2f);
		Screen3D.setColor(gl, 0.1f, 0.3f, 0.3f);
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3f(0, 0.5f, 0);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glEnd();
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glEnd();
	}
}

class BulletLineShape extends BulletShapeBase {
	public BulletLineShape(GL gl) {
		super(gl);
	}	
	protected void drawList(GL gl) {
		gl.glScalef(1.2f, 1.2f, 1.2f);
		gl.glBegin(GL.GL_LINES);
		gl.glVertex3f(0, 0.5f, 0);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, 0.5f, 0);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, 0.5f, 0);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glEnd();
		gl.glBegin(GL.GL_LINE_LOOP);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glEnd();
	}
}

class MiddleBulletShape extends BulletShapeBase {
	public MiddleBulletShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glScalef(1.1f, 1.0f, 1.1f);
		Screen3D.setColor(gl, 0, 0, 0);
		gl.glBegin(GL.GL_QUADS);
		gl.glVertex3f(-0.17f, 0.3f, -0.1f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glVertex3f(0, 0.3f, 0.2f);
		gl.glVertex3f(0, 0.3f, 0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(-0.17f, 0.3f, -0.1f);
		gl.glEnd();
		gl.glBegin(GL.GL_TRIANGLES);
		gl.glVertex3f(-0.17f, -0.3f, -0.1f);
		gl.glVertex3f(0.17f, -0.3f, -0.1f);
		gl.glVertex3f(0, -0.3f, 0.2f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glEnd();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
		gl.glScalef(1.4f, 1.3f, 1.4f);
		Screen3D.setColor(gl, 0.1f, 0.2f, 0.3f);
		gl.glBegin(GL.GL_QUADS);
		gl.glVertex3f(-0.17f, 0.3f, -0.1f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glVertex3f(0, 0.3f, 0.2f);
		gl.glVertex3f(0, 0.3f, 0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(-0.17f, 0.3f, -0.1f);
		gl.glEnd();
		gl.glBegin(GL.GL_TRIANGLES);
		gl.glVertex3f(-0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0, 0.3f, 0.2f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glEnd();
	}
}

class MiddleBulletLineShape extends BulletShapeBase {
	public MiddleBulletLineShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glScalef(1.4f, 1.3f, 1.4f);
		gl.glBegin(GL.GL_LINES);
		gl.glVertex3f(-0.17f, 0.3f, -0.1f);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, 0.3f, 0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glEnd();
		gl.glBegin(GL.GL_LINE_LOOP);
		gl.glVertex3f(-0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0.17f, 0.3f, -0.1f);
		gl.glVertex3f(0, 0.3f, 0.2f);
		gl.glEnd();
		gl.glBegin(GL.GL_LINE_LOOP);
		gl.glVertex3f(-0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0.34f, -0.3f, -0.2f);
		gl.glVertex3f(0, -0.3f, 0.4f);
		gl.glEnd();
	}
}

abstract class EnemyShape extends DisplayListShape {
	public EnemyShape(GL gl) {
		super(gl);
	}
	public void draw(GL gl, Vector3 pos, float cd, float deg, float cnt, Vector2 size) {
		gl.glPushMatrix();
		Screen3D.glTranslate(gl, pos);
		gl.glRotatef(cd * 180 / (float)Math.PI, 0, 1, 0);
		Screen3D.glRotate(gl, deg);
		gl.glScalef(size.x, size.y, 1);
		gl.glRotatef(cnt * 3.0f, 0, 1, 0);
		displayList.call(gl);
		gl.glPopMatrix();
	}
}

class Enemy1Shape extends EnemyShape {
	public Enemy1Shape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.6f, 0);
		gl.glScalef(0.5f, 1.4f, 0.5f);
		PyramidShape.drawShadow(gl, 0.5f, 0.5f, 0.3f);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(120, 0, 0, 1);
		gl.glTranslatef(0.5f, -0.2f, 0);
		gl.glScalef(0.4f, 1.0f, 0.4f);
		PyramidShape.drawShadow(gl, 0.2f, 0.2f, 0.5f);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 0.2f, 0.2f, 0.5f);
		gl.glPushMatrix();
		gl.glRotatef(240, 0, 0, 1);
		gl.glTranslatef(-0.5f, -0.2f, 0);
		gl.glScalef(0.4f, 1.0f, 0.4f);
		PyramidShape.drawShadow(gl, 0.2f, 0.2f, 0.5f);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 1, 1, 0.6f);
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.6f, 0);
		gl.glScalef(0.3f, 1.2f, 0.3f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 0.5f, 0.5f, 1);
		gl.glPushMatrix();
		gl.glRotatef(120, 0, 0, 1);
		gl.glTranslatef(0.5f, -0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 0.5f, 0.5f, 1);
		gl.glPushMatrix();
		gl.glRotatef(240, 0, 0, 1);
		gl.glTranslatef(-0.5f, -0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
	}
}

class Enemy1TrailShape extends EnemyShape {
	public Enemy1TrailShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.6f, 0);
		gl.glScalef(0.3f, 1.2f, 0.3f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(120, 0, 0, 1);
		gl.glTranslatef(0.5f, -0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(240, 0, 0, 1);
		gl.glTranslatef(-0.5f, -0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
	}
}

class Enemy2Shape extends EnemyShape {
	public Enemy2Shape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.5f, 0);
		gl.glScalef(0.5f, 1.2f, 0.5f);
		PyramidShape.drawShadow(gl, 0.5f, 0.4f, 0.5f);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(60, 0, 0, 1);
		gl.glTranslatef(0.6f, -0.7f, 0);
		gl.glScalef(0.4f, 1.4f, 0.4f);
		PyramidShape.drawShadow(gl, 0.9f, 0.6f, 0.5f);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(300, 0, 0, 1);
		gl.glTranslatef(-0.6f, -0.7f, 0);
		gl.glScalef(0.4f, 1.4f, 0.4f);
		PyramidShape.drawShadow(gl, 0.9f, 0.6f, 0.5f);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 1, 0.9f, 1.0f);
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.5f, 0);
		gl.glScalef(0.3f, 1.0f, 0.3f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 0.9f, 0.6f, 0.5f);
		gl.glPushMatrix();
		gl.glRotatef(60, 0, 0, 1);
		gl.glTranslatef(0.6f, -0.7f, 0);
		gl.glScalef(0.2f, 1.2f, 0.2f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 0.9f, 0.6f, 0.5f);
		gl.glPushMatrix();
		gl.glRotatef(300, 0, 0, 1);
		gl.glTranslatef(-0.6f, -0.7f, 0);
		gl.glScalef(0.2f, 1.2f, 0.2f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
	}
}

class Enemy2TrailShape extends EnemyShape {
	public Enemy2TrailShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.5f, 0);
		gl.glScalef(0.3f, 1.0f, 0.3f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(60, 0, 0, 1);
		gl.glTranslatef(0.6f, -0.7f, 0);
		gl.glScalef(0.2f, 1.2f, 0.2f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(300, 0, 0, 1);
		gl.glTranslatef(-0.6f, -0.7f, 0);
		gl.glScalef(0.2f, 1.2f, 0.2f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
	}
}

class Enemy3Shape extends EnemyShape {
	public Enemy3Shape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.4f, 0);
		gl.glScalef(0.5f, 1.4f, 0.5f);
		PyramidShape.drawShadow(gl, 0.5f, 0.5f, 0.3f);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(150, 0, 0, 1);
		gl.glTranslatef(0.5f, 0.2f, 0);
		gl.glScalef(0.4f, 1.0f, 0.4f);
		PyramidShape.drawShadow(gl, 0.2f, 0.2f, 0.5f);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 0.2f, 0.2f, 0.5f);
		gl.glPushMatrix();
		gl.glRotatef(210, 0, 0, 1);
		gl.glTranslatef(-0.5f, 0.2f, 0);
		gl.glScalef(0.4f, 1.0f, 0.4f);
		PyramidShape.drawShadow(gl, 0.2f, 0.2f, 0.5f);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 1, 0.6f, 0.9f);
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.4f, 0);
		gl.glScalef(0.3f, 1.2f, 0.3f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 0.3f, 0.5f, 1);
		gl.glPushMatrix();
		gl.glRotatef(150, 0, 0, 1);
		gl.glTranslatef(0.5f, 0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		Screen3D.setColor(gl, 0.3f, 0.5f, 1);
		gl.glPushMatrix();
		gl.glRotatef(210, 0, 0, 1);
		gl.glTranslatef(-0.5f, 0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.draw(gl);
		gl.glPopMatrix();
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
	}
}

class Enemy3TrailShape extends EnemyShape {
	public Enemy3TrailShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glPushMatrix();
		gl.glTranslatef(0, -0.4f, 0);
		gl.glScalef(0.3f, 1.2f, 0.3f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(150, 0, 0, 1);
		gl.glTranslatef(0.5f, 0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
		gl.glPushMatrix();
		gl.glRotatef(210, 0, 0, 1);
		gl.glTranslatef(-0.5f, 0.2f, 0);
		gl.glScalef(0.2f, 0.8f, 0.2f);
		PyramidShape.drawLineShape(gl);
		gl.glPopMatrix();
	}
}

class TriangleParticleShape extends DisplayListShape {
	public TriangleParticleShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glBegin(GL.GL_LINE_LOOP);
		gl.glVertex3f(0, 0.5f, 0);
		gl.glVertex3f(0.4f, -0.3f, 0);
		gl.glVertex3f(-0.4f, -0.3f, 0);
		gl.glEnd();
	}
}

abstract class PillarShape extends DisplayListShape {
	public static final float TICKNESS = 4.0f;
	private static final float RADIUS_RATIO = 0.3f;

	public PillarShape(GL gl) {
		super(gl);
	}
	
	protected void drawPillar(GL gl, float r, float g, float b) {
		drawPillar(gl, r, g, b, false);
	}
	protected void drawPillar(GL gl, float r, float g, float b, boolean outside) {
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE_MINUS_SRC_ALPHA);
		gl.glBegin(GL.GL_QUADS);
		Screen3D.setColor(gl, r, g, b);
		for (int i = 0; i < 8; i++) {
			float d = (float)Math.PI * 2 * i / 8;
			gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
					TICKNESS,
					Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
			d += Math.PI * 2 / 8;
			gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
					TICKNESS,
					Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
			gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
					-TICKNESS,
					Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
			d -= Math.PI * 2 / 8;
			gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
					-TICKNESS,
					Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
		}
		gl.glEnd();
		if (!outside) {
			Screen3D.setColor(gl, r, g, b);
			gl.glBegin(GL.GL_TRIANGLES);
			for (int i = 0; i < 8; i++) {
				float d = (float)Math.PI * 2 * i / 8;
				gl.glVertex3f(0, TICKNESS, 0);
				gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
						TICKNESS,
						Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
				d += Math.PI * 2 / 8;
				gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
						TICKNESS,
						Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
				d -= Math.PI * 2 / 8;
				gl.glVertex3f(0, -TICKNESS, 0);
				gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
						-TICKNESS,
						Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
				d += Math.PI * 2 / 8;
				gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
						-TICKNESS,
						Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
			}
			gl.glEnd();
		}
		Screen3D.setColor(gl, 0.1f, 0.1f, 0.1f);
		for (int i = 0; i < 8; i++) {
			float d = (float)Math.PI * 2 * i / 8; 
			gl.glBegin(GL.GL_LINE_STRIP);
			gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
					TICKNESS,
					Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
			d += Math.PI * 2 / 8;
			gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
					TICKNESS,
					Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
			gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
					-TICKNESS,
					Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
			d -= Math.PI * 2 / 8;
			gl.glVertex3d(Math.sin(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO,
					-TICKNESS,
					Math.cos(d) * Field.CIRCLE_RADIUS * RADIUS_RATIO);
			gl.glEnd();
		}
		gl.glBlendFunc(GL.GL_SRC_ALPHA, GL.GL_ONE);
	}

	public void draw(GL gl, float y, float deg) {
		gl.glPushMatrix();
		gl.glTranslatef(0, y, 0);
		gl.glRotated(deg * 180 / Math.PI, 0, 1, 0);
		displayList.call(gl);
		gl.glPopMatrix();
	}
}

class Pillar1Shape extends PillarShape {
	public Pillar1Shape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glScalef(0.6f, 1.0f, 0.6f);
		drawPillar(gl, 0.5f, 0.4f, 0.4f);
	}
}

class Pillar2Shape extends PillarShape {
	public Pillar2Shape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glScalef(0.8f, 1.0f, 0.8f);
		drawPillar(gl, 0.6f, 0.3f, 0.3f);
	}
}

class Pillar3Shape extends PillarShape {
	public Pillar3Shape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		drawPillar(gl, 0.5f, 0.5f, 0.4f);
	}
}

class Pillar4Shape extends PillarShape {
	public Pillar4Shape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glScalef(1.1f, 1.0f, 1.1f);
		drawPillar(gl, 0.5f, 0.4f, 0.5f);
	}
}

class OutsidePillarShape extends PillarShape {
	public OutsidePillarShape(GL gl) {
		super(gl);
	}
	protected void drawList(GL gl) {
		gl.glScalef(7.0f, 3.0f, 7.0f);
		drawPillar(gl, 0.2f, 0.2f, 0.3f, true);
	}
}
