/*
 * $Id: stage.d,v 1.2 2006/11/23 02:29:44 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import java.util.ArrayList;

import javax.media.opengl.GL;

import abagames.util.Rand;

//private import std.math;
//private import abagames.util.rand;
//private import abagames.util.vector;
//private import abagames.ttn.field;
//private import abagames.ttn.token;
//private import abagames.ttn.enemy;
//private import abagames.ttn.bullet;
//private import abagames.ttn.player;
//private import abagames.ttn.particle;
//private import abagames.ttn.pillar;
//private import abagames.ttn.shape;
//private import abagames.ttn.sound;
//private import abagames.ttn.frame;
//private import abagames.ttn.letter;

/**
 * Enemy appearance pattern handler.
 */
public class Stage {
	static final int PHASE_RESULT_SHOW_CNT = 150;
	static final int PHASE_START_SHOW_CNT = 90;
	Field field;
	EnemyPool enemies;
	BulletPool bullets;
	Player player;
	ParticlePool particles;
	PillarPool pillars;
	GameState gameState;
	Rand rand;
	int appCnt;
	EnemySpec middleEnemySpec, smallEnemy1Spec, smallEnemy2Spec;
	EnemyShape enemy1Shape, enemy2Shape, enemy3Shape;
	EnemyShape enemy1TrailShape, enemy2TrailShape, enemy3TrailShape;
	BulletSpec bulletSpec, middleBulletSpec;
	BulletShapeBase bulletShape, bulletLineShape, middleBulletShape, middleBulletLineShape;
	PillarSpec pillarSpec;
	ArrayList<PillarShape> pillarShapes;
	PillarShape outsidePillarShape;
	int smallEnemyNum;
	int smallEnemyFormationNum;
	float rank;
	int phaseTime;
	boolean stageStarted, waitNextFormationPhase;
	int middleEnemyAppInterval;
	int _attackSmallEnemyNum;
	float goingDownBeforeStandByRatio;
	int appCntInterval;
	int formationIdx;
	int cnt;
	float rankTrg;
	int phaseNum;
	int shotFiredNum, shotHitNum;
	int shotFiredNumRsl, shotHitNumRsl;
	int shotFiredNumTotal, shotHitNumTotal;
	float hitRatio;
	int hitRatioBonus;

	public Stage(GL gl, Field field, EnemyPool enemies, BulletPool bullets, Player player,
			ParticlePool particles, PillarPool pillars, GameState gameState) {
		this.field = field;
		this.enemies = enemies;
		this.bullets = bullets;
		this.player = player;
		this.particles = particles;
		this.pillars = pillars;
		this.gameState = gameState;
		rand = new Rand();
		enemy1Shape = new Enemy1Shape(gl);
		enemy2Shape = new Enemy2Shape(gl);
		enemy3Shape = new Enemy3Shape(gl);
		enemy1TrailShape = new Enemy1TrailShape(gl);
		enemy2TrailShape = new Enemy2TrailShape(gl);
		enemy3TrailShape = new Enemy3TrailShape(gl);
		bulletShape = new BulletShape(gl);
		bulletLineShape = new BulletLineShape(gl);
		middleBulletShape = new MiddleBulletShape(gl);
		middleBulletLineShape = new MiddleBulletLineShape(gl);
		bulletSpec = new BulletSpec(field, player, enemies, particles,
				bulletShape, bulletLineShape);
		middleBulletSpec = new BulletSpec(field, player, enemies, particles,
				middleBulletShape, middleBulletLineShape);
		pillarSpec = new PillarSpec(field);
		pillarShapes = new ArrayList<PillarShape>();
		pillarShapes.add(new Pillar1Shape(gl));
		pillarShapes.add(new Pillar2Shape(gl));
		pillarShapes.add(new Pillar3Shape(gl));
		pillarShapes.add(new Pillar4Shape(gl));
		outsidePillarShape = new OutsidePillarShape(gl);
	}

	public void close(GL gl) {
		enemy1Shape.close(gl);
		enemy2Shape.close(gl);
		enemy3Shape.close(gl);
		enemy1TrailShape.close(gl);
		enemy2TrailShape.close(gl);
		enemy3TrailShape.close(gl);
		bulletShape.close(gl);
		bulletLineShape.close(gl);
		middleBulletShape.close(gl);
		middleBulletLineShape.close(gl);
		for (PillarShape ps: pillarShapes)
			ps.close(gl);
		outsidePillarShape.close(gl);
	}

	public void start(long randSeed) {
		clear();
		rand.setSeed(randSeed);
		EnemySpec.rand.setSeed(randSeed);
		TurretSpec.rand.setSeed(randSeed);
		PlayerSpec.rand.setSeed(randSeed);
		ParticleSpec.rand.setSeed(randSeed);
		Sound.rand.setSeed(randSeed);
		rank = rankTrg = 0;
		phaseNum = 0;
		cnt = 0;
		shotFiredNumTotal = shotHitNumTotal = 0;
		for (int i = 0; i < 1000; i++) {
			cnt++;
			moveOutsidePillars();
			pillars.move();
		}
		startPhase();
	}

	private void clear() {
		smallEnemyNum = smallEnemyFormationNum = 0;
		rank = 0;
		phaseTime = 0;
		stageStarted = waitNextFormationPhase = false;
		middleEnemyAppInterval = 0;
		_attackSmallEnemyNum = 0;
		goingDownBeforeStandByRatio = 0;
		appCntInterval = 0;
		formationIdx = 0;
		cnt = 0;
		rankTrg = 0;
		phaseNum = 0;
		shotFiredNum = shotHitNum = 0;
		shotFiredNumRsl = shotHitNumRsl = 0;
		shotFiredNumTotal = shotHitNumTotal = 0;
		hitRatio = 0;
		hitRatioBonus = 0;
	}

	private void startPhase() {
		phaseTime = 0;
		phaseNum++;
		if (phaseNum > 1)
			calcHitRatioBonus();
		if (phaseNum % 10 == 0)
			Sound.fadeBgm();
		setEnemySpecs();
		initPillars();
	}

	private void calcHitRatioBonus() {
		shotFiredNumRsl = shotFiredNum;
		shotHitNumRsl = shotHitNum;
		shotFiredNum = shotHitNum = 0;
		if (shotFiredNumRsl <= 0) {
			hitRatio = 0;
		} else {
			hitRatio = (float) shotHitNumRsl / shotFiredNumRsl;
		}
		float r = (float) ((int) (hitRatio * 100)) / 100;
		if (r > 1)
			r = 1;
		hitRatioBonus = (int) (10000.0f * r * r * r * r);
		gameState.addScore(hitRatioBonus, true);
	}

	private void setEnemySpecs() {
		rankTrg += 1.25f;
		rank += (rankTrg - rank) * 0.33f;
		if (phaseNum % 5 == 0)
			rank *= 0.1f;
		long rs = phaseNum;
		rand.setSeed(rs);
		EnemySpec.rand.setSeed(rs);
		TurretSpec.rand.setSeed(rs);
		int en = 32 + (int) ((50 + rand.nextInt(10)) * Math.sqrt(rank) * 0.33f);
		smallEnemyNum = 7 + rand.nextInt(4);
		smallEnemyFormationNum = (int) (en / smallEnemyNum) + 1;
		middleEnemyAppInterval = 5 + rand.nextInt(2);
		middleEnemySpec = new MiddleEnemySpec
		(field, bullets, player, particles, enemies,
				this, enemy3Shape, enemy3TrailShape, middleBulletSpec);
		middleEnemySpec.setRank(rank * 0.15f);
		smallEnemy1Spec = new SE1Spec
		(field, bullets, player, particles, enemies,
				this, enemy1Shape, enemy1TrailShape, bulletSpec);
		((SE1Spec) smallEnemy1Spec).setRank(rank * 0.22f);
		smallEnemy2Spec = new SE2Spec
		(field, bullets, player, particles, enemies,
				this, enemy2Shape, enemy2TrailShape, bulletSpec);
		((SE2Spec) smallEnemy2Spec).setRank(rank * 0.22f);
		_attackSmallEnemyNum = (int) Math.sqrt(rank + 2);
		goingDownBeforeStandByRatio = 0;
		if (rand.nextFloat(rank + 1) > 2)
			goingDownBeforeStandByRatio = rand.nextFloat(0.2f) + 0.1f;
		appCntInterval = 48 + rand.nextSignedInt(10);
		appCntInterval *= (0.5f + 0.5f / Math.sqrt(rank));
		appCnt = 0;
		formationIdx = 0;
		stageStarted = false;
		waitNextFormationPhase = false;
	}

	private void initPillars() {
		pillars.setEnd();
		Pillar pp = null;
		int pln = 0;
		int pn = phaseNum;
//		int[] pshapes;
		ArrayList<Integer> pshapes = new ArrayList<Integer>();
		for (;;) {
			if (pn <= 0)
				break;
			if (pn >= 20) {
				pshapes.add(3);
				pn -= 20;
			} else if (pn >= 10) {
				pshapes.add(2);
				pn -= 10;
			} else if (pn >= 5) {
				pshapes.add(1);
				pn -= 5;
			} else {
				pshapes.add(0);
				pn--;
			}
			pln++;
		}
		float maxY = -15 + pln * 8;
		for (int i = 0; i < pln; i++) {
			Pillar p = pillars.getInstance();
			if (p == null)
				break;
			p.set(pillarSpec, -80 - i * 10, maxY, pp, pillarShapes.get(pshapes.get(i)), (pln - i) * 0.03f);
			pp = p;
		}
	}

	public void move() {
		if (appCnt <= 0) {
			if ((formationIdx % middleEnemyAppInterval) == middleEnemyAppInterval - 1) {
				Enemy me = enemies.getInstance();
				if (me == null)
					return;
				float x = rand.nextFloat(field.circularDistance());
				x = field.normalizeX(x);
				float sp = 0.1f + rand.nextSignedFloat(0.01f);
				float av = sp * 0.4f + rand.nextSignedFloat(0.005f);
				float er = rand.nextFloat(0.5f);
				float ed = rand.nextFloat((float)Math.PI * 2);
				me.set(middleEnemySpec, x, field.size().y * Field.PIT_SIZE_Y_RATIO, (float)Math.PI, sp);
				me.setMiddleEnemyState(sp, av, er, ed);
			}
			float x = rand.nextFloat(field.circularDistance());
			x = field.normalizeX(x);
			float sp = 0.15f + rand.nextSignedFloat(0.01f);
			float av = sp * 0.5f + rand.nextSignedFloat(0.005f);
			float dst = sp * 6.0f;
			float er = rand.nextFloat(0.8f);
			float ed = rand.nextFloat((float)Math.PI * 2);
			Enemy fe = null;
			for (int i = 0; i < smallEnemyNum; i++) {
				Enemy e = enemies.getInstance();
				if (e == null)
					break;
				SmallEnemySpec ses = null;
				int appPattern = formationIdx % 2;
				switch (formationIdx % 3) {
				case 0:
				case 1:
					ses = (SmallEnemySpec) smallEnemy1Spec;
					break;
				case 2:
					ses = (SmallEnemySpec) smallEnemy2Spec;
					break;
				}
				e.set(ses, x, field.size().y * Field.PIT_SIZE_Y_RATIO + i * dst, (float)Math.PI, sp);
				boolean gd = false;
				if (rand.nextFloat(1) < goingDownBeforeStandByRatio)
					gd = true;
				if (fe == null) {
					e.setSmallEnemyState(sp, av, (int) (i * (dst / sp)), appPattern,
							er, ed, gd);
					fe = e;
				} else {
					e.setSmallEnemyState(sp, av, (int) (i * (dst / sp)), appPattern,
							er, ed, gd, fe);
				}
			}
			smallEnemyFormationNum--;
			formationIdx++;
			if (smallEnemyFormationNum <= 0) {
				stageStarted = true;
				appCnt = 9999999;
			} else {
				appCnt += appCntInterval * (1 - 1 / (enemies.num() + 1));
			}
		}
		appCnt--;
		phaseTime++;
		if (phaseNum >= 10 && phaseNum % 10 == 0 && phaseTime == 120 &&
				gameState.isInGameAndNotGameOver())
			Sound.nextBgm();
		cnt++;
		moveOutsidePillars();
		if (enemies.numInScreen() > 0)
			gameState.mulMultiplier(0.999f);
		if (stageStarted && enemies.num() <= 0)
			startPhase();
	}

	public void moveOutsidePillars() {
		if (cnt % 120 == 0) {
			Pillar p = pillars.getInstance();
			if (p != null)
				p.set(pillarSpec, 180, 0, null, outsidePillarShape, (((int) cnt / 120) % 2 * 2 - 1) * 0.003f, true);
		}
	}

	public void countShotFired() {
		if (phaseTime >= PHASE_RESULT_SHOW_CNT) {
			shotFiredNum++;
			shotFiredNumTotal++;
		}
	}

	public void countShotHit() {
		if (phaseTime >= PHASE_RESULT_SHOW_CNT) {
			shotHitNum++;
			shotHitNumTotal++;
		}
	}

	public void draw(GL gl) {
		if (phaseTime < PHASE_RESULT_SHOW_CNT && phaseNum > 1) {
			Letter.drawString(gl, "SHOTS FIRED", 152, 250, 6, Letter.TO_RIGHT,
					false, 0, 1, 1, 0.33f);
			Letter.drawNum(gl, shotFiredNumRsl, 480, 250, 6);
			Letter.drawString(gl, "NUMBER OF HITS", 152, 280, 6, Letter.TO_RIGHT,
					false, 0, 1, 1, 0.33f);
			Letter.drawNum(gl, shotHitNumRsl, 480, 280, 6);
			Letter.drawString(gl, "HIT-MISS RATIO", 152, 310, 6);
			Letter.drawNum(gl, (int) (hitRatio * 10000), 480, 310, 6, 3, -1, 2);
			Letter.drawString(gl, "BONUS", 200, 350, 6, Letter.TO_RIGHT,
					false, 0, 1, 0.33f, 0.33f);
			Letter.drawNum(gl, hitRatioBonus, 440, 350, 6);
		} else if (phaseTime < PHASE_RESULT_SHOW_CNT + PHASE_START_SHOW_CNT) {
			Letter.drawNum(gl, phaseNum, 392, 200, 10);
			Letter.drawString(gl, "PHASE", 232, 200, 10);
		}
	}

	public void drawPhaseNum(GL gl) {
		Letter.drawNum(gl, phaseNum, 622, 448, 10);
	}

	public void drawGameover(GL gl) {
		float hr = 0;
		if (shotFiredNumTotal > 0)
			hr = (float) shotHitNumTotal / shotFiredNumTotal;
		Letter.drawString(gl, "SHOTS FIRED", 152, 250, 6, Letter.TO_RIGHT,
				false, 0, 1, 1, 0.33f);
		Letter.drawNum(gl, shotFiredNumTotal, 480, 250, 6);
		Letter.drawString(gl, "NUMBER OF HITS", 152, 280, 6, Letter.TO_RIGHT,
				false, 0, 1, 1, 0.33f);
		Letter.drawNum(gl, shotHitNumTotal, 480, 280, 6);
		Letter.drawString(gl, "HIT-MISS RATIO", 152, 310, 6);
		Letter.drawNum(gl, (int) (hr * 10000), 480, 310, 6, 3, -1, 2);
	}

	public int attackSmallEnemyNum() {
		return _attackSmallEnemyNum;
	}
}
