/*
 * $Id: title.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;

import javax.media.opengl.GL;
import javax.media.opengl.GLException;

import com.sun.opengl.util.texture.Texture;
import com.sun.opengl.util.texture.TextureCoords;
import com.sun.opengl.util.texture.TextureIO;

import abagames.util.*;
import abagames.util.sdl.*;
import abagames.util.sdl.Pad.PadState;
import abagames.util.sdl.Pad.RecordablePad;

/**
 * Title screen.
 */
public class Title {
	Preference preference;
	RecordablePad pad;
	Frame frame;
	int cnt;
	boolean aPressed;
	Texture titleTexture;
	Vector2 titlePos;
	float titleSize;

	public Title(Preference preference, Pad pad, Frame frame) {
		this.preference = preference;
		this.pad = (RecordablePad) pad;
		this.frame = frame;
		titlePos = new Vector2();
	}

	public void init() {
		InputStream is;
		try {
//			is = new FileInputStream("resource/images" + "title.bmp");
			is = getClass().getResourceAsStream("/resource/images/" + "title.png");
//			is.read();
			titleTexture = TextureIO.newTexture(is, false, "png");
		} catch (FileNotFoundException e) {
			e.printStackTrace();
		} catch (GLException e) {
			e.printStackTrace();
		} catch (IOException e) {
			e.printStackTrace();
		}
	}

	public void close() {
//		titleTexture.close();
		if (titleTexture != null)
			titleTexture.dispose();
	}

	public void start() {
		cnt = 0;
		aPressed = true;
		titlePos.x = 150;
		titlePos.y = 150;
		titleSize = 1.0f;
	}

	public void move() {
		PadState input;
		input = pad.getState(false);
		if ((input.button & PadState.A) != 0) {
			if (!aPressed) {
				aPressed = true;
				frame.startInGame();
			}
		} else {
			aPressed = false;
		}
		if (cnt > 180 && cnt < 235)
			titlePos.y -= 2;
		if (cnt > 600 && cnt < 675) {
			titlePos.x -= 2;
			titlePos.y++;
			titleSize -= 0.007f;
		}
		cnt++;
	}

	public void draw(GL gl) {
		Screen.setColor(gl, 1, 1, 1);
//		gl.glEnable(GL.GL_TEXTURE_2D);
if (titleTexture != null) {
	titleTexture.enable();
	titleTexture.bind();
}
drawBoard(gl, titlePos.x, titlePos.y, 280 * titleSize, 64 * titleSize, titleTexture.getImageTexCoords());
if (titleTexture != null) {
	titleTexture.disable();
}
//gl.glDisable(GL.GL_TEXTURE_2D);
if ((cnt % 120) < 60) {
	float x = 175, sz = 6;
	if (cnt >= 600) {
		int c = cnt - 600;
		if (c > 75)
			c = 75;
		x += c * 4.33f;
		sz -= c * 0.045f;
	}
	Letter.drawString(gl, "PUSH SHOT BUTTON TO START", x, 440, sz);
}
if (cnt >= 240) {
	drawRanking(gl);
}
	}

	private void drawBoard(GL gl, float x, float y, float w, float h, TextureCoords c) {
		gl.glBegin(GL.GL_TRIANGLE_FAN);
		gl.glTexCoord2f(c.left(), c.top());
		gl.glVertex3f(x, y, 0);
		gl.glTexCoord2f(c.right(), c.top());
		gl.glVertex3f(x + w, y, 0);
		gl.glTexCoord2f(c.right(), c.bottom());
		gl.glVertex3f(x + w, y + h, 0);
		gl.glTexCoord2f(c.left(), c.bottom());
		gl.glVertex3f(x, y + h, 0);
		gl.glEnd();
	}

	private void drawRanking(GL gl) {
		int rn = (cnt - 240) / 30;
		if (rn > Preference.RANKING_NUM)
			rn = Preference.RANKING_NUM;
		float y = 140;
		for (int i = 0; i < rn; i++) {
			if (cnt < 600) {
				String rstr;
				switch (i) {
				case 0:
					rstr = "1ST";
					break;
				case 1:
					rstr = "2ND";
					break;
				case 2:
					rstr = "3RD";
					break;
				default:
					rstr = (i + 1) + "TH";
				break;
				}
				if (i < 9)
					Letter.drawString(gl, rstr, 180, y, 7);
				else
					Letter.drawString(gl, rstr, 166, y, 7);
			}
			float sx = 450, sy = y, sz = 6;
			if (cnt >= 600) {
				int c = cnt - 600;
				if (c > 75)
					c = 75;
				sx += (int) (c * 2.35f);
				sz -= c * 0.03f;
			}
			Letter.drawNum(gl, preference.highScore()[i], sx, sy, sz);
			y += 24;
		}
	}
}
