/*
 * $Id: token.d,v 1.2 2006/11/23 02:29:44 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.ttn;

import javax.media.opengl.GL;

import abagames.util.Actor;
import abagames.util.Vector2;
import abagames.util.Vector3;

/**
 * Tokens of a player, enemies, bullets, particles, etc.
 * Handling these states (position, direction, speed, etc.) and
 *  specs (maneuver, method of attack, etc.).
 */
public class Token<ST extends TokenState, SP extends TokenSpec<ST>> extends Actor {
	protected ST state;
	protected SP spec;

	public void init(Object[] args) {
//		state = new ST();//TODO
	}

	public void set(SP spec, Vector2 pos, float deg, float speed) {
		set(spec, pos.x, pos.y, deg, speed);
	}

	public void set(SP spec, float x, float y, float deg, float speed) {
		this.spec = spec;
		set(x, y, deg, speed);
	}

	public void set(float x, float y, float deg, float speed) {
		state.clear();
		state.pos.x = x;
		state.pos.y = y;
		state.deg = deg;
		state.speed = speed;
		spec.set(state);
		_exists = true;
	}

	public void move() {
		if (!spec.move(state))
			remove();
	}

	public void remove() {
		_exists = false;
		spec.removed(state);
	}

	public void draw(GL gl) {
		spec.draw(gl, state);
	}

	public Vector2 pos() {
		return state.pos;
	}
}



/**
 * Holding a state of a token.
 */
class TokenState {
	boolean isInitialized = false;
	Vector2 pos;
	float deg;
	float speed;

//	invariant {
//		if (isInitialized) {
//			assert(pos.x <>= 0);
//			assert(pos.y <>= 0);
//			assert(deg <>= 0);
//			assert(speed <>= 0);
//		}
//	}

	public TokenState() {
		pos = new Vector2();
	}

	public void clear() {
		pos.x = pos.y = 0;
		deg = 0;
		speed = 0;
		isInitialized = true;
	}

	public void stepForward() {
		pos.x -= Math.sin(deg) * speed;
		pos.y += Math.cos(deg) * speed;
	}
}

/**
 * Base class of a token's specification.
 */
class TokenSpec<T extends TokenState> {
	protected Field field;
	protected Shape shape;

	public void set(T state) {}
	public void removed(T state) {}
	
	public boolean move(T state) {
		return true;
	}
	
	public void draw(GL gl, T state) {
		Vector3 p = field.calcCircularPos(state.pos);
		float cd = field.calcCircularDeg(state.pos.x);
		shape.draw(gl, p, cd, state.deg);
	}
}
