/*
 * $Id: rand.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.util;

//private import std.stream;
//private import std.date;

/**
 * Random number generator.
 */
//public template StaticRandImpl() {
// protected:
//  static Rand rand = null;
//
//  public static void setRandSeed(long seed) {
//    if (!rand)
//      rand = new Rand;
//    rand.setSeed(seed);
//  }
//}

public class Rand {

	public Rand() {
		init_genrand((int)System.currentTimeMillis());
	}

	public void setSeed(long n) {
		init_genrand((int)n);
	}

	public long nextInt32() {
		return genrand_int32();
	}

	public int nextInt(int n) {
		if (n == 0)
			return 0;
		else
			return genrand_int32() % n;
	}

	public int nextSignedInt(int n) {
		if (n == 0)
			return 0;
		else
			return genrand_int32() % (n * 2 + 1) - n;
	}

	public float nextFloat() {
		return nextFloat(1);
	}
	public float nextFloat(float n) {
		return (float)genrand_real1() * n;
	}

	public float nextSignedFloat() {
		return nextSignedFloat(1);
	}
	public float nextSignedFloat(float n) {
		return (float)genrand_real1() * (n * 2) - n;
	}

/* 
   MT.d
   Mersenne Twister random number generator -- D
   Based on code by Makoto Matsumoto, Takuji Nishimura, Shawn Cokus,
     Matthe Bellew, and Isaku Wada
   Andrew C. Edwards  v0.1  30 September 2003  edwardsac@ieee.org

   Before using, initialize the state by using init_genrand(seed) 
   or init_by_array(init_key, key_length).

   Copyright (C) 1997 - 2002, Makoto Matsumoto and Takuji Nishimura,
   Copyright (C) 2003, Andrew C. Edwards
   All rights reserved.

   Redistribution and use in source and binary forms, with or without
   modification, are permitted provided that the following conditions
   are met:

     1. Redistributions of source code must retain the above copyright
        notice, this list of conditions and the following disclaimer.

     2. Redistributions in binary form must reproduce the above copyright
        notice, this list of conditions and the following disclaimer in the
        documentation and/or other materials provided with the distribution.

     3. The names of its contributors may not be used to endorse or promote 
        products derived from this software without specific prior written 
        permission.

   THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
   "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
   LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
   A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE COPYRIGHT OWNER OR
   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
   EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
   PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
   PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
   LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
   NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
   SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

   The original code included the following notice:
  
     Any feedback is very welcome.
     http://www.math.keio.ac.jp/matumoto/emt.html
     email: matumoto@math.keio.ac.jp
  
   Please CC: edwardsac@ieee.org on all correspondence
*/

/* 
   Modified by Kenta Cho.
   Remove 'static' to wrap with a Rand class.
*/

	/* Period parameters */  
	private static final int N = 624;
	private static final int M = 397;
	private static final int MATRIX_A = 0x9908b0df;   /* constant vector a */
	private static final int UMASK = 0x80000000; /* most significant w-r bits */
	private static final int LMASK = 0x7fffffff; /* least significant r bits */
	private static final int MIXBITS(int u, int v) { return (u & UMASK) | (v & LMASK); }
	private static final int TWIST(int u, int v) { return (MIXBITS(u,v) >> 1) ^ ((v&1) != 0 ? MATRIX_A : 0); }

	int[] state = new int[N]; /* the array for the state vector  */
	int left = 1;
	int initf = 0;
	int next = 0;

/* initializes state[N] with a seed */
	void init_genrand(int s)
	{
		state[0]= s & 0xffffffff;
		for (int j=1; j<N; j++) {
			state[j] = (1812433253 * (state[j-1] ^ (state[j-1] >> 30)) + j); 
			/* See Knuth TAOCP Vol2. 3rd Ed. P.106 for multiplier. */
			/* In the previous versions, MSBs of the seed affect   */
			/* only MSBs of the array state[].                        */
			/* 2002/01/09 modified by Makoto Matsumoto             */
			state[j] &= 0xffffffff;  /* for >32 bit machines */
		}
		left = 1; initf = 1;
	}
	
	/* initialize by an array with array-length */
	/* init_key is the array for initializing keys */
	/* key_length is its length */
//	uint init_key[];
//	uint key_length;
	void init_by_array(int init_key[], int key_length)
	{
		int i, j, k;
		init_genrand(19650218);
		i=1; j=0;
		k = (N>key_length ? N : key_length);
		for (; k != 0; k--) {
			state[i] = (state[i] ^ ((state[i-1] ^ (state[i-1] >> 30)) * 1664525))
			+ init_key[j] + j; /* non linear */
			state[i] &= 0xffffffff; /* for WORDSIZE > 32 machines */
			i++; j++;
			if (i>=N) { state[0] = state[N-1]; i=1; }
			if (j>=key_length) j=0;
		}
		for (k=N-1; k != 0; k--) {
			state[i] = (state[i] ^ ((state[i-1] ^ (state[i-1] >> 30)) * 1566083941))
			- i; /* non linear */
			state[i] &= 0xffffffff; /* for WORDSIZE > 32 machines */
			i++;
			if (i>=N) { state[0] = state[N-1]; i=1; }
		}
		
		state[0] = 0x80000000; /* MSB is 1; assuring non-zero initial array */ 
		left = 1; initf = 1;
	}
	
	void next_state()
	{
		int p[]=state;
		
		/* if init_genrand() has not been called, */
		/* a default initial seed is used         */
		if (initf==0) init_genrand(5489);
		
		left = N;
		next = 0;
		
		for (int j=0; j<N-M; j++)
			p[j]=p[j+M] ^ TWIST(p[j], p[j+1]);
//		for (int j=N-M+1; --j; p++) 
//			*p = p[M] ^ TWIST(p[0], p[1]);

		for (int j=N-M; j<N-1; j++) 
			p[j] = p[j+M-N] ^ TWIST(p[j], p[j+1]);
//		for (int j=M; --j; p++) 
//			*p = p[M-N] ^ TWIST(p[0], p[1]);

		p[N-1] = p[M-1] ^ TWIST(p[N-1], state[0]);
//		*p = p[M-N] ^ TWIST(p[0], state[0]);
	}
	
	/* generates a random number on [0,0xffffffff]-interval */
	int genrand_int32()
	{
		int y;

		if (--left == 0) next_state();
		y = state[next++];

		/* Tempering */
		y ^= (y >> 11);
		y ^= (y << 7) & 0x9d2c5680;
		y ^= (y << 15) & 0xefc60000;
		y ^= (y >> 18);

		return y;
	}

	/* generates a random number on [0,0x7fffffff]-interval */
	long genrand_int31()
	{
		int y;

		if (--left == 0) next_state();
		y = state[next++];

		/* Tempering */
		y ^= (y >> 11);
		y ^= (y << 7) & 0x9d2c5680;
		y ^= (y << 15) & 0xefc60000;
		y ^= (y >> 18);

		return (long)(y>>1);
	}

	/* generates a random number on [0,1]-real-interval */
	double genrand_real1()
	{
		int y;

		if (--left == 0) next_state();
		y = state[next++];

		/* Tempering */
		y ^= (y >> 11);
		y ^= (y << 7) & 0x9d2c5680;
		y ^= (y << 15) & 0xefc60000;
		y ^= (y >> 18);
		
		return (double)y * (1.0/4294967295.0); 
		/* divided by 2^32-1 */ 
	}

	/* generates a random number on [0,1)-real-interval */
	double genrand_real2()
	{
		int y;
		
		if (--left == 0) next_state();
		y = state[next++];

		/* Tempering */
		y ^= (y >> 11);
		y ^= (y << 7) & 0x9d2c5680;
		y ^= (y << 15) & 0xefc60000;
		y ^= (y >> 18);
		
		return (double)y * (1.0/4294967296.0); 
		/* divided by 2^32 */
	}

	/* generates a random number on (0,1)-real-interval */
	double genrand_real3()
	{
		int y;

		if (--left == 0) next_state();
		y = state[next++];

		/* Tempering */
		y ^= (y >> 11);
		y ^= (y << 7) & 0x9d2c5680;
		y ^= (y << 15) & 0xefc60000;
		y ^= (y >> 18);

		return ((double)y + 0.5) * (1.0/4294967296.0); 
		/* divided by 2^32 */
	}

	/* generates a random number on [0,1) with 53-bit resolution*/
	double genrand_res53() 
	{ 
		int a=genrand_int32()>>5, b=genrand_int32()>>6; 
		return(a*67108864.0+b)*(1.0/9007199254740992.0); 
	}

}
