/*
 * $Id: vector.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2004 Kenta Cho. Some rights reserved.
 */
package abagames.util;

//private import std.math;
//private import std.string;

/**
 * 2D vector.
 */
public class Vector2 {
	public float x, y;
	private static Vector2 rsl = new Vector2();

	public Vector2() {
		clear();
	}

	public Vector2(float x, float y) {
		this.x = x;
		this.y = y;
	}

	public void clear() {
		x = y = 0;
	}

	public float opMul(Vector2 v) {
		return x * v.x + y * v.y;
	}

	public Vector2 getElement(Vector2 v) {
		float ll = v.opMul(v);
		if (ll > 0.1f) {
			float mag = this.opMul(v);
			rsl.x = mag * x / ll;
			rsl.y = mag * y / ll;
		} else {
			rsl.x = rsl.y = 0;
		}
		return rsl;
	}

	public Vector2 getElement(Vector2 v, float min, float max) {
		float ll = v.opMul(v);
		if (ll > 0.1f) {
			float mag = this.opMul(v);
			mag /= ll;
			rsl.x = mag * x;
			rsl.y = mag * y;
		} else {
			rsl.x = rsl.y = 0;
		}
		float d = rsl.vctSize();
		if (d > 0.1f && d < min)
			rsl.opMulAssign(min / d);
		else if (d > max)
			rsl.opMulAssign(max / d);
		return rsl;
  }

	public void opAddAssign(Vector2 v) {
		x += v.x;
		y += v.y;
	}

	public void opSubAssign(Vector2 v) {
		x -= v.x;
		y -= v.y;
	}

	public void opMulAssign(float a) {
		x *= a;
		y *= a;
	}

	public void opDivAssign(float a) {
		x /= a;
		y /= a;
	}

	public float checkSide(Vector2 pos1, Vector2 pos2) {
		float xo = pos2.x - pos1.x;
		float yo = pos2.y - pos1.y;
		if (xo == 0) {
			if (yo == 0)
				return 0;
			if (yo > 0)
				return x - pos1.x;
			else
				return pos1.x - x;
		} else if (yo == 0) {
			if (xo > 0)
				return pos1.y - y;
			else
				return y - pos1.y;
		} else {
			if (xo * yo > 0)
				return (x - pos1.x) / xo - (y - pos1.y) / yo;
			else
				return -(x - pos1.x) / xo + (y - pos1.y) / yo;
		}
	}

	public float checkSide(Vector2 pos1, Vector2 pos2, Vector2 ofs) {
		float xo = pos2.x - pos1.x;
		float yo = pos2.y - pos1.y;
		float mx = x + ofs.x;
		float my = y + ofs.y;
		if (xo == 0) {
			if (yo == 0)
				return 0;
			if (yo > 0)
				return mx - pos1.x;
			else
				return pos1.x - mx;
		} else if (yo == 0) {
			if (xo > 0)
				return pos1.y - my;
			else
				return my - pos1.y;
		} else {
			if (xo * yo > 0)
				return (mx - pos1.x) / xo - (my - pos1.y) / yo;
			else
				return -(mx - pos1.x) / xo + (my - pos1.y) / yo;
		}
	}

	public boolean checkCross(Vector2 p, Vector2 p1, Vector2 p2, float width) {
		float a1x, a1y, a2x, a2y;
		if (x < p.x) {
			a1x = x - width; a2x = p.x + width;
		} else {
			a1x = p.x - width; a2x = x + width;
		}
		if (y < p.y) {
			a1y = y - width; a2y = p.y + width;
		} else {
			a1y = p.y - width; a2y = y + width;
		}
		float b1x, b1y, b2x, b2y;
		if (p2.y < p1.y) {
			b1y = p2.y - width; b2y = p1.y + width;
		} else {
			b1y = p1.y - width; b2y = p2.y + width;
		}
		if (a2y >= b1y && b2y >= a1y) {
			if (p2.x < p1.x) {
				b1x = p2.x - width; b2x = p1.x + width;
			} else {
				b1x = p1.x - width; b2x = p2.x + width;
			}
			if (a2x >= b1x && b2x >= a1x) {
				float a = y - p.y;
				float b = p.x - x;
				float c = p.x * y - p.y * x;
				float d = p2.y - p1.y;
				float e = p1.x - p2.x;
				float f = p1.x * p2.y - p1.y * p2.x;
				float dnm = b * d - a * e;
				if (dnm != 0) {
					float x = (b*f - c*e) / dnm;
					float y = (c*d - a*f) / dnm;
					if (a1x <= x && x <= a2x && a1y <= y && y <= a2y &&
							b1x <= x && x <= b2x && b1y <= y && y <= b2y)
						return true;
				}
			}
		}
		return false;
	}

	public boolean checkHitDist(Vector2 p, Vector2 pp, float dist) {
		float bmvx, bmvy, inaa;
		bmvx = pp.x;
		bmvy = pp.y;
		bmvx -= p.x;
		bmvy -= p.y;
		inaa = bmvx * bmvx + bmvy * bmvy;
		if (inaa > 0.00001) {
			float sofsx, sofsy, inab, hd;
			sofsx = x;
			sofsy = y;
			sofsx -= p.x;
			sofsy -= p.y;
			inab = bmvx * sofsx + bmvy * sofsy;
			if (inab >= 0 && inab <= inaa) {
				hd = sofsx * sofsx + sofsy * sofsy - inab * inab / inaa;
				if (hd >= 0 && hd <= dist)
					return true;
			}
		}
		return false;
	}

	public float vctSize() {
		return (float)Math.sqrt(x * x + y * y);
	}

	public float dist(Vector2 v) {
		return dist(v.x, v.y);
	}

	public float dist() {
		return dist(0, 0);
	}
	public float dist(float px, float py) {
		float ax = Math.abs(x - px);
		float ay = Math.abs(y - py);
		if (ax > ay)
			return ax + ay / 2;
		else
			return ay + ax / 2;
	}

	public float distAcc(Vector2 v) {
		return (float)Math.sqrt((v.x - x) * (v.x - x) + (v.y - y) * (v.y - y));
	}

	public boolean contains(Vector2 p) {
		return contains(p, 1);
	}
	
	public boolean contains(Vector2 p, float r) {
		return contains(p.x, p.y, r);
	}

	public boolean contains(float px, float py) {
		return contains(px, py, 1);
	}
	public boolean contains(float px, float py, float r) {
		if (px >= -x * r && px <= x * r && py >= -y * r && py <= y * r)
			return true;
		else
			return false;
	}

	public void roll(float d) {
		double tx = x * Math.cos(d) - y * Math.sin(d);
		y = (float)(x * Math.sin(d) + y * Math.cos(d));
		x = (float)tx;
	}

	public String toString() {
		return "(" + x + ", " + y + ")";
	}

}