package abagames.util;

import abagames.ttn.Field;

/**
 * 3D vector.
 */
public class Vector3 {
	public float x, y, z;
	private static Vector3 rsl = new Vector3();

	public Vector3() {
		clear();
	}

	public Vector3(float x, float y, float z) {
		this.x = x;
		this.y = y;
		this.z = z;
	}

	final public void clear() {
		x = y = z = 0;
	}

	final public void rollX(float d) {
		double ty = y * Math.cos(d) - z * Math.sin(d);
		z = (float)(y * Math.sin(d) + z * Math.cos(d));
		y = (float)ty;
	}

	final public void rollY(float d) {
		double tx = x * Math.cos(d) - z * Math.sin(d);
		z = (float)(x * Math.sin(d) + z * Math.cos(d));
		x = (float)tx;
	}

	final public void rollZ(float d) {
		double tx = x * Math.cos(d) - y * Math.sin(d);
		y = (float)(x * Math.sin(d) + y * Math.cos(d));
		x = (float)tx;
	}

	final public void blend(Vector3 v1, Vector3 v2, float ratio) {
		x = v1.x * ratio + v2.x * (1 - ratio);
		y = v1.y * ratio + v2.y * (1 - ratio);
		z = v1.z * ratio + v2.z * (1 - ratio);
	}

	final public float vctSize() {
		return (float)Math.sqrt(x * x + y * y + z * z);
	}

	final public float dist(Vector3 v) {
		return dist(v.x, v.y, v.z);
	}

	final public float dist() {
		return dist(0, 0, 0);
	}
	
	final public float dist(float px, float py, float pz) {
		float ax = Math.abs(x - px);
		float ay = Math.abs(y - py);
		float az = Math.abs(z - pz);
		float axy;
		if (ax > ay)
			axy = ax + ay / 2;
		else
			axy = ay + ax / 2;
		if (axy > az)
			return axy + az / 2;
		else
			return az + axy / 2;
	}

	final public Vector3 getElement(Vector3 v) {
		float ll = v.opMul(v);
		if (ll != 0) {
			float mag = this.opMul(v);
			rsl.x = mag * v.x / ll;
			rsl.y = mag * v.y / ll;
			rsl.z = mag * v.z / ll;
		} else {
			rsl.x = rsl.y = rsl.z = 0;
		}
		return rsl;
	}
	
	final public float opMul(Vector3 v) {
		return x * v.x + y * v.y + z * v.z;
	}

	final public void opAddAssign(Vector3 v) {
		x += v.x;
		y += v.y;
		z += v.z;
	}

	final public void opSubAssign(Vector3 v) {
		x -= v.x;
		y -= v.y;
		z -= v.z;
	}

	final public void opMulAssign(float a) {
		x *= a;
		y *= a;
		z *= a;
	}

	final public void opDivAssign(float a) {
		x /= a;
		y /= a;
		z /= a;
	}

	final public String toString() {
		return "(" + (x) + ", " + (y) + ", " + (z) + ")";
	}


	final public void rollX2(float d) {
		double ty = y * Field.cos(d) - z * Field.sin(d);
		z = (float)(y * Field.sin(d) + z * Field.cos(d));
		y = (float)ty;
	}

	final public void rollY2(float d) {
		double tx = x * Field.cos(d) - z * Field.sin(d);
		z = (float)(x * Field.sin(d) + z * Field.cos(d));
		x = (float)tx;
	}

	final public void rollZ2(float d) {
		double tx = x * Field.cos(d) - y * Field.sin(d);
		y = (float)(x * Field.sin(d) + y * Field.cos(d));
		x = (float)tx;
	}
}