/**
 * 
 */
package abagames.util.sdl;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.util.ArrayList;

public class InputRecord<T extends Recordable> {
	private ArrayList<Record<T>> record;
	private int idx, series;
	private T replayData;
	private T org;

	public InputRecord(T t) {
//		replayData = T.newInstance();
		org = t;
		replayData = (T)org.clone();
		record = new ArrayList<Record<T>>(30 * 60 * 10);
	}

	public void clear() {
		record.clear();
	}

	public void add(T d) {
		if (record != null && record.size() > 0 && record.get(record.size() - 1).data.equals(d)) {
			record.get(record.size() - 1).series++;
		} else {
			Record<T> r = new Record<T>();
			r.data = (T)org.clone();
			r.data.set(d);
			r.series = 1;
			record.add(r);
		}
	}

	public void reset() {
		idx = 0;
		series = 0;
	}

	public boolean hasNext() {
		if (idx >= record.size())
			return false;
		else
			return true;
	}

	public T next() {
		if (idx >= record.size())
			throw new NoRecordDataException("No more items");
		if (series <= 0)
			series = record.get(idx).series;
		replayData.set(record.get(idx).data);
		series--;
		if (series <= 0)
			idx++;
		return replayData;
	}

	public void save(DataOutputStream fd) {
		try {
			fd.write(record.size());
			for (Record r: record) {
				fd.writeInt(r.series);
				r.data.write(fd);
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}

	@SuppressWarnings("unchecked")
	public void load(DataInputStream fd) {
		clear();
		int l, s;
		T d;
		try {
			l = fd.readInt();
			for (int i = 0; i < l; i++) {
				s = fd.readInt();
				d = (T)org.clone();
				d.read(fd);
				Record r = new Record();
				r.series = s;
				r.data = d;
				record.add(r);
			}
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
	}
	static class Record<A extends Recordable> {
		int series;
		A data;
	}
}