/*
 * $Id: mainloop.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.util.sdl;

import javax.media.opengl.GL;

import sdl.SDL_Event;

import abagames.util.Logger;
import abagames.util.Preference;

//private import SDL;
//private import abagames.util.logger;
//private import abagames.util.rand;
//private import abagames.util.preference;
//private import abagames.util.sdl.frame;
//private import abagames.util.sdl.screen;
//private import abagames.util.sdl.input;
//private import abagames.util.sdl.sound;
//private import abagames.util.sdl.sdlexception;

/**
 * SDL main loop.
 */
public class MainLoop {
	public static final int INTERVAL_BASE = 16;
	public static final int MAX_SKIP_FRAME = 5;
	public boolean noSlowdown = false;
	public SDL_Event event;
	private Screen screen;
	private Input input;
	private Frame frame;
	private Preference preference;
	private float slowdownRatio;
	private float interval = INTERVAL_BASE;
	private float _slowdownStartRatio = 1;
	private float _slowdownMaxRatio = 1.5f;
	private long prvTickCount = -1;

	public MainLoop(Screen screen, Input input, Frame frame, Preference preference) {
		this.screen = screen;
		this.input = input;
		this.frame = frame;
		this.preference = preference;
		frame.setMainLoop(this);
		frame.setUIs(screen, input);
		frame.setPreference(preference);
		event = new SDL_Event();
	}

  // Initialize and load a preference.
	private void initFirst(GL gl) {
		preference.load();
		try {
			Sound.init();
		} catch (SDLInitFailedException e) {
			Logger.error(e);
		}
		frame.init(gl);
		initInterval();
	}

	// Quit and save a preference.
	private void quitLast(GL gl) {
		frame.quit(gl);
		Sound.close();
		preference.save();
		screen.closeSDL();
//		SDL_Quit();
	}

	private boolean done;

	public void breakLoop() {
		done = true;
	}

//	public void loop(GL gl) {
//		done = false;
//		long prvTickCount = 0;
//		int i;
//		long nowTick;
//		int frameNum;
//		screen.initSDL();
//		initFirst(gl);
//		frame.start(gl);
//		while (!done) {
//			//TODO implements main loop
////			if (SDL_PollEvent(&event) == 0)
////				event.type = SDL_USEREVENT;
////			input.handleEvent(&event);
////			if (event.type == SDL_QUIT)
////				breakLoop();
////			nowTick = SDL_GetTicks();
////			int itv = cast(int) interval;
////			frameNum = cast(int) (nowTick - prvTickCount) / itv;
////			if (frameNum <= 0) {
////				frameNum = 1;
////				SDL_Delay(prvTickCount + itv - nowTick);
////				prvTickCount += interval;
////			} else if (frameNum > MAX_SKIP_FRAME) {
////				frameNum = MAX_SKIP_FRAME;
////				prvTickCount = nowTick;
////			} else {
////				//prvTickCount += frame * interval;
////				prvTickCount = nowTick;
////			}
////			slowdownRatio = 0;
////			for (i = 0; i < frameNum; i++)
////				frame.move();
////			slowdownRatio /= frameNum;
////			screen.clear();
////			frame.draw();
////			screen.flip();
////			if (!noSlowdown)
////				calcInterval();
//
//			//TODO implements main loop
////			if (SDL_PollEvent(&event) == 0)
////				event.type = SDL_USEREVENT;
////			input.handleEvent(&event);
////			if (event.type == SDL_QUIT)
////				breakLoop();
////			nowTick = SDL_GetTicks();
////			int itv = cast(int) interval;
////			frameNum = cast(int) (nowTick - prvTickCount) / itv;
////			if (frameNum <= 0) {
////				frameNum = 1;
////				SDL_Delay(prvTickCount + itv - nowTick);
////				prvTickCount += interval;
////			} else if (frameNum > MAX_SKIP_FRAME) {
////				frameNum = MAX_SKIP_FRAME;
////				prvTickCount = nowTick;
////			} else {
////				//prvTickCount += frame * interval;
////				prvTickCount = nowTick;
////			}
////			slowdownRatio = 0;
////			for (i = 0; i < frameNum; i++)
////				frame.move();
////			slowdownRatio /= frameNum;
////			screen.clear();
////			frame.draw();
////			screen.flip();
////			if (!noSlowdown)
////				calcInterval();
////			if (SDL_PollEvent(&event) == 0)
////				event.type = SDL_USEREVENT;
////			input.handleEvent(&event);
////			if (event.type == SDL_QUIT)
////				breakLoop();
//			nowTick = System.currentTimeMillis();
//			int itv = (int) interval;
//			frameNum = (int) (nowTick - prvTickCount) / itv;
//			if (frameNum <= 0) {
//				frameNum = 1;
////				SDL_Delay(prvTickCount + itv - nowTick);
//				try {
////					System.out.println("sleep " + (prvTickCount + itv - nowTick));
//					Thread.sleep(prvTickCount + itv - nowTick);
//				} catch (InterruptedException e) {
//					e.printStackTrace();
//				}
//				prvTickCount += interval;
//			} else if (frameNum > MAX_SKIP_FRAME) {
//				frameNum = MAX_SKIP_FRAME;
//				prvTickCount = nowTick;
//			} else {
//				//prvTickCount += frame * interval;
//				prvTickCount = nowTick;
//			}
//			slowdownRatio = 0;
//			for (i = 0; i < frameNum; i++)
//				frame.move(gl);
//			slowdownRatio /= frameNum;
//			screen.clear();
//			frame.draw(gl);
//			screen.flip();
//			if (!noSlowdown)
//				calcInterval();
//		}
//		quitLast(gl);
//	}
//	

	public void init(GL gl) {
		screen.initSDL();
		initFirst(gl);
		frame.start(gl);
		prvTickCount = 0;
	}
	
	private long start = Long.MIN_VALUE; 
	public long getTick() {
		if (start == Long.MIN_VALUE)
			start = System.nanoTime() / 1000000;
		return System.nanoTime() / 1000000 - start;
	}
	
	public boolean step(GL gl) {
//		System.out.println("step");
		done = false;
		int i;
		long nowTick;
		int frameNum;
//		long t0 = System.nanoTime();
//		if (SDL_PollEvent(&event) == 0)
//		event.type = SDL_USEREVENT;
		input.handleEvent(event);
		if (event.type == SDL_Event.EventType.SDL_QUIT)
			breakLoop();
//		long t1 = System.nanoTime();
		nowTick = getTick();
		int itv = (int) interval;
		frameNum = (int) (nowTick - prvTickCount) / itv;
//		System.out.println(nowTick + " " + prvTickCount + " " + itv + " " + frameNum);
		if (prvTickCount < 0)
			System.out.println("****************");
		if (frameNum <= 0) {
			frameNum = 1;
//			SDL_Delay(prvTickCount + itv - nowTick);
			try {
				long slp = prvTickCount + itv - nowTick;
//				System.out.println("sleep " + slp);
				if (slp > 0 && slp < 1000)
					Thread.sleep(slp);
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			prvTickCount += interval;
		} else if (frameNum > MAX_SKIP_FRAME) {
			frameNum = MAX_SKIP_FRAME;
			prvTickCount = nowTick;
		} else {
			//prvTickCount += frame * interval;
			prvTickCount = nowTick;
		}
//		long t2 = System.nanoTime();
		slowdownRatio = 0;
		for (i = 0; i < frameNum; i++)
			frame.move(gl);
		slowdownRatio /= frameNum;
//		long t3 = System.nanoTime();
		screen.clear();
//		long t4 = System.nanoTime();
		frame.draw(gl);
//		long t5 = System.nanoTime();
		screen.flip();
		if (!noSlowdown)
			calcInterval();
//		long t6 = System.nanoTime();
//		System.out.println("step " + (t6 - t0) / 1000000.0 + " " + (t5 - t4) / 1000000.0);
//		System.out.println((t1 - t0) / 1000000.0 + " " + (t2 - t1) / 1000000.0 + " " + (t3 - t2) / 1000000.0
//				+ " " + (t4 - t3) / 1000000.0 + " " + (t5 - t4) / 1000000.0 + " " + (t6 - t5) / 1000000.0);
		return done;
	}
	public void quit(GL gl) {
		quitLast(gl);
	}

	// Intentional slowdown.

	public void initInterval() {
		interval = INTERVAL_BASE;
	}

	public void addSlowdownRatio(float sr) {
		slowdownRatio += sr;
	}

	private void calcInterval() {
		if (slowdownRatio > _slowdownStartRatio) {
			float sr = slowdownRatio / _slowdownStartRatio;
			if (sr > _slowdownMaxRatio)
				sr = _slowdownMaxRatio;
			interval += (sr * INTERVAL_BASE - interval) * 0.1;
		} else {
			interval += (INTERVAL_BASE - interval) * 0.08;
		}
	}

	public float slowdownStartRatio(float v) {
		return _slowdownStartRatio = v;
	}

	public float slowdownMaxRatio(float v) {
		return _slowdownMaxRatio = v;
	}
}
