/*
 * $Id: pad.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2006 Kenta Cho. Some rights reserved.
 */
package abagames.util.sdl;

import java.awt.event.KeyEvent;
import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;

import sdl.SDL_Event;

/**
 * Inputs from a joystick and a keyboard.
 */
public class Pad implements Input {
	public static class SDL_Joystick {
		
	}
	public int[] keys;
	public boolean buttonsExchanged = false;
	
	private SDL_Joystick stick = null;
	private final int JOYSTICK_AXIS = 16384;
	private PadState state;

	public Pad() {
		state = new PadState();
		keys = new int[256];
	}

//	public SDL_Joystick openJoystick(SDL_Joystick st) {
	public SDL_Joystick openJoystick() {
//		if (st == null) {
//			if (SDL_InitSubSystem(SDL_INIT_JOYSTICK) < 0)
//				return null;
//			stick = SDL_JoystickOpen(0);
//		} else {
//		stick = st;
//		}
		return stick;
	}

	public void handleEvent(SDL_Event event) {
//		keys = SDL_GetKeyState(null);
	}

	public void handleEvent(KeyEvent event) {
		final int code = event.getKeyCode();
		if (event.getID() == KeyEvent.KEY_PRESSED) {
			if (code >= 0 && code < keys.length) {
				keys[code] = SDL_PRESSED;
//				System.out.println(code + " pressed");
			}
		}
		if (event.getID() == KeyEvent.KEY_RELEASED) {
			if (code >= 0 && code < keys.length) {
				keys[code] = 0;
//				System.out.println(code + " released")
			};
		}
//		System.out.println(code + " handled " + event.getID());
	}
	
	public final static int SDL_PRESSED = 1;
	public final static int SDLK_LEFT = KeyEvent.VK_LEFT;
	public final static int SDLK_RIGHT = KeyEvent.VK_RIGHT;
	public final static int SDLK_UP = KeyEvent.VK_UP;
	public final static int SDLK_DOWN = KeyEvent.VK_DOWN;
	public final static int SDLK_KP2 = KeyEvent.VK_NUMPAD2;
	public final static int SDLK_KP4 = KeyEvent.VK_NUMPAD4;
	public final static int SDLK_KP6 = KeyEvent.VK_NUMPAD6;
	public final static int SDLK_KP8 = KeyEvent.VK_NUMPAD8;
	public final static int SDLK_LALT = KeyEvent.VK_ALT;
	public final static int SDLK_LCTRL = KeyEvent.VK_CONTROL;
	public final static int SDLK_LSHIFT = KeyEvent.VK_SHIFT;
	public final static int SDLK_PERIOD = KeyEvent.VK_PERIOD;
	public final static int SDLK_RALT = KeyEvent.VK_ALT;
	public final static int SDLK_RCTRL = KeyEvent.VK_CONTROL;
	public final static int SDLK_RETURN = KeyEvent.VK_ENTER;
	public final static int SDLK_RSHIFT = KeyEvent.VK_SHIFT;
	public final static int SDLK_SLASH = KeyEvent.VK_SLASH;
	public final static int SDLK_ESCAPE = KeyEvent.VK_ESCAPE;
	public final static int SDLK_a = KeyEvent.VK_A;
	public final static int SDLK_d = KeyEvent.VK_D;
	public final static int SDLK_i = KeyEvent.VK_I;
	public final static int SDLK_j = KeyEvent.VK_J;
	public final static int SDLK_k = KeyEvent.VK_K;
	public final static int SDLK_l = KeyEvent.VK_L;
	public final static int SDLK_p = KeyEvent.VK_P;
	public final static int SDLK_s = KeyEvent.VK_S;
	public final static int SDLK_w = KeyEvent.VK_W;
	public final static int SDLK_x = KeyEvent.VK_X;
	public final static int SDLK_z = KeyEvent.VK_Z;
	
	public PadState getState() {
		int x = 0, y = 0;
		state.dir = 0;
		if (stick != null) {
//			x = SDL_JoystickGetAxis(stick, 0);
//			y = SDL_JoystickGetAxis(stick, 1);
		}
		if (keys[SDLK_RIGHT] == SDL_PRESSED || keys[SDLK_KP6] == SDL_PRESSED || 
				keys[SDLK_d] == SDL_PRESSED || keys[SDLK_l] == SDL_PRESSED ||
				x > JOYSTICK_AXIS)
			state.dir |= PadState.RIGHT;
		if (keys[SDLK_LEFT] == SDL_PRESSED || keys[SDLK_KP4] == SDL_PRESSED ||
				keys[SDLK_a] == SDL_PRESSED || keys[SDLK_j] == SDL_PRESSED ||
				x < -JOYSTICK_AXIS)
			state.dir |= PadState.LEFT;
		if (keys[SDLK_DOWN] == SDL_PRESSED || keys[SDLK_KP2] == SDL_PRESSED ||
				keys[SDLK_s] == SDL_PRESSED || keys[SDLK_k] == SDL_PRESSED ||
				y > JOYSTICK_AXIS)
			state.dir |= PadState.DOWN;
		if (keys[SDLK_UP] == SDL_PRESSED ||  keys[SDLK_KP8] == SDL_PRESSED ||
				keys[SDLK_w] == SDL_PRESSED || keys[SDLK_i] == SDL_PRESSED ||
				y < -JOYSTICK_AXIS)
			state.dir |= PadState.UP;
//		System.out.println(state.dir);
		state.button = 0;
		int btn1 = 0, btn2 = 0;
		float leftTrigger = 0, rightTrigger = 0;
		if (stick != null) {
//			btn1 = SDL_JoystickGetButton(stick, 0) + SDL_JoystickGetButton(stick, 2) +
//			SDL_JoystickGetButton(stick, 4) + SDL_JoystickGetButton(stick, 6) +
//			SDL_JoystickGetButton(stick, 8) + SDL_JoystickGetButton(stick, 10);
//			btn2 = SDL_JoystickGetButton(stick, 1) + SDL_JoystickGetButton(stick, 3) +
//			SDL_JoystickGetButton(stick, 5) + SDL_JoystickGetButton(stick, 7) +
//			SDL_JoystickGetButton(stick, 9) + SDL_JoystickGetButton(stick, 11);
		}
		if (keys[SDLK_z] == SDL_PRESSED || keys[SDLK_PERIOD] == SDL_PRESSED ||
				keys[SDLK_LCTRL] == SDL_PRESSED || keys[SDLK_RCTRL] == SDL_PRESSED || 
				btn1 != 0) {
			if (!buttonsExchanged)
				state.button |= PadState.A;
			else
				state.button |= PadState.B;
		}
		if (keys[SDLK_x] == SDL_PRESSED || keys[SDLK_SLASH] == SDL_PRESSED ||
				keys[SDLK_LALT] == SDL_PRESSED || keys[SDLK_RALT] == SDL_PRESSED ||
				keys[SDLK_LSHIFT] == SDL_PRESSED || keys[SDLK_RSHIFT] == SDL_PRESSED ||
				keys[SDLK_RETURN] == SDL_PRESSED ||
				btn2 != 0) {
			if (!buttonsExchanged)
				state.button |= PadState.B;
			else
				state.button |= PadState.A;
		}
		return state;
	}
	
	public PadState getNullState() {
		state.clear();
		return state;
	}


	static public class PadState implements Recordable {
		
		public static final int UP = 1, DOWN = 2, LEFT = 4, RIGHT = 8;
		public static final int A = 16, B = 32, ANY = 48;
		
		public int dir, button;
				
		public static PadState newInstance() {
			return new PadState();
		}
		
		public static PadState newInstance(PadState s) {
			return new PadState(s);
		}
		
		public PadState() {
		}
		
		public PadState(PadState s) {
			this();
			set(s);
		}
			
		public void set(PadState s) {
			dir = s.dir;
			button = s.button;
		}
			
		public void clear() {
			dir = button = 0;
		}
			
		public void read(DataInputStream fd) {
			int s = 0;;
			try {
				s = fd.readInt();
			} catch (IOException e) {
				e.printStackTrace();
			}
			dir = s & (UP | DOWN | LEFT | RIGHT);
			button = s & ANY;
		}
		
		public void write(DataOutputStream fd) {
			int s = dir | button;
			try {
				fd.writeInt(s);
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		
		public boolean equals(Object o) {
			if (o == null)
				return false;
			if (o == this)
				return true;
			if (o instanceof PadState) {
				PadState s = (PadState)o;
				if (dir == s.dir && button == s.button)
					return true;
				else
					return false;
			}
			return false;
		}

		public void set(Recordable r) {
			PadState s = (PadState)r;
			dir = s.dir;
			button = s.button;
		}
		
		@Override
		public PadState clone() {
			return new PadState();
		}
	}


	public static class RecordablePad extends Pad {
//		mixin RecordableInput!(PadState);
		public RecordableInput<PadState> input;
		
		public RecordablePad() {
			PadState p = new PadState();
			input = new RecordableInput<PadState>(p);
		}
		
		public PadState getState() {
			return getState(true);
		}
		
		public PadState getState(boolean doRecord) {
			PadState s = super.getState();
			if (doRecord)
				input.record(s);
			return s;
		}
	}
}
