/*
 * $Id: screen3d.d,v 1.1.1.1 2006/11/19 07:54:55 kenta Exp $
 *
 * Copyright 2005 Kenta Cho. Some rights reserved.
 */
package abagames.util.sdl;

import javax.media.opengl.GL;
import javax.media.opengl.GLCanvas;

import abagames.util.Vector2;
import abagames.util.Vector3;

//private import std.string;
//private import std.math;
//private import SDL;
//private import opengl;
//private import abagames.util.vector;
//private import abagames.util.sdl.screen;
//private import abagames.util.sdl.sdlexception;

/**
 * SDL screen handler (3D, OpenGL).
 */
abstract public class Screen3D extends GLCanvas implements Screen, SizableScreen {
	static float _brightness = 1f;
	float _farPlane = 1000f;
	float _nearPlane = 0.1f;
	int _width = 640;
	int _height = 480;
	boolean _windowMode = true;
	
	protected abstract void init(GL gl);
	protected abstract void close();
	protected void setIcon() {}
	
	public void initSDL() {
		GL gl = getGL();
		// Initialize SDL.
//		if (SDL_Init(SDL_INIT_VIDEO) < 0) {
//			throw new SDLInitFailedException(
//					"Unable to initialize SDL: " ~ std.string.toString(SDL_GetError()));
//		}
		setIcon();
		// Create an OpenGL screen.
//		Uint32 videoFlags;
//		if (_windowMode) {
//			videoFlags = SDL_OPENGL | SDL_RESIZABLE;
//		} else {
//			videoFlags = SDL_OPENGL | SDL_FULLSCREEN;
//		} 
//		if (SDL_SetVideoMode(_width, _height, 0, videoFlags) == null) {
//			throw new SDLInitFailedException
//			("Unable to create SDL screen: " ~ std.string.toString(SDL_GetError()));
//		}
		gl.glViewport(0, 0, width(), height());
		gl.glClearColor(0.0f, 0.0f, 0.0f, 0.0f);
		resized(_width, _height);
//		SDL_ShowCursor(SDL_DISABLE);
		init(gl);
	}

	// Reset a viewport when the screen is resized.
	public void screenResized() {
		GL gl = getGL();
		gl.glViewport(0, 0, _width, _height);
		gl.glMatrixMode(GL.GL_PROJECTION);
		setPerspective();
		gl.glMatrixMode(GL.GL_MODELVIEW);
	}

	public void setPerspective() {
		GL gl = getGL();
		gl.glLoadIdentity();
		//gluPerspective(45.0f, cast(GLfloat) width / cast(GLfloat) height, nearPlane, farPlane);
		gl.glFrustum(-_nearPlane,
				_nearPlane,
				-_nearPlane * (float) _height / (float) _width,
				_nearPlane * (float) _height / (float) _width,
				0.1f, _farPlane);
	}

	public void resized(int w, int h) {
		_width = w;
		_height = h;
		screenResized();
	}

	public void closeSDL() {
		close();
//		SDL_ShowCursor(SDL_ENABLE);
	}

	public void flip() {
		handleError();
//		SDL_GL_SwapBuffers();//TODO
	}

	public void clear() {
		GL gl = getGL();
		gl.glClear(GL.GL_COLOR_BUFFER_BIT);
	}

	public void handleError() {
		GL gl = getGL();
		int error = gl.glGetError();
		if (error == GL.GL_NO_ERROR)
			return;
		closeSDL();
		throw new RuntimeException("OpenGL error(" + error + ")");
	}

	protected void setCaption(String name) {
//		SDL_WM_SetCaption(std.string.toStringz(name), null);//TODO
	}

	public boolean windowMode(boolean v) {
		return _windowMode = v;
	}

	public boolean windowMode() {
		return _windowMode;
	}

	public int width(int v) {
		return _width = v;
	}

	public int width() {
		return _width;
	}

	public int height(int v) {
		return _height = v;
	}

	public int height() {
		return _height;
	}

	public static void glVertex(GL gl, Vector2 v) {
		gl.glVertex3f(v.x, v.y, 0);
	}

	public static void glVertex(GL gl, Vector3 v) {
		gl.glVertex3f(v.x, v.y, v.z);
	}

	public static void glTranslate(GL gl, Vector2 v) {
		gl.glTranslatef(v.x, v.y, 0);
	}

	public static void glTranslate(GL gl, Vector3 v) {
		gl.glTranslatef(v.x, v.y, v.z);
	}

	public static void glRotate(GL gl, float d) {
		glRotate(gl, d, 0, 0, 1);
	}
	
	public static void glRotate(GL gl, float d, float x, float y, float z) {
		gl.glRotated(d * 180 / Math.PI, x, y, z);
	}

	public static void setColor(GL gl, float r, float g, float b) {
		setColor(gl, r, g, b, 1);
	}
	public static void setColor(GL gl, float r, float g, float b, float a) {
		gl.glColor4f(r * _brightness, g * _brightness, b * _brightness, a);
	}

	public static void setClearColor(GL gl, float r, float g, float b) {
		setClearColor(gl, r, g, b, 1);
	}
	
	public static void setClearColor(GL gl, float r, float g, float b, float a) {
		gl.glClearColor(r * _brightness, g * _brightness, b * _brightness, a);
	}

	public static float brightness(float v) {
		return _brightness = v;
	}
}
